import axios from "axios";
import { ethers, } from "ethers";
import sdkConfig from "../config";
import { Bridge__factory, ERC20__factory } from "../contracts";
import { getFxTokenSymbolFromAddress } from "../utils/fxToken";
export default class Bridge {
    constructor(c) {
        this.deposit = async (args, signer, options = {}) => {
            const depositBridge = this.getBridgeContract(args.fromNetwork, signer);
            const tokenAddress = this.getTokenAddress(args.tokenSymbol);
            const isAmountSufficient = await this.doesDepositMeetMinimumAmount(args.tokenSymbol, args.amount, args.toNetwork);
            if (!isAmountSufficient)
                throw new Error("Deposit amount lower than withdrawal fee");
            return depositBridge.deposit(tokenAddress, args.amount, this.config.byNetwork[args.toNetwork].id, options);
        };
        /// A minimum amount must be enforced due to the withdrawal fees.
        this.doesDepositMeetMinimumAmount = async (tokenSymbol, amount, withdrawNetwork) => amount.gt(await this.getTokenFee(tokenSymbol, withdrawNetwork));
        /// Returns the withdraw token fee for the input network.
        this.getTokenFee = async (tokenSymbol, withdrawNetwork) => this.getBridgeContract(withdrawNetwork, sdkConfig.providers[withdrawNetwork]).tokenFees(this.getTokenAddress(tokenSymbol));
        this.withdraw = async (args, signer, options = {}) => {
            const bridgeContract = this.getBridgeContract(args.toNetwork, signer);
            const tokenAddress = this.getTokenAddress(args.tokenSymbol);
            const address = args.address ?? (await signer.getAddress());
            return bridgeContract.withdraw(address, tokenAddress, args.amount, args.nonce, this.config.byNetwork[args.fromNetwork].id, ethers.utils.arrayify(args.signature), options);
        };
        this.requestAutomaticWithdraw = async (network, txHash) => {
            const endpoint = `${this.config.apiBaseUrl}/withdraw`;
            const queryString = `network=${network}&transactionHash=${txHash}`;
            const response = await axios.put(`${endpoint}?${queryString}`);
            if (!response.data || !response.data.withdrawalTransactionHash)
                throw new Error("Automatic withdrawal request failed");
            return response.data.withdrawalTransactionHash;
        };
        this.getDepositAllowance = (account, token, network, signer) => {
            const tokenAddress = this.getTokenAddress(token);
            const bridgeAddress = this.config.byNetwork[network].address;
            const contract = ERC20__factory.connect(tokenAddress, signer);
            return contract.allowance(account, bridgeAddress);
        };
        this.setDepositAllowance = (token, network, amount, signer, options = {}) => {
            const tokenAddress = this.getTokenAddress(token);
            const bridgeAddress = this.config.byNetwork[network].address;
            const tokenContract = ERC20__factory.connect(tokenAddress, signer);
            return tokenContract.approve(bridgeAddress, amount, options);
        };
        this.getPendingWithdrawals = async (account, signers) => {
            const depositEventPromises = Object.keys(signers).map((n) => {
                const network = n;
                return this.getPendingWithdrawsForNetwork(account, network, signers);
            });
            const results = await Promise.all(depositEventPromises);
            return results.reduce((acc, curr) => [...acc, ...curr], []);
        };
        this.getWithdrawSignature = async (network, txHash) => {
            const { data } = await axios.get(`${this.config.apiBaseUrl}/sign?network=${network}&transactionHash=${txHash}`);
            return data.signature;
        };
        this.getWithdrawNonce = async (args, signer) => {
            const bridgeContract = this.getBridgeContract(args.toNetwork, signer);
            const account = args.address ?? (await signer.getAddress());
            return bridgeContract.withdrawNonce(account, this.config.byNetwork[args.fromNetwork].id);
        };
        this.getPendingWithdrawsForNetwork = async (account, network, signers) => {
            const signer = signers[network];
            const bridgeContract = this.getBridgeContract(network, signer);
            const fromBlock = network === "polygon"
                ? (await signer.provider.getBlockNumber()) - 1990
                : undefined;
            const filter = bridgeContract.filters.Deposit(account);
            const rawEvents = await bridgeContract.queryFilter(filter, fromBlock);
            const events = rawEvents.map((event) => ({
                txHash: event.transactionHash,
                ...bridgeContract.interface.parseLog(event)
                    .args,
            }));
            const eventsByNetwork = events.reduce((progress, event) => {
                const property = this.bridgeIdToNetwork(event.toId.toNumber());
                const values = progress[property] || [];
                return {
                    ...progress,
                    [property]: [...values, event],
                };
            }, {});
            const networksWithEvents = Object.keys(eventsByNetwork);
            const withdrawCounts = await Promise.all(networksWithEvents
                .filter((n) => n !== network)
                .map((n) => {
                const bc = this.getBridgeContract(n, signers[n]);
                return bc.withdrawNonce(account, this.config.byNetwork[network].id);
            }));
            const withdrawCountsByNetwork = withdrawCounts.reduce((progress, count, index) => {
                return {
                    ...progress,
                    [networksWithEvents[index]]: Number(count.toString()),
                };
            }, {});
            const pendingWithdraws = networksWithEvents.reduce((progress, network) => {
                const events = eventsByNetwork[network];
                const depositCount = events?.length || 0;
                const withdrawCount = withdrawCountsByNetwork[network || 0];
                const pendingCount = depositCount - (withdrawCount || 0);
                const pendingEvents = events?.slice(events.length - pendingCount, events.length) || [];
                return [...progress, ...pendingEvents];
            }, []);
            return pendingWithdraws.map((pw) => ({
                txHash: pw.txHash,
                tokenSymbol: this.getTokenSymbolFromAddress(pw.token),
                amount: pw.amount,
                nonce: pw.nonce,
                fromNetwork: this.bridgeIdToNetwork(pw.fromId.toNumber()),
                toNetwork: this.bridgeIdToNetwork(pw.toId.toNumber()),
            }));
        };
        this.getTokenAddress = (token) => {
            const tokenAddress = token === "FOREX"
                ? this.config.forexAddress
                : this.config.fxTokenAddresses[token];
            if (!tokenAddress) {
                throw new Error(`Invalid token symbol: ${token}`);
            }
            return tokenAddress;
        };
        this.getTokenSymbolFromAddress = (tokenAddress) => {
            if (tokenAddress === this.config.forexAddress) {
                return "FOREX";
            }
            return getFxTokenSymbolFromAddress(tokenAddress, this.config.fxTokenAddresses);
        };
        this.getBridgeContract = (network, signerOrProvider) => Bridge__factory.connect(this.config.byNetwork[network].address, signerOrProvider);
        this.bridgeIdToNetwork = (bridgeId) => {
            const networkNames = Object.keys(this.config.byNetwork);
            const ids = Object.values(this.config.byNetwork).map((x) => x.id);
            const index = ids.indexOf(bridgeId);
            return networkNames[index];
        };
        this.config = c || {
            apiBaseUrl: sdkConfig.bridge.apiBaseUrl,
            forexAddress: sdkConfig.forexAddress,
            byNetwork: sdkConfig.bridge.byNetwork,
            fxTokenAddresses: sdkConfig.fxTokenAddresses,
        };
    }
}
