import sdkConfig from "../config";
import { createERC20MulticallContract, createMulticallProtocolContracts, callMulticallObject, callMulticallObjects, } from "../utils/contract";
import Graph from "./graph";
import { ERC20__factory } from "../contracts";
import { BENTOBOX_ADDRESS } from "@sushiswap/core-sdk";
import { getTokensFromConfig } from "../utils/collateral";
import { NETWORK_NAME_TO_CHAIN_ID } from "../constants";
export default class Collaterals {
    constructor(c) {
        this.getCollateral = async (collateralSymbol, provider) => {
            const { provider: multicallProvider } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const collateral = this.findAvailable(collateralSymbol);
            const collateralMulticall = this.getCollateralMulticall(collateralSymbol, provider);
            const rawCollateral = await callMulticallObject(collateralMulticall, multicallProvider);
            return this.toCollateral(collateral, rawCollateral);
        };
        this.getCollaterals = async (provider) => {
            const { provider: multicallProvider } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const collateralMulticalls = this.tokens.map((a) => this.getCollateralMulticall(a.symbol, provider));
            const raw = await callMulticallObjects(collateralMulticalls, multicallProvider);
            return raw.map((c, index) => this.toCollateral(this.tokens[index], c));
        };
        this.getIndexedCollaterals = async () => {
            const collaterals = await this.graph.collateralGraphClient.query({});
            return collaterals.map(this.indexedToCollateral).filter((col) => {
                return this.tokens.find((t) => t.address.toLowerCase() === col.address.toLowerCase());
            });
        };
        this.getDepositAllowance = async (collateralSymbol, account, action, signer) => {
            const contract = this.getCollateralContract(collateralSymbol, signer);
            return contract.allowance(account, action === "deposit"
                ? this.config.protocolAddresses.treasury
                : this.config.protocolAddresses.comptroller);
        };
        this.getSingleCollateralDepositAllowance = async (vaultSymbol, account, network, signer) => {
            const kashiPool = sdkConfig.singleCollateralVaults[network][vaultSymbol];
            if (!kashiPool) {
                throw new Error(`Unable to find vault: ${network}-${vaultSymbol}`);
            }
            const collateral = ERC20__factory.connect(kashiPool.collateral.address, signer);
            const chainId = NETWORK_NAME_TO_CHAIN_ID[network];
            return collateral.allowance(account, BENTOBOX_ADDRESS[chainId]);
        };
        this.setDepositAllowance = (collateralSymbol, amount, action, signer, options = {}) => {
            const collateralContract = this.getCollateralContract(collateralSymbol, signer);
            return collateralContract.approve(action === "deposit"
                ? this.config.protocolAddresses.treasury
                : this.config.protocolAddresses.comptroller, amount, options);
        };
        this.setSingleCollateralDepositAllowance = (vaultSymbol, amount, network, signer, options = {}) => {
            const kashiPool = sdkConfig.singleCollateralVaults[network][vaultSymbol];
            if (!kashiPool) {
                throw new Error(`Unable to find vault: ${network}-${vaultSymbol}`);
            }
            const collateral = ERC20__factory.connect(kashiPool.collateral.address, signer);
            const chainId = NETWORK_NAME_TO_CHAIN_ID[network];
            return collateral.approve(BENTOBOX_ADDRESS[chainId], amount, options);
        };
        this.getCollateralContract = (collateralSymbol, signer) => {
            const avail = this.findAvailable(collateralSymbol);
            return ERC20__factory.connect(avail.address, signer);
        };
        this.getCollateralMulticall = (collateralSymbol, provider) => {
            const collateralAddress = this.config.collaterals[collateralSymbol]?.address;
            if (!collateralAddress) {
                throw new Error(`Collaterals not initialised with collateral that matches: ${collateralSymbol}`);
            }
            const { contracts } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const erc20MulticallContract = createERC20MulticallContract(collateralAddress);
            return {
                decimals: erc20MulticallContract.decimals(),
                collateralDetails: contracts.handle.getCollateralDetails(collateralAddress),
                price: contracts.handle.getTokenPrice(collateralAddress),
            };
        };
        this.toCollateral = (token, collateral) => {
            const { decimals, collateralDetails, price } = collateral;
            return {
                symbol: token.symbol,
                address: token.address,
                decimals,
                price,
                mintCR: collateralDetails.mintCR,
                liquidationFee: collateralDetails.liquidationFee,
                interestRate: collateralDetails.interestRate,
                chainId: token.chainId,
                name: token.name,
            };
        };
        this.indexedToCollateral = (collateral) => {
            return {
                symbol: collateral.symbol,
                address: collateral.address,
                decimals: collateral.decimals,
                mintCR: collateral.mintCollateralRatio,
                liquidationFee: collateral.liquidationFee,
                interestRate: collateral.interestRate,
                price: collateral.rate,
                name: collateral.name,
                chainId: collateral.chainId,
            };
        };
        this.findAvailable = (collateralSymbol) => {
            const avail = this.tokens.find((a) => a.symbol === collateralSymbol);
            if (!avail) {
                throw new Error(`Collaterals not initialised with collateral that matches: ${collateralSymbol}`);
            }
            return avail;
        };
        this.config = c || {
            protocolAddresses: sdkConfig.protocol.arbitrum.protocol,
            collaterals: sdkConfig.protocol.arbitrum.collaterals,
            chainId: NETWORK_NAME_TO_CHAIN_ID.arbitrum,
            graphEndpoint: sdkConfig.theGraphEndpoints.arbitrum.fx,
        };
        this.tokens = getTokensFromConfig(Object.values(this.config.collaterals).map(({ address }) => address));
        this.graph = new Graph(this.config.graphEndpoint);
    }
}
