import { ethers } from "ethers";
import sdkConfig from "../config";
import { NETWORK_NAME_TO_CHAIN_ID } from "../constants";
import { FxKeeperPool__factory } from "../contracts";
import { callMulticallObject, callMulticallObjects, createMulticallProtocolContracts, } from "../utils/contract";
import { fetchCacheApi } from "../utils/sdk";
export default class FxKeeperPool {
    constructor(c) {
        this.getPool = async (account, fxTokenSymbol, provider = sdkConfig.providers.arbitrum) => {
            const { provider: multicallProvider } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const multicall = this.getFxKeeperPoolMulticall(account, fxTokenSymbol, provider);
            const multicallResponse = await callMulticallObject(multicall, multicallProvider);
            return this.toFxKeeperPoolPool(fxTokenSymbol, multicallResponse);
        };
        this.getPools = async (account, provider = sdkConfig.providers.arbitrum) => {
            if (!account && sdkConfig.sdk.shouldUseCacheServer) {
                return fetchCacheApi("/earn-pools/fxkeeper");
            }
            const { provider: multicallProvider } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const fxTokenSymbols = Object.keys(this.config.fxTokenAddresses);
            const multicalls = fxTokenSymbols.map((fx) => this.getFxKeeperPoolMulticall(account, fx, provider));
            const multicallResponses = await callMulticallObjects(multicalls, multicallProvider);
            return multicallResponses.map((r, index) => this.toFxKeeperPoolPool(fxTokenSymbols[index], r));
        };
        this.stake = (args, signer, options = {}) => {
            const contract = this.getContract(signer);
            const fxTokenAddress = this.config.fxTokenAddresses[args.fxTokenSymbol];
            return contract.stake(args.amount, fxTokenAddress, ethers.constants.AddressZero, options);
        };
        this.unstake = (args, signer, options = {}) => {
            const contract = this.getContract(signer);
            const fxTokenAddress = this.config.fxTokenAddresses[args.fxTokenSymbol];
            return contract.unstake(args.amount, fxTokenAddress, options);
        };
        this.claim = (args, signer, options = {}) => {
            const contract = this.getContract(signer);
            const fxTokenAddress = this.config.fxTokenAddresses[args.fxTokenSymbol];
            return contract.withdrawCollateralReward(fxTokenAddress, options);
        };
        this.getFxKeeperPoolMulticall = (account, fxTokenSymbol, provider) => {
            const { contracts } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const base = {
                totalDeposited: contracts.fxKeeperPool.getPoolTotalDeposit(this.config.fxTokenAddresses[fxTokenSymbol]),
            };
            if (account) {
                return {
                    ...base,
                    accountBalance: contracts.fxKeeperPool.balanceOfStake(account, this.config.fxTokenAddresses[fxTokenSymbol]),
                    accountRewards: contracts.fxKeeperPool.balanceOfRewards(account, this.config.fxTokenAddresses[fxTokenSymbol]),
                };
            }
            return base;
        };
        this.toFxKeeperPoolPool = (fxTokenSymbol, multicallResponse) => {
            return {
                fxToken: fxTokenSymbol,
                totalDeposited: multicallResponse.totalDeposited,
                account: multicallResponse.accountBalance && multicallResponse.accountRewards
                    ? {
                        fxLocked: multicallResponse.accountBalance,
                        rewards: multicallResponse.accountRewards,
                    }
                    : undefined,
            };
        };
        this.getContract = (signer) => {
            return FxKeeperPool__factory.connect(this.config.protocolAddresses.fxKeeperPool, signer);
        };
        this.config = c || {
            protocolAddresses: sdkConfig.protocol.arbitrum.protocol,
            fxTokenAddresses: sdkConfig.fxTokenAddresses,
            chainId: NETWORK_NAME_TO_CHAIN_ID.arbitrum,
        };
    }
}
