import sdkConfig from "../config";
import { callMulticallObjects, createERC20MulticallContract, createMultiCallContract, } from "../utils/contract";
import { Staking__factory } from "../contracts";
import stakingContractAbi from "../abis/handle/Staking.json";
import { Provider as MultiCallProvider } from "ethers-multicall";
import { NETWORK_NAME_TO_CHAIN_ID } from "../constants";
import { fetchCacheApi } from "../utils/sdk";
export default class LpStaking {
    constructor(c) {
        this.getPools = async (account, signer) => {
            if (!account && sdkConfig.sdk.shouldUseCacheServer) {
                return fetchCacheApi("earn-pools/lpstaking");
            }
            const poolNames = Object.keys(this.config.pools);
            const multicalls = poolNames.map((poolName) => {
                return this.getMulticall(account, this.config.pools[poolName]);
            });
            const provider = new MultiCallProvider(signer.provider, this.config.chainId);
            const multicallResponses = await callMulticallObjects(multicalls, provider);
            const tokenBalances = await this.getTokenDepositedIntoLP(signer);
            return multicallResponses.map((poolData, index) => {
                const poolName = poolNames[index];
                const balances = tokenBalances[index];
                const pool = this.config.pools[poolName];
                return {
                    name: poolName,
                    title: pool.title,
                    address: pool.stakingContractAddress,
                    stakingContractAddress: pool.stakingContractAddress,
                    platform: pool.platform,
                    totalDeposited: poolData.totalDeposited,
                    distributionRate: poolData.distributionRate,
                    lpTokenTotalSupply: poolData.lpTokenTotalSupply,
                    distributionDuration: poolData.distributionDuration,
                    distributionPeriodEnds: poolData.distributionPeriodEnds,
                    rewardsBalance: poolData.rewardsBalance,
                    lpToken: pool.lpToken,
                    tokensInLp: pool.tokensInLp.map((token) => ({
                        symbol: token.symbol,
                        address: token.address,
                        decimals: token.decimals,
                        balance: balances[token.symbol],
                    })),
                    url: pool.url,
                    account: poolData.claimableRewards && poolData.deposited
                        ? {
                            claimableRewards: poolData.claimableRewards,
                            deposited: poolData.deposited,
                        }
                        : undefined,
                };
            });
        };
        this.stake = (args, signer, options = {}) => {
            const contract = this.getContract(args.poolName, signer);
            return contract.stake(args.amount, options);
        };
        this.unstake = (args, signer, options = {}) => {
            const contract = this.getContract(args.poolName, signer);
            return contract.withdraw(args.amount, options);
        };
        this.claim = (poolName, signer, options = {}) => {
            const contract = this.getContract(poolName, signer);
            return contract.getReward(options);
        };
        this.getMulticall = (account, poolDetails) => {
            const stakingContract = createMultiCallContract(poolDetails.stakingContractAddress, stakingContractAbi);
            const lpToken = createERC20MulticallContract(poolDetails.lpToken.address);
            const rewardToken = createERC20MulticallContract(this.config.forexAddress);
            const base = {
                totalDeposited: lpToken.balanceOf(poolDetails.stakingContractAddress),
                lpTokenTotalSupply: lpToken.totalSupply(),
                distributionRate: stakingContract.rewardRate(),
                distributionPeriodEnds: stakingContract.periodFinish(),
                distributionDuration: stakingContract.rewardsDuration(),
                rewardsBalance: rewardToken.balanceOf(poolDetails.stakingContractAddress),
            };
            if (account) {
                return {
                    ...base,
                    deposited: stakingContract.balanceOf(account),
                    claimableRewards: stakingContract.earned(account),
                };
            }
            return base;
        };
        this.getTokenDepositedIntoLP = async (signer) => {
            const poolNames = Object.keys(this.config.pools);
            const multicalls = poolNames.map((poolName) => {
                const pool = this.config.pools[poolName];
                return pool.tokensInLp.reduce((progress, token) => {
                    return {
                        ...progress,
                        [token.symbol]: createERC20MulticallContract(token.address).balanceOf(pool.lpToken.address),
                    };
                }, {});
            });
            const provider = new MultiCallProvider(signer.provider, this.config.chainId);
            return callMulticallObjects(multicalls, provider);
        };
        this.getContract = (poolName, signer) => {
            return Staking__factory.connect(this.config.pools[poolName].stakingContractAddress, signer);
        };
        this.config = c || {
            pools: sdkConfig.lpStaking.arbitrum,
            forexAddress: sdkConfig.forexAddress,
            chainId: NETWORK_NAME_TO_CHAIN_ID.arbitrum,
        };
    }
}
