import { ethers } from "ethers";
import Graph from "./graph";
import { Provider as MultiCallProvider } from "ethers-multicall";
import { createMulticallProtocolContracts, getCdpContracts, callMulticallObject, callMulticallObjects, } from "../utils/contract";
import sdkConfig from "../config";
import CollateralsSDK from "./Collaterals";
import FxTokensSDK from "./FxTokens";
import { getDeadline } from "../utils/general";
import { createSingleCollateralVault, createVault } from "../utils/vault";
import { getFxTokenByAddress, getFxTokenBySymbol } from "../utils/fxToken";
import { getCollateralByAddress, getCollateralBySymbol, } from "../utils/collateral";
import { ProtocolSDK } from "..";
import { getKashiPoolMulticall, kashiMulticallResultToSingleCollateralVaultData, } from "../utils/sushiswap";
import KashiCooker from "../utils/KashiCooker";
import { NETWORK_NAME_TO_CHAIN_ID } from "../constants";
export default class Vaults {
    constructor(c) {
        this.fxTokens = [];
        this.collaterals = [];
        this.initialised = false;
        this.initAsync = async (provider = sdkConfig.providers.arbitrum) => {
            const fxTokensPromise = this.fxTokensSDK.getFxTokens(provider);
            const collateralsPromise = this.collateralsSDK.getCollaterals(provider);
            const protocolParametersPromise = this.protocolSDK.getProtocolParameters(provider);
            const [fxTokens, collaterals, protocolParameters] = await Promise.all([
                fxTokensPromise,
                collateralsPromise,
                protocolParametersPromise,
            ]);
            this.fxTokens = fxTokens;
            this.collaterals = collaterals;
            this.protocolParameters = protocolParameters;
            this.initialised = true;
        };
        this.initSync = (protocolParamters, fxTokens, collaterals) => {
            this.protocolParameters = protocolParamters;
            this.fxTokens = fxTokens;
            this.collaterals = collaterals;
            this.initialised = true;
        };
        this.getVaults = async (account, provider = sdkConfig.providers.arbitrum, shouldAutoInitialise = true) => {
            if (!shouldAutoInitialise) {
                this.initialisationCheck();
            }
            else if (!this.initialised) {
                await this.initAsync(provider);
            }
            const { provider: multicallProvider } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, provider);
            const vaultMulticalls = this.fxTokens.map((t) => this.getVaultMulitcall(account, t.address, provider));
            const collateralBalancesMulticalls = this.fxTokens.map((fxToken) => this.createMulticallObjectForVaultCollateralBalance(account, fxToken.address, provider));
            const vaultsPromise = callMulticallObjects(vaultMulticalls, multicallProvider);
            const collateralsPromise = callMulticallObjects(collateralBalancesMulticalls, multicallProvider);
            const [vaultData, collaterals] = await Promise.all([
                vaultsPromise,
                collateralsPromise,
            ]);
            return vaultData.map((vault, index) => this.chainDataToVault(account, this.fxTokens[index].address, vault, collaterals[index]));
        };
        this.getVault = async (account, fxTokenSymbol, signer) => {
            this.initialisationCheck();
            const provider = new MultiCallProvider(signer.provider, this.config.chainId);
            const fxToken = this.fxTokens.find((t) => t.symbol === fxTokenSymbol);
            if (!fxToken) {
                throw new Error(`Could not find fxToken with symbol: ${fxTokenSymbol}`);
            }
            const vaultMulticall = this.getVaultMulitcall(account, fxToken.address, signer);
            const collateralBalanceMulticall = this.createMulticallObjectForVaultCollateralBalance(account, fxToken.address, signer);
            const vaultMulticallPromise = callMulticallObject(vaultMulticall, provider);
            const collateralMulticallResponsePromise = callMulticallObject(collateralBalanceMulticall, provider);
            const [vaultData, collateralData] = await Promise.all([
                vaultMulticallPromise,
                collateralMulticallResponsePromise,
            ]);
            return this.chainDataToVault(account, fxToken.address, vaultData, collateralData);
        };
        this.getSingleCollateralVault = async (account, vaultSymbol, network, signer) => {
            this.initialisationCheck();
            const chainId = NETWORK_NAME_TO_CHAIN_ID[network];
            const provider = new MultiCallProvider(signer.provider, chainId);
            const pool = this.config.singleCollateralVaults[network][vaultSymbol];
            if (!pool) {
                throw new Error(`Unable to find vault: ${network}-${vaultSymbol}`);
            }
            const fxToken = getFxTokenBySymbol(this.fxTokens, pool.fxToken);
            const multicall = getKashiPoolMulticall(account, pool, chainId);
            const result = await callMulticallObject(multicall, provider);
            return createSingleCollateralVault(kashiMulticallResultToSingleCollateralVaultData(account, pool, result), fxToken);
        };
        this.getIndexedVault = async (account, fxTokenSymbol) => {
            this.initialisationCheck();
            const fxToken = getFxTokenBySymbol(this.fxTokens, fxTokenSymbol);
            const indexedVault = await this.graph.vaults.queryOne({
                where: {
                    account: account.toLowerCase(),
                    fxToken: fxToken.address.toLowerCase(),
                },
            });
            return this.indexedDataToVault(indexedVault || this.createEmptyIndexedVault(account, fxToken.address));
        };
        this.getIndexedVaults = async (account) => {
            this.initialisationCheck();
            const indexedVaults = await this.graph.vaults.query({
                where: { account: account.toLowerCase() },
            });
            const allVaults = this.fxTokens.map((fxToken) => {
                const existingVault = indexedVaults.find((vault) => vault.fxToken.toLowerCase() === fxToken.address.toLowerCase());
                return (existingVault || this.createEmptyIndexedVault(account, fxToken.address));
            });
            return allVaults.map(this.indexedDataToVault);
        };
        this.getSingleCollateralVaults = async (account, network, signer) => {
            this.initialisationCheck();
            const chainId = NETWORK_NAME_TO_CHAIN_ID[network];
            const provider = new MultiCallProvider(signer.provider, chainId);
            const poolConfigs = this.config.singleCollateralVaults[network];
            const vaultsSymbols = Object.keys(poolConfigs);
            const multicalls = vaultsSymbols.map((symbol) => getKashiPoolMulticall(account, poolConfigs[symbol], chainId));
            const results = await callMulticallObjects(multicalls, provider);
            return results.map((r, index) => {
                const poolConfig = poolConfigs[vaultsSymbols[index]];
                const fxToken = getFxTokenBySymbol(this.fxTokens, poolConfig.fxToken);
                return createSingleCollateralVault(kashiMulticallResultToSingleCollateralVaultData(account, poolConfig, r), fxToken);
            });
        };
        this.mint = async (args, signer, options = {}) => {
            this.initialisationCheck();
            const protocolContracts = getCdpContracts(this.config.protocolAddresses, signer);
            const contract = protocolContracts.comptroller;
            const fxToken = getFxTokenBySymbol(this.fxTokens, args.fxToken);
            const deadline = getDeadline(args.deadline);
            const referral = args.referral ?? ethers.constants.AddressZero;
            if (args.collateral) {
                if (args.collateral.symbol === "ETH") {
                    return contract.mintWithEth(args.amount, fxToken.address, deadline, referral, {
                        ...options,
                        value: args.collateral.amount,
                    });
                }
                const collateral = getCollateralBySymbol(this.collaterals, args.collateral.symbol);
                return contract.mint(args.amount, fxToken.address, collateral.address, args.collateral.amount, deadline, referral, options);
            }
            return contract.mintWithoutCollateral(args.amount, fxToken.address, deadline, referral, options);
        };
        this.mintAndDepositSingleCollateral = async (args, signer, options = {}) => {
            if (!args.mintAmount && !args.depositAmount) {
                throw new Error("Must supply at least one of mintAmount or depositAmount");
            }
            if (args.mintAmount?.isZero() && args.depositAmount?.isZero()) {
                throw new Error("One of mintAmount or depositAmount must be non-zero");
            }
            const kashiPool = this.config.singleCollateralVaults[args.network][args.vaultSymbol];
            if (!kashiPool) {
                throw new Error(`Unable to find vault: ${args.network}-${args.vaultSymbol}`);
            }
            const account = await signer.getAddress();
            const chainId = NETWORK_NAME_TO_CHAIN_ID[args.network];
            const fxToken = getFxTokenBySymbol(this.fxTokens, kashiPool.fxToken);
            const cooker = new KashiCooker(kashiPool, account, fxToken, chainId);
            if (args.approveKashiSignature) {
                cooker.approve(args.approveKashiSignature);
            }
            if (args.depositAmount?.gt(0)) {
                cooker.addCollateral(args.depositAmount);
            }
            if (args.mintAmount?.gt(0)) {
                cooker.borrow(args.mintAmount);
            }
            return cooker.cook(signer, options);
        };
        this.burnAndWithdrawSingleCollateral = async (args, signer, options = {}) => {
            if (!args.burnAmount && !args.withdrawAmount) {
                throw new Error("Must supply at least one of mintAmount or depositAmount");
            }
            if (args.burnAmount?.isZero() && args.withdrawAmount?.isZero()) {
                throw new Error("One of mintAmount or depositAmount must be non-zero");
            }
            const kashiPool = this.config.singleCollateralVaults[args.network][args.vaultSymbol];
            if (!kashiPool) {
                throw new Error(`Unable to find vault ${args.network}-${args.vaultSymbol}`);
            }
            const account = await signer.getAddress();
            const chainId = NETWORK_NAME_TO_CHAIN_ID[args.network];
            const fxToken = getFxTokenBySymbol(this.fxTokens, kashiPool.fxToken);
            const cooker = new KashiCooker(kashiPool, account, fxToken, chainId);
            if (args.burnAmount?.gt(0)) {
                cooker.repay(args.burnAmount);
            }
            if (args.withdrawAmount?.gt(0)) {
                cooker.removeCollateral(args.withdrawAmount);
            }
            return cooker.cook(signer, options);
        };
        this.supplyFxTokenSingleCollateral = async (args, signer, options = {}) => {
            const kashiPool = this.config.singleCollateralVaults[args.network][args.vaultSymbol];
            if (!kashiPool) {
                throw new Error(`Unable to find vault ${args.network}-${args.vaultSymbol}`);
            }
            const account = await signer.getAddress();
            const chainId = NETWORK_NAME_TO_CHAIN_ID[args.network];
            const fxToken = getFxTokenBySymbol(this.fxTokens, kashiPool.fxToken);
            const cooker = new KashiCooker(kashiPool, account, fxToken, chainId);
            if (args.approveKashiSignature) {
                cooker.approve(args.approveKashiSignature);
            }
            cooker.addAsset(args.amount);
            return cooker.cook(signer, options);
        };
        this.depositCollateral = (args, signer, options = {}) => {
            this.initialisationCheck();
            const protocolContracts = getCdpContracts(this.config.protocolAddresses, signer);
            const fxToken = getFxTokenBySymbol(this.fxTokens, args.fxToken);
            const contract = protocolContracts.treasury;
            const referral = args.referral ?? ethers.constants.AddressZero;
            if (args.collateral === "ETH") {
                return contract.depositCollateralETH(args.account, fxToken.address, referral, {
                    ...options,
                    value: args.amount,
                });
            }
            const collateral = getCollateralBySymbol(this.collaterals, args.collateral);
            return contract.depositCollateral(args.account, args.amount, collateral.address, fxToken.address, referral, options);
        };
        this.burn = (args, signer, options = {}) => {
            this.initialisationCheck();
            const protocolContracts = getCdpContracts(this.config.protocolAddresses, signer);
            const fxToken = getFxTokenBySymbol(this.fxTokens, args.fxToken);
            return protocolContracts.comptroller.burn(args.amount, fxToken.address, getDeadline(args.deadline), options);
        };
        this.withdrawCollateral = (args, signer, options = {}) => {
            this.initialisationCheck();
            this.initialisationCheck();
            const protocolContracts = getCdpContracts(this.config.protocolAddresses, signer);
            const fxToken = getFxTokenBySymbol(this.fxTokens, args.fxToken);
            const contract = protocolContracts.treasury;
            if (args.collateral === "ETH") {
                return contract.withdrawCollateralETH(args.account, args.amount, fxToken.address, options);
            }
            const collateral = getCollateralBySymbol(this.collaterals, args.collateral);
            return contract.withdrawCollateral(collateral.address, args.account, args.amount, fxToken.address, options);
        };
        this.getVaultMulitcall = (account, fxTokenAddress, signerOrProvider) => {
            const { contracts } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, signerOrProvider);
            return {
                debt: contracts.handle.getDebt(account, fxTokenAddress),
            };
        };
        this.createMulticallObjectForVaultCollateralBalance = (account, fxTokenAddress, signerOrProvider) => {
            const { contracts } = createMulticallProtocolContracts(this.config.protocolAddresses, this.config.chainId, signerOrProvider);
            return this.collaterals.reduce((progress, collateral) => {
                return {
                    ...progress,
                    [collateral.symbol]: contracts.handle.getCollateralBalance(account, collateral.address, fxTokenAddress),
                };
            }, {});
        };
        this.chainDataToVault = (account, fxTokenAddress, vault, collateralMap) => {
            const { debt } = vault;
            const collateral = this.collaterals.map((c) => ({
                ...c,
                amount: collateralMap[c.symbol],
            }));
            const fxToken = getFxTokenByAddress(this.fxTokens, fxTokenAddress);
            const vaultData = {
                account: account.toLowerCase(),
                fxToken: fxToken,
                debt,
                collateral,
            };
            return createVault(vaultData, this.protocolParameters, fxToken, this.collaterals);
        };
        this.indexedDataToVault = (vault) => {
            const fxToken = getFxTokenByAddress(this.fxTokens, vault.fxToken);
            const collateral = vault.collateralTokens.map((c) => {
                const collateral = getCollateralByAddress(this.collaterals, c.address);
                return {
                    address: c.address.toLowerCase(),
                    symbol: collateral.symbol,
                    amount: ethers.BigNumber.from(c.amount),
                    decimals: collateral.decimals,
                    name: collateral.name,
                    chainId: collateral.chainId,
                };
            });
            const vaultData = {
                account: vault.account.toLowerCase(),
                fxToken: fxToken,
                debt: ethers.BigNumber.from(vault.debt),
                collateral,
            };
            return createVault(vaultData, this.protocolParameters, fxToken, this.collaterals);
        };
        this.initialisationCheck = () => {
            if (!this.initialised) {
                throw new Error("Vaults SDK not initialised");
            }
        };
        this.createEmptyIndexedVault = (account, fxTokenAddress) => {
            return {
                account: account,
                fxToken: fxTokenAddress,
                debt: "0",
                redeemableTokens: "0",
                collateralAsEther: "0",
                collateralRatio: "0",
                minimumRatio: "0",
                isRedeemable: false,
                isLiquidatable: false,
                collateralTokens: this.collaterals.map((c) => ({
                    address: c.address,
                    amount: "0",
                })),
            };
        };
        this.config = c || {
            forexTokenAddress: sdkConfig.forexAddress,
            protocolAddresses: sdkConfig.protocol.arbitrum.protocol,
            fxTokenAddresses: sdkConfig.fxTokenAddresses,
            collaterals: sdkConfig.protocol.arbitrum.collaterals,
            chainId: NETWORK_NAME_TO_CHAIN_ID.arbitrum,
            graphEndpoint: sdkConfig.theGraphEndpoints.arbitrum.fx,
            singleCollateralVaults: sdkConfig.singleCollateralVaults,
        };
        this.fxTokensSDK = new FxTokensSDK(c);
        this.collateralsSDK = new CollateralsSDK(c);
        this.protocolSDK = new ProtocolSDK(c);
        this.graph = new Graph(this.config.graphEndpoint);
    }
}
