import websocket from "websocket";
import { DATA_FEED_API_WS_URL_QUOTES } from "../../config";
import { allSettledResults, pairToString } from "../../utils/general";
import { BigNumber } from "ethers";
import { fetchApiQuote } from "./fetcher";
import { toParsedDatafeedPair } from "./toParsedDatafeedPair";
import { config } from "../../index";
import { setupPingInterval } from "../../utils/ws";
const RECONNECT_DELAY = 1000;
const MAX_RECONNECT_ATTEMPTS = 10;
/**
 * Watches for price updates on the h2so server at oracle.handle.fi,
 * allowing a client to subscribe to price updates.
 */
class PriceFeed {
    constructor(options) {
        this.subscriptions = [];
        this.feed = {};
        this.connectionId = 0;
        this.isConnecting = false;
        this.options = options;
        this.callback = options.callback;
        if (options.connect) {
            this.connectPersisting(options?.connectMaxAttempts, options?.reconnectDelayMillis).catch(console.error);
        }
    }
    async connect(reconnectDelayMillis = RECONNECT_DELAY) {
        if (this.isConnecting) {
            await this.waitForConnection();
            return;
        }
        if (this.isConnected) {
            try {
                this.close();
            }
            catch (_) { }
        }
        this.isConnecting = true;
        try {
            await this.setupConnection(reconnectDelayMillis);
        }
        finally {
            this.isConnecting = false;
        }
    }
    async setupConnection(reconnectDelayMillis = RECONNECT_DELAY) {
        this.client = new websocket.w3cwebsocket(this.options?.websocketUrl ?? DATA_FEED_API_WS_URL_QUOTES);
        if (this.options.autoReconnect) {
            this.client.onclose = (e) => {
                console.debug("[handle-sdk] [feed] onclose (w/autoReconnect)", e);
                setTimeout(() => this.connectPersisting(MAX_RECONNECT_ATTEMPTS, reconnectDelayMillis), reconnectDelayMillis);
            };
        }
        else {
            this.client.onclose = (e) => {
                console.debug("[handle-sdk] [feed] onclose", e);
            };
        }
        // Set the error handler.
        this.client.onerror = (error) => {
            console.error("[handle-sdk] [feed] onerror", error);
            // Call both the internal & user callbacks, if defined.
            this.internalErrorCallback?.(error);
            this.userErrorCallback?.(error);
        };
        this.client.onmessage = this.onMessage.bind(this);
        this.clientOpenPromise = new Promise((resolve, reject) => {
            this.client.onopen = () => {
                console.debug("[handle-sdk] [feed] onopen");
                setupPingInterval(this.client);
                resolve();
            };
            this.internalErrorCallback = reject;
        });
        await this.waitForConnection();
        this.connectionId += 1;
    }
    async connectPersisting(maxAttempts = MAX_RECONNECT_ATTEMPTS, reconnectDelayMillis = RECONNECT_DELAY) {
        let lastError;
        for (let attempt = 1; attempt < maxAttempts + 1; attempt++) {
            try {
                await this.connect(reconnectDelayMillis);
                return;
            }
            catch (error) {
                lastError = error;
                console.error(error);
                if (config.sdk.printLogs) {
                    console.debug(`[handle-sdk] [feed] re-trying connection (attempt ${attempt})`);
                }
                await new Promise((resolve) => setTimeout(resolve, reconnectDelayMillis * attempt));
            }
        }
        throw lastError;
    }
    close() {
        if (!this.isConnected || !this.client) {
            return;
        }
        this.client.onclose = () => null;
        this.client.close();
    }
    onError(callback) {
        if (!this.client)
            return;
        this.userErrorCallback = callback;
    }
    onMessage(message) {
        if (typeof message.data !== "string")
            return;
        const data = JSON.parse(message.data);
        const value = BigNumber.from(String(data.value));
        // Set instance state.
        this.feed[pairToString(data.pair)] = {
            value,
            timestamp: data.timestamp,
        };
        // Trigger callback.
        this.callback({
            pair: data.pair,
            ...this.feed[pairToString(data.pair)],
        });
    }
    async subscribe(pairs) {
        if (!this.isConnected) {
            throw new Error("PriceFeed: websocket is not connected");
        }
        this.removeOldSubscriptions();
        // Filter out pairs that are already subscribed.
        pairs = pairs.filter((pair) => !this.subscriptions.find((subscription) => pairToString(subscription.pair) == pairToString(pair) &&
            subscription.connectionId == this.connectionId));
        // Return early if there are no new subscriptions.
        if (pairs.length === 0) {
            // Wait for subscription promise to finish to ensure that
            // feed values are fetched.
            await this.subscriptionInitialPromise;
            return;
        }
        this.client.send(JSON.stringify({
            action: "subscribe",
            params: {
                pairs: pairs.map(pairToString).join(","),
            },
        }));
        const subscriptions = pairs.map((pair) => ({
            pair,
            connectionId: this.connectionId,
        }));
        // Push subscriptions to state.
        this.subscriptions.push(...subscriptions);
        // Initially fetch all prices instead of waiting for next websocket update.
        const requests = pairs.map(async (pair) => {
            const data = await fetchApiQuote(pair, false);
            const feed = {
                value: BigNumber.from(data.data.result),
                timestamp: Math.floor(Date.now() / 1000),
            };
            this.feed[pairToString(pair)] = feed;
            if (!this.callback)
                return;
            this.callback({
                pair,
                ...feed,
            });
        });
        this.subscriptionInitialPromise = this.subscriptionInitialPromise
            ? Promise.all([this.subscriptionInitialPromise, ...requests])
            : Promise.all(requests);
        await this.subscriptionInitialPromise;
        console.debug("[handle-sdk] [feed] subscribed to pairs", pairs);
    }
    unsubscribe(pairs) {
        if (!this.isConnected) {
            throw new Error("PriceFeed: websocket is not connected");
        }
        // Assert that pairs are valid.
        this.client.send(JSON.stringify({
            action: "unsubscribe",
            params: {
                pairs: pairs.map(pairToString).join(","),
            },
        }));
        // Delete subscriptions from state.
        this.subscriptions = this.subscriptions.filter((subscription) => !pairs.some((pair) => pairToString(pair) === pairToString(subscription.pair)));
    }
    async waitForConnection() {
        await this.clientOpenPromise;
    }
    get isConnected() {
        return (this.client && this.client.readyState === websocket.w3cwebsocket.OPEN);
    }
    async getPrice(pair) {
        const pairString = pairToString(pair);
        // Ensure pair subscription price availability.
        await this.subscribe([pair]);
        if (!this.feed[pairString]) {
            throw new Error(`PriceFeed: no price for ${pairString}`);
        }
        return this.feed[pairString].value;
    }
    /**
     * Gets a (nullable) price from the feed, and does not subscribe to the pair
     * if missing.
     */
    getPriceSync(pair) {
        return this.feed[pairToString(pair)]?.value;
    }
    removeOldSubscriptions() {
        const indicesToRemove = this.subscriptions
            .map((s, i) => ({ subscription: s, i }))
            .filter((v) => v.subscription.connectionId !== this.connectionId)
            .map((v) => v.i);
        for (let i = this.subscriptions.length; i >= 0; i--) {
            if (!indicesToRemove.includes(i)) {
                continue;
            }
            this.subscriptions.splice(i, 1);
        }
    }
}
let feed;
const subscriptions = [];
// TODO: remove the need for `forceSocketReconnection` this to be used by
// the client, due to the background process having its WebSocket connection
// killed by the OS. The problem is that when the app comes to the foreground,
// the connection has been killed but it appears to be live still with a
// delay.
export const fetchPrices = async (pairs, forceSocketReconnection = false) => {
    pairs = pairs.map(toParsedDatafeedPair);
    const shouldConstructFeed = !feed;
    if (shouldConstructFeed) {
        console.debug("[handle-sdk] [feed] connecting socket");
        feed = new PriceFeed({
            autoReconnect: true,
            connect: false,
            callback: publishPriceUpdate,
        });
        await feed.connectPersisting();
    }
    if (!feed) {
        // This should never happen; it is mostly to make the compiler happy.
        throw new Error("[handle-sdk] [feed] not constructed");
    }
    if (!shouldConstructFeed && (!feed.isConnected || forceSocketReconnection)) {
        console.debug("[handle-sdk] [feed] re-connecting socket");
        await feed.connectPersisting();
    }
    // This is required as there are many-to-one subscriptions to the feed,
    // and subsequent ones will not call `connectPersisting`, so they must
    // use this promise to ensure the connection is live.
    await feed.waitForConnection();
    const prices = await allSettledResults(pairs.map(feed.getPrice.bind(feed)));
    return prices.reduce((object, price, i) => ({
        ...object,
        [pairToString(pairs[i])]: price,
    }), {});
};
export const fetchPrice = async (pair) => {
    const prices = await fetchPrices([pair]);
    const price = prices[pairToString(pair)];
    if (!price) {
        throw new Error(`getPrice: no price for ${pairToString(pair)}`);
    }
    return price;
};
/// Subscribes to h2so feed and returns existing prices for pairs requested.
export const subscribe = (pairs, callback, forceSocketReconnection = false) => ({
    subscriberId: subscriptions.push({
        callback,
        arg: pairs,
    }) - 1,
    prices: fetchPrices(pairs, forceSocketReconnection),
});
export const unsubscribe = (id) => {
    if (!subscriptions[id])
        throw new Error("h2so feed: not subscribed");
    subscriptions[id] = null;
};
/// Gets the latest socket price; throws if not available.
export const getLatestPrice = (pair) => {
    if (!feed)
        throw new Error("h2so: no feed");
    const price = feed.getPriceSync(pair);
    if (!price)
        throw new Error(`h2so: no price for ${pairToString(pair)}`);
    return price;
};
const publishPriceUpdate = (data) => {
    const pair = pairToString(data.pair);
    for (let subscription of subscriptions) {
        if (!subscription)
            continue;
        for (let subscriptionPair of subscription.arg) {
            if (pairToString(subscriptionPair) !== pair)
                continue;
            subscription.callback(data.pair, data.value);
        }
    }
};
