import { BigNumber, ethers } from "ethers";
import { config } from "../..";
import axios from "axios";
import { DATA_FEED_API_BASE_URL, DATA_FEED_SIGNING_ADDRESS, } from "../../config";
import { toParsedDatafeedPair } from "./toParsedDatafeedPair";
import { feed } from "./index";
import { pairToString } from "../../utils/general";
export const fetchEncodedSignedQuotes = async (pairs) => {
    const signedQuotes = await fetchSignedQuotes(pairs);
    return {
        encoded: encodeQuotes(signedQuotes),
        decoded: signedQuotes,
    };
};
export const fetchUnsignedQuote = async (pair) => {
    try {
        // Try fetching the price from the H2SO feed first.
        pair = toParsedDatafeedPair(pair);
        const prices = await feed.fetchPrices([pair]);
        const price = prices?.[pairToString(pair)]?.toNumber();
        if (!price)
            throw new Error(`fetchUnsignedQuote: no price in h2so feed for ${pairToString(pair)}`);
        return price;
    }
    catch (error) {
        // In case of failure, fetch from the API server directly.
        console.error(error);
        return (await fetchApiQuote(pair, false)).data.result;
    }
};
export const fetchApiQuote = async (pair, sign = true) => {
    pair = toParsedDatafeedPair(pair);
    const result = await axios.get(`${DATA_FEED_API_BASE_URL}/${pair.baseSymbol}/${pair.quoteSymbol}?sign=${sign ? "true" : "false"}`);
    if (sign) {
        return result.data;
    }
    return result.data;
};
const fetchSignedQuotes = async (pairs) => {
    const responses = await Promise.all(pairs.map(async (pair) => {
        const response = await fetchApiQuote(pair, true);
        if (response.data.signed) {
            return response;
        }
        else {
            throw new Error(`No signature returned for ${pair}`);
        }
    }));
    return responses.map((response, i) => quoteApiResponseToSignedQuote(pairs[i], response));
};
const quoteApiResponseToSignedQuote = (pair, { data: { result, signed: { signatureParams, signature, message }, }, }, verifySigner = true) => {
    if (verifySigner) {
        const untrustedSigner = ethers.utils
            .verifyMessage(ethers.utils.arrayify(`0x${message}`), `0x${signature}`)
            .toLowerCase();
        if (untrustedSigner !== DATA_FEED_SIGNING_ADDRESS.toLowerCase())
            throw new Error(`Message is not signed by authorised signer (signed by "${untrustedSigner}")`);
    }
    return {
        pair,
        signature: signature.startsWith("0x")
            ? ethers.utils.arrayify(signature)
            : ethers.utils.arrayify(`0x${signature}`),
        signatureParams: {
            value: BigNumber.from(result),
            signedTimestamp: BigNumber.from(signatureParams.signedTimestamp),
            chainId: signatureParams.chainId,
            validFromTimestamp: BigNumber.from(signatureParams.validFromTimestamp),
            durationSeconds: BigNumber.from(signatureParams.durationSeconds),
        },
    };
};
export const encodeQuotes = (quotes) => {
    const concatenatedSignatures = quotes.reduce((buffer, quote, i) => {
        for (let j = 0; j < 65; j++) {
            const offset = i * 65;
            buffer[offset + j] = quote.signature[j];
        }
        return buffer;
    }, new Uint8Array(quotes.length * 65));
    const tokenAddresses = quotes.map((quote) => symbolToAddress(quote.pair.baseSymbol));
    return ethers.utils.defaultAbiCoder.encode([
        "uint256",
        "address[]",
        "uint256[]",
        "uint256[]",
        "uint256[]",
        "uint256[]",
        "bytes",
    ], [
        tokenAddresses.length,
        tokenAddresses,
        quotes.map((quote) => quote.signatureParams.value),
        quotes.map((quote) => quote.signatureParams.signedTimestamp),
        quotes.map((quote) => quote.signatureParams.validFromTimestamp),
        quotes.map((quote) => quote.signatureParams.durationSeconds),
        concatenatedSignatures,
    ]);
};
const symbolToAddress = (symbol) => {
    const eth = (symbol) => {
        if (symbol !== "ETH" && symbol !== "WETH")
            return;
        return config.protocol.arbitrum.collaterals.WETH.address;
    };
    const fxToken = (symbol) => {
        const fxSymbol = symbol.startsWith("fx") ? symbol : `fx${symbol}`;
        return config.fxTokenAddresses[fxSymbol];
    };
    // Find a valid address conversion.
    const result = [eth, fxToken]
        .map((converter) => converter(symbol))
        .find((address) => address != null);
    if (!result)
        throw new Error(`Couldn't get address for symbol "${symbol}"`);
    return result;
};
/// Returns the encoded signature after waiting (if needed) for
/// it to become valid.
export const fetchTimedEncodedSignedQuotes = async (pairs, provider) => {
    const signature = await fetchEncodedSignedQuotes(pairs);
    const maxValidTime = Math.max(...signature.decoded.map((quote) => +quote.signatureParams.validFromTimestamp));
    await waitForBlockWithTimestamp(maxValidTime, provider);
    return signature;
};
const waitForBlockWithTimestamp = async (timestamp, provider) => new Promise(async (resolve) => {
    const checkBlock = async () => {
        const block = await provider.getBlock("latest");
        if (block?.timestamp && block.timestamp >= timestamp) {
            provider.removeListener("block", checkBlock);
            resolve();
        }
    };
    await checkBlock();
    provider.addListener("block", checkBlock);
});
