import config from "../../config";
import { ChainlinkAggregator__factory } from "../../contracts";
import { H2SO_PRICE_DECIMALS, NETWORK_NAME_TO_CHAIN_ID } from "../../constants";
import { isFxTokenSymbol, stripFx } from "../../utils/fxToken";
import * as coingecko from "../prices/sources/coingecko";
import { formatUnits, parseUnits } from "ethers/lib/utils";
import { h2so } from "../../index";
import { getUsdQuotedPair, replaceWrappedSymbolForNative, transformDecimals, } from "../../utils/general";
import { hlp } from "../trade/platforms";
const CONFIG = {
    feeds: config.protocol.arbitrum.chainlinkFeeds,
    chainId: NETWORK_NAME_TO_CHAIN_ID.arbitrum,
};
const fetchEthUsdPrice = async (signer) => {
    const aggregator = ChainlinkAggregator__factory.connect(CONFIG.feeds.ETH, signer ?? config.providers.arbitrum);
    const bn = await aggregator.latestAnswer();
    const number = chainlinkPriceToNumber(bn);
    return {
        bn,
        number,
    };
};
/**
 * Returns the prices for all fxTokens in USD, using the
 * official Chainlink feed.
 * These prices are used for the CDP vaults.
 */
const fetchFxTokenTargetChainlinkUsdPrices = async () => {
    const fxTokenSymbols = Object.keys(CONFIG.feeds).filter((key) => isFxTokenSymbol(key));
    const calls = fxTokenSymbols
        // Remove fxUSD as it is a constant (one).
        .filter((symbol) => symbol !== "fxUSD")
        .map(fetchFxTokenTargetChainlinkUsPrice);
    const response = await Promise.all(calls);
    const result = response.reduce((progress, bn, index) => {
        return {
            ...progress,
            [fxTokenSymbols[index]]: {
                bn,
                number: chainlinkPriceToNumber(bn),
            },
        };
    }, {});
    return {
        ...result,
        fxUSD: {
            bn: parseUnits("1", H2SO_PRICE_DECIMALS),
            number: 1,
        },
    };
};
const fetchFxTokenTargetChainlinkUsPrice = (symbol) => ChainlinkAggregator__factory.connect(CONFIG.feeds[symbol], config.providers.arbitrum).latestAnswer();
const chainlinkPriceToNumber = (price) => +formatUnits(price, 8);
/// Tries to fetch a price, from any source available, for a token symbol.
/// Allows passing an optional TokenInfo, which makes it more likely the
/// price will be found, specially if the token is a non-handle token.
export const fetchTokenPriceUsd = async (symbol, token) => {
    const pair = getUsdQuotedPair(replaceWrappedSymbolForNative(stripFx(symbol)));
    // If the pair is USD/USD, return one.
    if (pair.baseSymbol === pair.quoteSymbol) {
        return parseUnits("1", H2SO_PRICE_DECIMALS);
    }
    // First, try to fetch from the pre-cached H2SO feed.
    try {
        return h2so.feed.getLatestPrice(pair);
    }
    catch (_) { }
    if (isFxTokenSymbol(symbol)) {
        // The fxToken was not present in H2SO. Use the Chainlink price.
        try {
            return await fetchFxTokenTargetChainlinkUsPrice(symbol);
        }
        catch (_) { }
    }
    // If ETH, fetch from the official Chainlink oracle on Arbitrum One,
    // if it could not be fetched from H2SO.
    if (symbol.toUpperCase() === "ETH" || symbol.toUpperCase() === "WETH") {
        try {
            return await fetchEthUsdPrice().then((response) => response.bn);
        }
        catch (_) { }
    }
    // If the hLP token, fetch via the hlp module.
    if (isHlpToken(symbol, token)) {
        try {
            return await hlp.internals
                .getHlpPrice()
                .then((price) => transformDecimals(price.maximum, 30, H2SO_PRICE_DECIMALS));
        }
        catch (_) { }
    }
    return null;
};
export const isHlpToken = (symbol, token) => {
    const isSymbolCorrect = symbol.toLowerCase().endsWith("hlp");
    if (!isSymbolCorrect) {
        return false;
    }
    if (!token) {
        // Only consider the symbol check, as the token is not available.
        return true;
    }
    return !!token.extensions?.isLiquidityToken;
};
export { coingecko, fetchEthUsdPrice, fetchFxTokenTargetChainlinkUsdPrices };
