import { Referrals__factory } from "../../../contracts";
import config from "../../../config";
import { referralsNetworks } from "../../../types/network";
import { getEventData, validateNetworkType } from "../../../utils/web3";
import { formatBytes32String } from "ethers/lib/utils";
export const openReferrerAccount = async (recipientAddress, signer) => {
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.openReferrerAccount(recipientAddress);
    return fetchOpenAccountTxResponse(tx, referrals);
};
export const openChildReferrerAccount = async (recipientAddress, parentReferrerId, weight, signer) => {
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.openChildReferrerAccount(recipientAddress, parentReferrerId, weight);
    return fetchOpenAccountTxResponse(tx, referrals);
};
export const closeReferrerAccount = async (referrerId, signer) => {
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.closeReferrerAccount(referrerId);
    return tx.wait(1);
};
const fetchOpenAccountTxResponse = async (tx, referrals) => {
    const receipt = await tx.wait(1);
    const eventData = getEventData("ReferrerAccountOpened", referrals, receipt);
    if (!eventData) {
        throw new Error("no event data");
    }
    return {
        receipt,
        referrerId: eventData.args.referrerId.toNumber(),
    };
};
export const setChildWeights = async (parentReferrerId, args, signer) => {
    const childReferrerIds = args.map((a) => a.referrerId);
    const weights = args.map((a) => a.weight);
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.setChildWeights(parentReferrerId, childReferrerIds, weights);
    return tx.wait(1);
};
export const configureReferralCodes = async (referrerId, args, signer) => {
    const codes = args.map((a) => a.code);
    const rebates = args.map((a) => a.userRebate);
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.configureReferralCodes(codes.map(codeStringToBytes32), rebates, referrerId);
    return tx.wait(1);
};
export const deleteReferralCodes = async (referrerId, codes, signer) => {
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.deleteReferralCodes(codes.map(codeStringToBytes32), referrerId);
    return tx.wait(1);
};
export const obtainChildReferrerOwnership = async (childReferrerId, parentReferrerId, signer) => {
    const network = await validateNetworkType(signer, referralsNetworks);
    const referrals = getReferralsContract(network, signer);
    const tx = await referrals.obtainChildReferrerOwnership(childReferrerId, parentReferrerId);
    return tx.wait(1);
};
const getReferralsContract = (network, signer) => Referrals__factory.connect(config.protocol[network].referrals, signer ?? config.providers[network]);
const codeStringToBytes32 = (code) => {
    if (code.length > 32) {
        throw new Error("code cannot be more than 32 characters");
    }
    return formatBytes32String(code);
};
