import { TokenInfo } from "@uniswap/token-lists";
import TokenManager from ".";
import { Network } from "../..";
/**
 * TokenManager that comes with native, wrapped native, and handle tokens by default.
 * Handle supported token extensions:
 * - isNative: true if token is native for that network (e.g. ETH, MATIC)
 * - isHlpToken: true if the token is a hlp token in a handle Vault contract
 * - isWrappedNative: true if the token is a wrapped version of a native token (e.g. WETH)
 * - isStable: true if the token is a USD stablecoin, false otherwise,
 * - isShortable: true if the token is shortable, false otherwise
 * - isFxToken: true if token is a fx token, false otherwise
 * - isLiquidityToken: true if token is the Handle Liquidity Pool token (symbol hLP), false otherwise
 * @note the Handle Liquidity Pool token (symbol hLP) has isHlpToken set to false,
 * as it is not technically in the liquidity pool. Instead, it has isLiquidityToken set to true
 */
declare class HandleTokenManager extends TokenManager {
    constructor(config?: import(".").TokenManagerConfig);
    /**
     * Creates an instance of HandleTokenList from a TokenList instance.
     * @param tokenManager the tokenList from which to construct the HandleTokenList
     * @returns an instance of HandleTokenList with the cache of the tokenList
     */
    static from(tokenManager: TokenManager): HandleTokenManager;
    /**
     * Gets all hLP tokens for a network
     * @param network the network on which to search for tokens
     * @returns all hLP tokens for the network
     */
    getHlpTokens(network: number | Network): TokenInfo[];
    /**
     * Checks if a token is a supported hLP token by its symbol
     * @param symbol the symbol of the token to check
     * @param network the network on which to check the token
     * @returns wheteher the token is a hlpToken
     */
    isHlpTokenBySymbol(symbol: string, network: Network | number): boolean;
    /**
     * Checks if a token is a supported hLP token by its address
     * @param address the address of the token to check
     * @param network the network on which to check the token
     * @returns wheteher the token is a hlpToken
     */
    isHlpTokenByAddress(address: string, network: Network | number): boolean;
    /**
     * Checks if the token is a stable hLP token
     * @param symbol the symbol of the token to check
     * @param network the network on which to check the token
     * @returns whether there exists a stable hLP token with the given symbol
     */
    isHlpStableTokenBySymbol(symbol: string, network: Network | number): boolean;
    /**
     * Checks if the token is a stable hLP token
     * @param address the address of the token to check
     * @param network the network on which to check the token
     * @returns whether there exists a stable hLP token with the given address
     */
    isHlpStableTokenByAddress(address: string, network: Network | number): boolean;
    /**
     * Finds the hLP compatible wrapped native token for a network
     * @param network the network from which to get the token
     * @returns the wrapped native token if one exists, otherwise undefined
     */
    getHlpWrappedNativeToken(network: Network | number): TokenInfo | undefined;
    /**
     * Gets the handle liquidity pool token (symbol hLP) for a network
     * @param network the network on which to find the token
     * @returns the handle liquidity pool token if it exists, otherwise undefined
     */
    getHandleLiquidityToken(network: Network | number): TokenInfo | undefined;
    /**
     * Parses a token into a hLP wrapped native token if it is a native token
     * @param token the token to check
     * @returns an object with the properties isNative and hlpAddress. If the token is native, isNative is
     * true and hlpAddress is the address of a hlp compatible wrapped native token. If the token is a hLP
     * token, isNative is false and hlpAddress is the address of the hLP token.
     * @throws if the token is neither native, nor a hLP supported token, nor the handle liquidity token
     * @throws if the token is native, and no hLP compatible wrapped native token exists
     * @deprecated in favour of parseNativeToWrapped
     */
    checkForHlpNativeToken(token: TokenInfo): {
        isNative: boolean;
        hlpAddress: string;
    };
    /**
     * Parses a token into a wrapped native token if it is native
     * @param token the token to parse
     * @returns an object with the properties isInputNative and parsedToken. If the token passed into the
     * function is native, isInputNative will be set to true, and parsedToken will be the native wrapped token
     * for the network. If the token passed in is not native, isInputNative will be set to false and parsedToken
     * will be the token passed into the function
     */
    parseNativeToWrapped(token: TokenInfo): {
        isInputNative: boolean;
        parsedToken: TokenInfo;
    };
}
export declare const HandleTokenManagerInstance: HandleTokenManager;
export default HandleTokenManager;
