import { TokenList, TokenInfo } from "@uniswap/token-lists";
import { Network } from "../../types/network";
type TokenListCache = {
    [url: string]: TokenList;
};
export declare const DEFAULT_TOKEN_LIST_URLS: string[];
type SearchTokenAddress = {
    address: string;
    network?: Network | number;
};
type SearchTokenSymbols = {
    symbol: string;
    network?: Network | number;
};
export type TokenManagerConfig = {
    tokenListUrls?: string[];
    tokens?: TokenInfo[];
};
export declare const DEFAULT_CONFIG: TokenManagerConfig;
/**
 * The TokenList class is used to fetch and validate token lists.
 */
declare class TokenManager {
    /** Caches fetched results indefinitely */
    protected cache: TokenListCache;
    protected customTokens: TokenInfo[];
    initialLoad: Promise<TokenList[]>;
    /** Called whenever the cache, or custom tokens changes */
    onTokensChange: () => void;
    constructor({ tokenListUrls, tokens }?: TokenManagerConfig);
    /**
     * extracts tokens from the token token lists to a flat array of all the tokens
     * @param tokenLists token lists to get tokens from
     * @returns the tokens from all token lists
     */
    protected getTokensFromLists(tokenLists: TokenList[]): TokenInfo[];
    /**
     * @param urls urls to fetch
     * @returns cached token lists if the cached urls exist, otherwise fetches and returns the tokenlist
     */
    getTokensFromUrls(urls: string[]): Promise<TokenInfo[]>;
    /**
     * Removes tokens that are duplicates of each other by comparing their symbol and chainId
     * @param tokens tokens from which to remove duplicates
     * @returns the tokens with duplicates removed
     */
    protected static removeDuplicates(tokens: TokenInfo[]): TokenInfo[];
    /**
     * @returns all tokens from all cached token lists
     */
    getLoadedTokens(network?: Network | number, removeDuplicates?: boolean): TokenInfo[];
    /**
     * @param symbol symbol of the token to get
     * @param network the network to get the token from
     * @returns the first occurence of the token with the given symbol, or undefined if not found
     */
    getTokenBySymbol<Symbol extends string>(symbol: Symbol, network: Network | number): TokenInfo & {
        symbol: Symbol;
    };
    /**
     * @param symbol symbol of the token to get
     * @param network the network to get the token from
     * @returns the first occurence of the token with the given symbol, or undefined if not found
     */
    tryGetTokenBySymbol<Symbol extends string>(symbol: Symbol, network: Network | number): (TokenInfo & {
        symbol: Symbol;
    }) | undefined;
    /**
     * @param address address of the token to get
     * @param network the network to get the token from
     * @param find optional function to find the token with specific criteria
     * @returns the first occurence of the token with the given address, or undefined if not found
     */
    getTokenByAddress(address: string, network: Network | number, find?: (t: TokenInfo) => boolean): TokenInfo;
    /**
     * @param address address of the token to get
     * @param network the network to get the token from
     * @param find optional function to find the token with specific criteria
     * @returns the first occurence of the token with the given address, or undefined if not found
     */
    tryGetTokenByAddress(address: string, network: Network | number, find?: (t: TokenInfo) => boolean): TokenInfo | undefined;
    /**
     * Finds the native token for a network
     * @param network the network from which to get the token
     * @returns the native token if one exists, otherwise undefined
     */
    getNativeToken(network: Network | number): TokenInfo | undefined;
    /**
     * Gets the wrapped native token for a network
     * @param network the network from which to get the token
     * @returns the wrapped native token if one exists, otherwise undefined
     */
    getWrappedNativeToken(network: Network | number): TokenInfo;
    /**
     * Tries to get the wrapped native token for a network
     * @param network the network from which to get the token
     * @returns the wrapped native token if one exists, otherwise undefined
     */
    tryGetWrappedNativeToken(network: Network | number): TokenInfo | undefined;
    /**
     * Returns an array of tokens with the given addresses. Order is not guaranteed.
     * If a token cannot be found with the given address, it will be omitted from the array.
     * If multiple tokens are found with the same address and network, they will all be included.
     * @param search an array of objects with address as the address of the token, and network as the
     * network of the token
     * @returns an array of tokens with the given addresses.
     * @note this is not optimised for large addresses arrays
     */
    getTokensByAddresses(search: SearchTokenAddress[]): TokenInfo[];
    /**
     * Returns an array of tokens with the given symbol. Order is not guaranteed.
     * If a token cannot be found with the given address, it will be omitted from the array.
     * If multiple tokens are found with the same address and network, they will all be included.
     * @param search an array of objects with address as the address of the token, and network as the
     * network of the token
     * @returns an array of tokens with the given symbol.
     * @note this is not optimised for large symbol arrays
     */
    getTokensBySymbols(search: SearchTokenSymbols[]): TokenInfo[];
    /**
     * fetches a token list
     * @param url url to fetch
     * @returns the token list
     * @throws if the token list is invalid
     */
    fetchTokenList(url: string): Promise<TokenList>;
    /**
     * Fetches a list of tokenLists from the given urls
     * @param urls urls to fetch
     * @returns The fetched token lists
     */
    fetchTokenLists(urls: string[]): Promise<TokenList[]>;
    /**
     * Adds custom tokens to the token manager
     * @param tokens tokens to add
     */
    addCustomTokens(tokens: TokenInfo[]): void;
    /**
     * Clears all custom tokens
     */
    clearCustomTokens: () => void;
    /**
     * Sets custom tokens as a list of tokens
     * @param tokens tokens to set
     */
    setCustomTokens: (tokens: TokenInfo[]) => void;
    /**
     * @returns all custom tokens
     */
    getCustomTokens: () => TokenInfo[];
    /**
     * Gets a token list from the cache
     * @param key the key of the token list in the cache
     * @returns the token list with the given key
     */
    getFromCache(key: string): TokenList | undefined;
    /**
     * Sets a cache key to a token list
     * @param key the key in the cache for which to set the tokenList
     * @param tokenList the tokenList to set
     */
    setTokenList(key: string, tokenList: TokenList): void;
    /**
     * Deletes a tokenList in the cache
     * @param key the key in the cache to delete
     */
    deleteTokenList(key: string): void;
    /**
     * Clears the token list cache
     */
    clearCache(): void;
    /**
     * Gets the cache
     * @returns the cache
     */
    getCache(): TokenListCache;
}
export declare const getNativeTokenInfo: (network: Network) => TokenInfo;
export default TokenManager;
