import { BigNumber, Signer } from "ethers";
import { DepositOrWithdrawal, PeriodicFee, TradeAdapter, TradeLiquidityPool } from ".";
import { TradeProtocol } from "./protocol";
import { TradeReader } from "./reader";
import { TokenInfo } from "@uniswap/token-lists";
import { ActiveOrder, GetHistoryArgs, OrderArgs, SimulateTradeArgs, TradeAction, TradeArgs, TradeEffect, TradePairId, TradeSimulation } from "./index";
import { Position } from "./position";
export type AccountOrderArgs = OrderArgs & {
    lpId: string;
};
export type DepositArgs = {
    signer: Signer;
    amount: BigNumber;
    token: TokenInfo;
    useGasless?: boolean;
    psmToken?: string;
};
export type WithdrawArgs = {
    signer: Signer;
    amount: BigNumber;
    receiver: string;
    token: TokenInfo;
    psmToken?: string;
};
/**
 * @property userAddress User address to which the role will be granted.
 * @property role The trade account role to grant.
 * @property signer The account signer to perform the action.
 */
export type GrantRoleArgs = {
    userAddress: string;
    role: TradeAccountRole;
    signer: Signer;
};
/**
 * @property userAddress User address from which the role will be revoked.
 * @property role The trade account role to grant.
 * @property signer The account signer to perform the action.
 */
export type RevokeRoleArgs = {
    userAddress: string;
    role: TradeAccountRole;
    signer: Signer;
};
export type TransferArgs = {};
export type GetRolesArgs = {
    userAddress: string;
};
export type AccountTradeResponse = {
    tradeEffect: TradeEffect;
};
export declare enum TradeAccountRole {
    None = 0,
    Owner = 1,
    Trader = 2,
    Withdraw = 3,
    Deposit = 4,
    Open = 5,
    ProtocolAdmin = 6
}
interface ITradeAccount {
    readonly id: number;
    getPosition(pair: TradePairId): Position;
    getAllPositions(): Position[];
    getTradeHistory(args: GetHistoryArgs): Promise<TradeAction[]>;
    getDepositWithdrawHistory(args: GetHistoryArgs): Promise<DepositOrWithdrawal[]>;
    getPeriodicFeeHistory(args: GetHistoryArgs): Promise<PeriodicFee[]>;
    getActiveOrders(): Promise<ActiveOrder[]>;
    getEquity(lps?: TradeLiquidityPool[]): BigNumber;
    getRealisedEquity(): BigNumber;
    getUnrealisedEquity(lps?: TradeLiquidityPool[]): BigNumber;
    getAvailableEquity(lps?: TradeLiquidityPool[]): BigNumber;
    getInitialMarginRequirement(): BigNumber;
    getMaintenanceMarginRequirement(): BigNumber;
    getOpenInterest(): BigNumber;
    getLeverage(): BigNumber;
    getPositionLeverage(id: TradePairId): BigNumber;
    trade(args: TradeArgs): Promise<AccountTradeResponse>;
    simulateTrade(args: SimulateTradeArgs): TradeSimulation;
    deposit(args: DepositArgs): Promise<void>;
    withdraw(args: WithdrawArgs): Promise<void>;
    order(args: AccountOrderArgs): Promise<void>;
    grantRole(args: GrantRoleArgs): Promise<void>;
    revokeRole(args: RevokeRoleArgs): Promise<void>;
    transfer(args: TransferArgs): Promise<void>;
    getRolesArgs(args: GetRolesArgs): Promise<TradeAccountRole[]>;
}
export type AccountOpenArgs = {
    amount: BigNumber;
    wallet: Signer;
    protocol: TradeProtocol;
    adapter: TradeAdapter;
    equityToken: TokenInfo;
    reader: TradeReader;
    referralCode?: string;
    useGasless?: boolean;
    psmToken?: string;
};
/**
 * A trade account instance.
 * The instance is connected to a specific equity token, and will
 * only have access to LPs with that same underlying token.
 */
export declare class TradeAccount implements ITradeAccount {
    readonly id: number;
    readonly equityTokenAddress: string;
    private readonly protocol;
    private readonly adapter;
    private readonly reader;
    private readonly remoteTimestamp;
    private realizedEquity;
    private positions;
    onUpdate?: () => void;
    private subsriptionId;
    private isProcessingSubscriptionRequest;
    constructor(id: number, realizedEquity: BigNumber, positions: Position[], protocol: TradeProtocol, adapter: TradeAdapter, reader: TradeReader, equityTokenAddress: string, remoteTimestamp?: number);
    subscribeToRemoteUpdates(): void;
    cancelSubscriptionToRemoteUpdates(): void;
    private handleAccountPublication;
    getPosition(pair: TradePairId): Position;
    getAllPositions(): Position[];
    getTradeHistory(args: GetHistoryArgs): Promise<TradeAction[]>;
    getDepositWithdrawHistory(args: GetHistoryArgs): Promise<DepositOrWithdrawal[]>;
    getPeriodicFeeHistory(args: GetHistoryArgs): Promise<PeriodicFee[]>;
    getActiveOrders(): Promise<ActiveOrder[]>;
    /**
     * Returns the available equity for the account.
     * @param lps Liquidity pool references used for calculating unrealised
     * borrow and funding position fees. Optional, but if
     * not passed the position fees will not be included in the calculation.
     */
    getEquity(lps?: TradeLiquidityPool[]): BigNumber;
    getRealisedEquity(): BigNumber;
    /**
     * Returns the available equity for the account.
     * @param lps Liquidity pool references used for calculating unrealised
     * borrow and funding position fees. Optional, but if
     * not passed the position fees will not be included in the calculation.
     */
    getAvailableEquity(lps?: TradeLiquidityPool[]): BigNumber;
    /**
     * Returns the unrealised equity for the account.
     * @param lps Liquidity pool references used for calculating unrealised
     * borrow and funding position fees. Optional, but if
     * not passed the position fees will not be included in the calculation.
     */
    getUnrealisedEquity(lps?: TradeLiquidityPool[]): BigNumber;
    getInitialMarginRequirement(): BigNumber;
    getMaintenanceMarginRequirement(): BigNumber;
    /**
     * @returns The total value of all positions in the liquidity token currency
     * (currently USD).
     */
    getOpenInterest(): BigNumber;
    getLeverage(): BigNumber;
    getPositionLeverage(id: TradePairId): BigNumber;
    trade(args: TradeArgs): Promise<AccountTradeResponse>;
    private applyTradeEffect;
    simulateTrade(args: SimulateTradeArgs): TradeSimulation;
    deposit(args: DepositArgs): Promise<void>;
    withdraw(args: WithdrawArgs): Promise<void>;
    static open({ amount, wallet, protocol, adapter, equityToken, reader, referralCode, useGasless, psmToken, }: AccountOpenArgs): Promise<TradeAccount>;
    static fromId(id: number, reader: TradeReader, protocol: TradeProtocol, adapter: TradeAdapter, equityTokenAddress: string): Promise<TradeAccount>;
    order(_: AccountOrderArgs): Promise<void>;
    grantRole(args: GrantRoleArgs): Promise<void>;
    revokeRole(args: RevokeRoleArgs): Promise<void>;
    transfer(_: TransferArgs): Promise<void>;
    getRolesArgs(_: GetRolesArgs): Promise<TradeAccountRole[]>;
    timestamp(): number;
}
export {};
