import { Bytes, ethers } from "ethers";
import { Pair, PairStringified } from "../../types/trade";
import { TradeAccountRole } from "./account";
import { ClearSystemParamArgs, LpId, SetSystemParamArgs, TradeSize } from "./interface";
export type Request = {
    id?: string;
    method: string;
    params?: unknown;
};
export type Response<C> = {
    id?: string;
    result?: {
        type: string;
        content: C;
    };
    error?: string;
};
export type PublicationResponse<C> = Response<{
    topic: string;
    content: C;
}>;
export type OpenAccountResponse = Response<{
    openAccount: {
        accountId: number;
        amount: string;
        timestampUnixMillis: number;
        owner: string;
        token: string;
        referralCode: string | undefined;
        signature: string;
    };
}>;
export type DepositResponse = Response<{
    deposit: {
        accountId: number;
        amount: string;
        timestampUnixMillis: number;
        depositor: string;
        token: string;
        signature: string;
    };
}>;
export type WithdrawResponse = Response<{
    withdraw: {
        accountId: number;
        amount: string;
        timestampUnixMillis: number;
        accountUser: string;
        token: string;
        recipient: string;
        signature: string;
    };
}>;
export type TradeResponse = Response<{
    trade: {
        pair: string;
        accountId: number;
        lpId: string;
        price: string;
        size: string;
        marginFee: string;
        timestampUnixMillis: number;
        accountUser: string;
        signature: string;
    };
}>;
export type GrantAccountUserRoleResponse = Response<{
    grantAccountUserRole: {
        accountId: number;
        user: string;
        role: TradeAccountRole;
        accountOwner: string;
        signature: string;
    };
}>;
export type RevokeAccountUserRoleResponse = Response<{
    revokeAccountUserRole: {
        accountId: number;
        user: string;
        role: TradeAccountRole;
        accountOwner: string;
        signature: string;
    };
}>;
export type LpConfigUpdateResponse = Response<{
    lpConfigUpdate: {
        lpId: string;
        pair: string;
        config: {
            initialMarginFraction: string;
            maintenanceMarginFraction: string;
            incrementalInitialMarginFraction: string;
            baselinePositionSize: string;
            incrementalPositionSize: string;
            marginFeeFraction: string;
            symmetricalSpreadFraction: string;
            isActive: boolean;
            isReduceOnly: boolean;
            borrowFeeFactor: string;
            fundingFactor: string;
            fundingExponent: string;
            maxOpenInterestDiff: string | null;
            maxOpenInterestLong: string | null;
            maxOpenInterestShort: string | null;
            usePriceImpact: boolean;
            priceImpactFraction: string | null;
            skewScale: string | null;
        };
    }[];
}>;
export interface TradeAdapter {
    getOpenAccountMessage(signerAddress: string): Promise<Bytes>;
    getDepositMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getWithdrawMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getTradeMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getOwnerMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    openAccount(amount: string, tokenAddress: string, ownerAddress: string, signature: string, referralCode?: string, useGasless?: boolean, psmToken?: string): Promise<OpenAccountResponse>;
    deposit(amount: string, accountId: number, tokenAddress: string, depositorAddress: string, signature: string, useGasless?: boolean, psmToken?: string): Promise<DepositResponse>;
    withdraw(amount: string, accountId: number, accountUser: string, tokenAddress: string, recipientAddress: string, signature: string, psmToken?: string): Promise<WithdrawResponse>;
    trade(accountId: number, liquidityPoolId: string, size: TradeSize, pair: Pair, accountUser: string, signature: string): Promise<TradeResponse>;
    grantAccountUserRole(accountId: number, user: string, role: TradeAccountRole, accountOwner: string, signature: string): Promise<GrantAccountUserRoleResponse>;
    revokeAccountUserRole(accountId: number, user: string, role: TradeAccountRole, accountOwner: string, signature: string): Promise<RevokeAccountUserRoleResponse>;
    setSystemParam(args: SetSystemParamArgs): Promise<void>;
    clearSystemParam(args: ClearSystemParamArgs): Promise<void>;
    getLpConfig(): Promise<LpConfigUpdateResponse>;
    waitForConnect(): Promise<void>;
    subscribeToAccount(accountId: number, listener: AccountListener): Promise<TradeSubscriptionId>;
    subscribeToLp(lpId: LpId, listener: LpListener): Promise<TradeSubscriptionId>;
    cancelSubscription(subscriptionId: string): void;
}
export type LpPairTradeabilityListener = (lpId: LpId, pair: Pair, isTradeable: boolean) => void;
export type SubscriptionListener<C> = (content: C, topic: string) => void;
export type AccountSubscriptionContent = {
    id: number;
    realizedEquity: string;
    positions: SerializedPosition[];
};
export type SerializedPosition = {
    pair: string;
    lpId: string;
    size: string;
    entryPrice: string;
    snapshotSumFractionFunding: string;
    snapshotSumFractionBorrow: string;
};
export type AccountListener = SubscriptionListener<AccountSubscriptionContent>;
export type LpPairPublication<C> = {
    lpPair: SerializedLpPair;
    content: C;
};
export type SerializedLpPair = {
    lpId: LpId;
    pair: PairStringified;
};
export type LpPublication = LpPairPublication<boolean> | SerializedPairState;
export type LpListener = SubscriptionListener<LpPublication>;
export type SerializedPairState = {
    lpPair: SerializedLpPair;
    sumFractionFunding: SerializedTimestampedBigDecimalMarketSide;
    sumFractionBorrow: SerializedTimestampedBigDecimalMarketSide;
    openInterest: SerializedBigDecimalMarketSide;
};
export type SerializedTimestampedBigDecimalMarketSide = {
    value: SerializedBigDecimalMarketSide;
    timestamp: number;
};
export type SerializedBigDecimalMarketSide = {
    long: string;
    short: string;
};
export type TradeSubscriptionId = string;
export type SubscriptionResponse = Response<TradeSubscriptionId>;
export type WebsocketListener = (message: string) => void;
export declare class TradeAdapterWebsocket implements TradeAdapter {
    private socket;
    private connectionPromise;
    private listeners;
    private activeSubscriptions;
    private subscriptionRequests;
    protected resubscriptionResults: string[];
    provider: ethers.providers.Provider;
    private config;
    private constructor();
    static create(wsUrl: string, provider: ethers.providers.Provider, config?: {
        accountAddress: string;
        timeout: number;
    }): Promise<TradeAdapterWebsocket>;
    waitForConnect(): Promise<void>;
    reconnect(newWsUrl?: string): Promise<void>;
    private getUserNonce;
    getOpenAccountMessage(signerAddress: string): Promise<Bytes>;
    getDepositMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getWithdrawMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getTradeMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getAdminMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    getOwnerMessage(signerAddress: string, accountId: number): Promise<Bytes>;
    private getUserRoleMessage;
    openAccount(amount: string, tokenAddress: string, ownerAddress: string, signature: string, referralCode?: string, useGasless?: boolean, psmToken?: string): Promise<OpenAccountResponse>;
    deposit(amount: string, accountId: number, tokenAddress: string, depositorAddress: string, signature: string, useGasless?: boolean, psmToken?: string): Promise<DepositResponse>;
    withdraw(amount: string, accountId: number, accountUser: string, tokenAddress: string, recipientAddress: string, signature: string, psmToken?: string): Promise<WithdrawResponse>;
    trade(accountId: number, liquidityPoolId: string, size: TradeSize, pair: Pair, accountUser: string, signature: string): Promise<TradeResponse>;
    grantAccountUserRole(accountId: number, user: string, role: TradeAccountRole, accountOwner: string, ownerSignature: string): Promise<GrantAccountUserRoleResponse>;
    revokeAccountUserRole(accountId: number, user: string, role: TradeAccountRole, accountOwner: string, signature: string): Promise<RevokeAccountUserRoleResponse>;
    setSystemParam(args: SetSystemParamArgs): Promise<void>;
    clearSystemParam(args: ClearSystemParamArgs): Promise<void>;
    getLpConfig(): Promise<LpConfigUpdateResponse>;
    /** This function has no validation that the returned type is T */
    private waitForRequestId;
    close(): void;
    subscribeToAccount(accountId: number, listener: AccountListener): Promise<TradeSubscriptionId>;
    subscribeToLp(lpId: LpId, listener: LpListener): Promise<TradeSubscriptionId>;
    cancelSubscription(subscriptionId: TradeSubscriptionId): void;
    private subscribe;
    private pushUniqueSubscriptionRequest;
    private resubscribe;
    private sendRequestRetry;
    private sendRequest;
    /**
     * Sets up a listener for the subscription.
     * The listener will loop through sever updates for as long as the client
     * keeps the subscription active.
     * When called, this function should not be awaited as it is meant to be
     * executed as a background task.
     * @private
     */
    private setupSubscriptionListener;
}
