import { ethers, utils } from "ethers";
import websocket from "websocket";
import { Account__factory } from "../../contracts";
import { TradeAccountRole } from "./account";
import { pairToString } from "../../utils/general";
import config from "../../config";
const MESSAGE_SCOPE = utils.keccak256(utils.toUtf8Bytes("HANDLE_SYNTH_ACCOUNT_MESSAGE"));
const OPEN_ACCOUNT_ID = 0;
const MAX_SOCKET_SEND_TRIES = 5;
const SOCKET_RETRY_DELAY = 500;
const DEFAULT_CONFIG = {
    // The address of the Account contract.
    accountAddress: config.protocol["arbitrum-sepolia"].tradeAccount,
    // The timeout for waiting for a response from the server ws.
    timeout: 15000,
};
export class TradeAdapterWebsocket {
    constructor(provider, config = DEFAULT_CONFIG) {
        this.listeners = [];
        // TODO: manage one-to-many server-to-client subscriptions?
        this.activeSubscriptions = new Set();
        this.subscriptionRequests = [];
        this.resubscriptionResults = [];
        this.config = DEFAULT_CONFIG;
        this.provider = provider;
        this.config = config;
    }
    static async create(wsUrl, provider, config = DEFAULT_CONFIG) {
        const adapter = new TradeAdapterWebsocket(provider, config);
        await adapter.reconnect(wsUrl);
        return adapter;
    }
    async waitForConnect() {
        const isOpen = () => this.socket.readyState === websocket.w3cwebsocket.OPEN;
        if (isOpen()) {
            return;
        }
        // Check every 250 ms whether it is connected, or times out.
        return new Promise(async (resolve, reject) => {
            let hasResolved = false;
            let hasTimedOut = false;
            setTimeout(() => {
                if (hasResolved) {
                    return;
                }
                hasTimedOut = true;
                reject();
            }, this.config.timeout);
            while (!hasTimedOut && !hasResolved) {
                await new Promise((r) => setTimeout(r, 250));
                if (isOpen()) {
                    hasResolved = true;
                    resolve();
                }
            }
        });
    }
    async reconnect(newWsUrl) {
        if (!newWsUrl && !this.socket) {
            throw new Error("No WS URL given for connection");
        }
        console.debug("[trade adapter] ws (re-)connection triggered");
        return new Promise((resolve, reject) => {
            this.socket = new websocket.w3cwebsocket(newWsUrl ?? this.socket.url);
            this.socket.onmessage = (message) => {
                this.listeners.forEach((l) => l(message.data.toString()));
            };
            this.socket.onclose = (event) => {
                console.debug("[trade adapter] connection closed: ", event);
                // Only attempt to reconnect if the close was not expected/intentional.
                this.reconnect();
            };
            this.socket.onopen = async () => {
                await this.resubscribe();
                console.debug("[trade adapter]: connection (re-)established");
                resolve();
            };
            this.socket.onerror = (error) => {
                console.error("[trade adapter] error:", error);
                reject(error);
            };
        });
    }
    async getUserNonce(signerAddress) {
        return Account__factory.connect(this.config.accountAddress, this.provider).userNonce(signerAddress);
    }
    async getOpenAccountMessage(signerAddress) {
        const userNonce = await this.getUserNonce(signerAddress);
        return this.getUserRoleMessage(userNonce, OPEN_ACCOUNT_ID, TradeAccountRole.Open);
    }
    async getDepositMessage(signerAddress, accountId) {
        const userNonce = await this.getUserNonce(signerAddress);
        return this.getUserRoleMessage(userNonce, accountId, TradeAccountRole.Deposit);
    }
    async getWithdrawMessage(signerAddress, accountId) {
        const userNonce = await this.getUserNonce(signerAddress);
        return this.getUserRoleMessage(userNonce, accountId, TradeAccountRole.Withdraw);
    }
    async getTradeMessage(signerAddress, accountId) {
        const userNonce = await this.getUserNonce(signerAddress);
        return this.getUserRoleMessage(userNonce, accountId, TradeAccountRole.Trader);
    }
    async getAdminMessage(signerAddress, accountId) {
        const userNonce = await this.getUserNonce(signerAddress);
        return this.getUserRoleMessage(userNonce, accountId, TradeAccountRole.ProtocolAdmin);
    }
    async getOwnerMessage(signerAddress, accountId) {
        const userNonce = await this.getUserNonce(signerAddress);
        return this.getUserRoleMessage(userNonce, accountId, TradeAccountRole.Owner);
    }
    getUserRoleMessage(nonce, accountId, role) {
        const message = utils.defaultAbiCoder.encode(["bytes32", "uint256", "uint256", "uint8"], [MESSAGE_SCOPE, nonce, accountId, role]);
        const messageHash = utils.keccak256(message);
        return ethers.utils.arrayify(messageHash);
    }
    async openAccount(amount, tokenAddress, ownerAddress, signature, referralCode, useGasless, psmToken) {
        return this.sendRequest({
            method: "openAccount",
            params: {
                amount,
                token: tokenAddress,
                owner: ownerAddress,
                useGasless: useGasless,
                referralCode,
                signature,
                psmToken,
            },
        });
    }
    async deposit(amount, accountId, tokenAddress, depositorAddress, signature, useGasless, psmToken) {
        return this.sendRequest({
            method: "deposit",
            params: {
                amount,
                accountId: accountId,
                depositor: depositorAddress,
                token: tokenAddress,
                signature,
                useGasless,
                psmToken,
            },
        });
    }
    async withdraw(amount, accountId, accountUser, tokenAddress, recipientAddress, signature, psmToken) {
        return this.sendRequest({
            method: "withdraw",
            params: {
                amount,
                accountId: accountId,
                accountUser: accountUser,
                token: tokenAddress,
                recipient: recipientAddress,
                signature,
                psmToken,
            },
        });
    }
    async trade(accountId, liquidityPoolId, size, pair, accountUser, signature) {
        return this.sendRequest({
            method: "trade",
            params: {
                id: accountId,
                lpId: liquidityPoolId,
                size: size.serialize(),
                pair: pairToString(pair),
                accountUser: accountUser,
                signature,
            },
        });
    }
    async grantAccountUserRole(accountId, user, role, accountOwner, ownerSignature) {
        return this.sendRequest({
            method: "grantAccountUserRole",
            params: {
                // The trade server expects a string, not a number.
                role: TradeAccountRole[role],
                accountId,
                user,
                accountOwner,
                ownerSignature,
            },
        });
    }
    async revokeAccountUserRole(accountId, user, role, accountOwner, signature) {
        return this.sendRequest({
            method: "revokeAccountUserRole",
            params: {
                accountId,
                user,
                role,
                accountOwner,
                signature,
            },
        });
    }
    async setSystemParam(args) {
        const signerAddress = await args.signer.getAddress();
        const message = await this.getAdminMessage(signerAddress, args.adminTradeAccountId);
        const signature = await args.signer.signMessage(message);
        const response = await this.sendRequest({
            method: "setSystemParam",
            params: {
                paramId: args.paramId,
                paramValue: args.paramValue.toHexString(),
                adminRequest: {
                    tradeAccountId: args.adminTradeAccountId,
                    admin: signerAddress,
                    signature,
                },
            },
        });
        if (response?.error) {
            throw response.error;
        }
    }
    async clearSystemParam(args) {
        const signerAddress = await args.signer.getAddress();
        const message = await this.getAdminMessage(signerAddress, args.adminTradeAccountId);
        const signature = await args.signer.signMessage(message);
        const response = await this.sendRequest({
            method: "clearSystemParam",
            params: {
                paramId: args.paramId,
                adminRequest: {
                    tradeAccountId: args.adminTradeAccountId,
                    admin: signerAddress,
                    signature,
                },
            },
        });
        if (response?.error) {
            throw response.error;
        }
    }
    async getLpConfig() {
        return this.sendRequest({
            method: "getLpConfig",
        });
    }
    /** This function has no validation that the returned type is T */
    waitForRequestId(requestId, timeout = this.config.timeout) {
        // Make a ref object so that the callback can be updated in the promise.
        const cancelRef = {
            cancel: () => { },
            result: null,
        };
        cancelRef.result = new Promise((resolve, reject) => {
            let resolved = false;
            const removeListener = (listener) => {
                this.listeners = this.listeners.filter((l) => l !== listener);
            };
            const listener = (raw) => {
                setTimeout(() => {
                    if (resolved)
                        return;
                    removeListener(listener);
                    reject(new Error("Timeout"));
                }, timeout);
                cancelRef.cancel = () => {
                    removeListener(listener);
                };
                try {
                    const response = JSON.parse(raw);
                    if (response.id !== requestId) {
                        return;
                    }
                    removeListener(listener);
                    resolved = true;
                    resolve(response);
                }
                catch (e) {
                    removeListener(listener);
                    return reject(e);
                }
            };
            this.listeners.push(listener);
        });
        return cancelRef;
    }
    close() {
        this.socket.close();
    }
    async subscribeToAccount(accountId, listener) {
        const subscriptionId = await this.subscribe({
            topic: "tradeAccount",
            params: accountId,
            listener,
        });
        return subscriptionId;
    }
    subscribeToLp(lpId, listener) {
        return this.subscribe({
            topic: "liquidityPool",
            params: lpId,
            listener,
        });
    }
    cancelSubscription(subscriptionId) {
        this.activeSubscriptions.delete(subscriptionId);
    }
    async subscribe(request) {
        // Store the request for re-subscription
        this.pushUniqueSubscriptionRequest(request);
        const payload = {
            method: "subscribe",
            params: {
                topic: request.topic,
                params: request.params,
            },
        };
        console.debug("[trade adapter] sending (re-)subscription request", request);
        const response = await this.sendRequest(payload);
        if (response.error) {
            throw new Error(response.error);
        }
        const subscriptionId = response.result.content;
        // Setup listener.
        this.setupSubscriptionListener(subscriptionId, request.listener).catch(console.error);
        console.debug("[trade adapter] received subscription id", response);
        return subscriptionId;
    }
    pushUniqueSubscriptionRequest(request) {
        const isExisting = this.subscriptionRequests.some((r) => isSameSubscriptionRequest(r, request));
        if (isExisting) {
            return;
        }
        this.subscriptionRequests.push(request);
    }
    async resubscribe() {
        console.debug("[trade adapter] ws re-subscription triggered");
        this.resubscriptionResults = await Promise.all(this.subscriptionRequests.map((request) => this.subscribe(request)));
        return this.resubscriptionResults;
    }
    async sendRequestRetry(stringifiedRequest) {
        let lastError;
        let success = false;
        for (let i = 0; i < MAX_SOCKET_SEND_TRIES; i++) {
            try {
                this.socket.send(stringifiedRequest);
                success = true;
                break;
            }
            catch (e) {
                lastError = e;
            }
            await new Promise((r) => setTimeout(r, SOCKET_RETRY_DELAY));
        }
        if (!success) {
            throw lastError;
        }
    }
    async sendRequest(request) {
        const requestId = request.id ?? randomId();
        if (!request.id) {
            request.id = requestId;
        }
        await this.connectionPromise;
        await this.sendRequestRetry(JSON.stringify(request));
        return this.waitForRequestId(requestId).result;
    }
    /**
     * Sets up a listener for the subscription.
     * The listener will loop through sever updates for as long as the client
     * keeps the subscription active.
     * When called, this function should not be awaited as it is meant to be
     * executed as a background task.
     * @private
     */
    async setupSubscriptionListener(subscriptionId, listener) {
        // Because this is setting up the subscription, it sets it as active.
        this.activeSubscriptions.add(subscriptionId);
        const wsListener = (raw) => {
            const response = JSON.parse(raw);
            if (response.id !== subscriptionId) {
                return;
            }
            if (response.error) {
                console.error(`[trade adapter] subscription error:`, response);
                return;
            }
            const result = response.result;
            const content = result.content.content;
            listener(content, result.content.topic);
        };
        this.listeners.push(wsListener);
        // Over the lifetime of this function, the subscription may be cancelled.
        const isActive = () => this.activeSubscriptions.has(subscriptionId);
        // Fire background task to periodically check for subscription status.
        while (isActive()) {
            // While the subscription is active, wait.
            // The waiting interval here is the timeout but this is not really
            // a timeout, but rather it is just an arbitrary waiting period.
            await new Promise((resolve) => setTimeout(resolve, this.config.timeout));
        }
        // The subscription has been cancelled by the client.
        this.listeners = this.listeners.filter((l) => l !== wsListener);
    }
}
const randomId = (len = 6) => Math.random()
    .toString(36)
    .substring(len + 1);
const isSameSubscriptionRequest = (a, b) => a.topic === b.topic &&
    a.listener === b.listener &&
    JSON.stringify(a.params) === JSON.stringify(b.params);
