import { BigNumber, Signer } from "ethers";
import { Pair } from "../../types/trade";
import { Subscription, SubscriptionId } from "../../types/general";
import { TokenInfo } from "@uniswap/token-lists";
import { TradeAccount } from "./account";
import { LpPairTradeabilityListener } from "./adapter";
import { MarketSide, PairState } from "./pairState";
import { z } from "zod";
export type LpId = string;
export type TradePairIdState = {
    pair: Pair;
    lpId: LpId;
};
export declare class TradePairId implements TradePairIdState {
    readonly pair: Pair;
    readonly lpId: string;
    constructor(pair: Pair, lpId: string);
    eq(other: TradePairId): boolean;
    toString(): string;
    static fromString(s: string): TradePairId;
    static fromState(state: TradePairIdState): TradePairId;
}
export type H2SOPairData = {
    pair: Pair;
};
/**
 * This is analogous to the `PairConfig` type in the
 * server.
 */
export declare class TradePair {
    id: TradePairId;
    initialMarginFraction: BigNumber;
    maintenanceMarginFraction: BigNumber;
    incrementalInitialMarginFraction: BigNumber;
    baselinePositionSize: BigNumber;
    incrementalPositionSize: BigNumber;
    marginFeeFraction: BigNumber;
    spreadFraction: BigNumber;
    isActive: boolean;
    isReduceOnly: boolean;
    maxOpenInterestDiff: BigNumber | null;
    maxOpenInterestLong: BigNumber | null;
    maxOpenInterestShort: BigNumber | null;
    borrowFeeFactor: BigNumber;
    fundingFactor: BigNumber;
    fundingExponent: BigNumber;
    h2soPairData: H2SOPairData;
    usePriceImpact: boolean;
    priceImpactFraction: BigNumber | null;
    skewScale: BigNumber | null;
    readonly pair: Pair;
    private static readonly divisor;
    constructor(id: TradePairId, initialMarginFraction: BigNumber, maintenanceMarginFraction: BigNumber, incrementalInitialMarginFraction: BigNumber, baselinePositionSize: BigNumber, incrementalPositionSize: BigNumber, marginFeeFraction: BigNumber, spreadFraction: BigNumber, isActive: boolean, isReduceOnly: boolean, maxOpenInterestDiff: BigNumber | null, maxOpenInterestLong: BigNumber | null, maxOpenInterestShort: BigNumber | null, borrowFeeFactor: BigNumber, fundingFactor: BigNumber, fundingExponent: BigNumber, h2soPairData: H2SOPairData, usePriceImpact: boolean, priceImpactFraction: BigNumber | null, skewScale: BigNumber | null);
    private getLargerSideFundingRate;
    shouldUsePriceImpact(): boolean;
    /**
     * @returns the funding rates with `AMOUNT_DECIMALS` precision.
     */
    getFundingRate(openInterest: OpenInterest): MarketSide<BigNumber>;
    getLongBorrowRate(openInterest: OpenInterest): BigNumber;
    getShortBorrowRate(openInterest: OpenInterest): BigNumber;
    /**
     * @returns the borrow rates with `AMOUNT_DECIMALS` precision.
     */
    getBorrowRate(openInterest: OpenInterest): MarketSide<BigNumber>;
    getInitialMargin(size: BigNumber, price: BigNumber): BigNumber;
    getMaintenanceMargin(size: BigNumber, price: BigNumber): BigNumber;
    getMarginFee(size: BigNumber, price: BigNumber): BigNumber;
}
export type TradeArgs = {
    pairId: TradePairId;
    size: TradeSize;
    signer: Signer;
};
export type SimulateTradeArgs = {
    pairId: TradePairId;
    size: BigNumber;
    price: Readonly<MarketPrice>;
    gasFee: BigNumber;
};
export type LpTradeArgs = TradeArgs & {
    oldSize: BigNumber;
    accountId: number;
};
export type SetSystemParamArgs = {
    signer: Signer;
    adminTradeAccountId: number;
    paramId: string;
    paramValue: BigNumber;
};
export type ClearSystemParamArgs = {
    signer: Signer;
    adminTradeAccountId: number;
    paramId: string;
};
export type OrderArgs = {
    pairId: TradePairId;
    size: BigNumber;
    limitPrice: BigNumber;
};
export declare class MarketPrice {
    index: BigNumber;
    private skew?;
    bestBid: BigNumber;
    bestAsk: BigNumber;
    constructor(index: BigNumber, bestBid: BigNumber, bestAsk: BigNumber, skew?: BigNumber);
    /**
     * Calculates the inverse of this market price.
     * @note This does not modify the current object.
     * @returns The inverse of this market price.
     */
    inverse(): MarketPrice;
    get marketPrice(): BigNumber;
    static fromIndex(index: BigNumber, spread?: BigNumber): MarketPrice;
    static zero(): MarketPrice;
    getTradePrice(isBuying: boolean): BigNumber;
    getTradePriceFromTradeSize(tradeSize: BigNumber): BigNumber;
}
export type MarketPriceSubscription = Subscription<MarketPriceCallback, Pair>;
export type SubscribeMarketPriceArgs = {
    callback: MarketPriceCallback;
    pair: Pair;
};
export type MarketPriceCallback = (pair: Readonly<Pair>, price: Readonly<MarketPrice>) => any;
/**
 * Available liquidity, in the base token, for this trade pair.
 * long is returned as an amount of the index token
 * short is returned as a usd amount, to 30 decimals
 */
export type AvailableLiquidity = {
    long: BigNumber;
    short: BigNumber;
};
export type GetAvailableLiquidityArgs = {
    pair: Pair;
    forceFetch?: boolean;
};
export type GetAvailabilityArgs = {
    pair: Pair;
};
export type Availability = {
    isAvailable: true;
    reason?: undefined;
} | {
    isAvailable: false;
    reason?: string;
};
export type GetHistoryArgs = Partial<TradePairId> & {
    skip?: number;
    limit?: number;
    pairId?: TradePairId;
    startTimestamp?: number;
};
export declare class TradeAction {
    readonly pairId: TradePairId;
    readonly size: BigNumber;
    readonly price: BigNumber;
    readonly timestamp: number;
    readonly realisedEquity: BigNumber;
    readonly marginFee: BigNumber;
    readonly type: "trade" | "liquidation";
    readonly txHash: string;
    readonly didOpenPosition: boolean;
    readonly didClosePosition: boolean;
    constructor(pairId: TradePairId, size: BigNumber, price: BigNumber, timestamp: number, realisedEquity: BigNumber, marginFee: BigNumber, type: "trade" | "liquidation", txHash: string, didOpenPosition: boolean, didClosePosition: boolean);
    eq(other: TradeAction): boolean;
    isLiquidation(): boolean;
}
/**
 * A USD deposit or withdrawal.
 * @note amount will be negative for withdrawals.
 */
export type DepositOrWithdrawal = {
    amount: BigNumber;
    timestamp: number;
    txHash: string;
};
export type PeriodicFee = {
    type: "borrow" | "funding";
    pairId: TradePairId;
    amount: BigNumber;
    timestamp: number;
};
export type ActiveOrder = TradePairId & {
    size: BigNumber;
    limitPrice: BigNumber;
};
export type TradeEffect = {
    fillPrice: BigNumber;
    marginFee: BigNumber;
};
export type TradeSimulation = {
    effect: TradeEffect;
    nextAccount: TradeAccount;
    spreadFee: BigNumber;
    failureReason?: string;
};
export type GetTradePriceArgs = {
    pair: Pair;
    size: BigNumber;
};
export declare class OpenInterest {
    long: BigNumber;
    short: BigNumber;
    constructor(long: BigNumber, short: BigNumber);
    total(): BigNumber;
    static zero(): OpenInterest;
}
/**
 * An interface for liquidity pools within the synths platform.
 */
export interface TradeLiquidityPool {
    id: string;
    onUpdate?: () => void;
    trade(args: LpTradeArgs): Promise<TradeEffect>;
    order(args: OrderArgs): Promise<void>;
    getLpTokenPrice(): Promise<BigNumber>;
    getPrice(pair: Pair, size?: BigNumber): Readonly<MarketPrice>;
    getTradePrice(args: GetTradePriceArgs): BigNumber;
    subscribeToMarketPrice(args: SubscribeMarketPriceArgs): SubscriptionId;
    unsubscribeFromMarketPrice(id: SubscriptionId): void;
    getTradePairs(): TradePair[];
    getUnderlyingToken(): TokenInfo;
    getTradePair(pair: Pair): TradePair;
    getPairState(pair: Pair): PairState;
    getAvailableLiquidity(args: GetAvailableLiquidityArgs): Promise<AvailableLiquidity>;
    getPairAvailability(args: GetAvailabilityArgs): Availability;
    getOpenInterest(args: {
        pair: Pair;
    }): OpenInterest;
    listenToLpPairTradeability(listener: LpPairTradeabilityListener): void;
}
export type GetMarginArgs = {
    size: BigNumber;
};
export type TradeSizeType = "Lpc" | "Lot";
export declare class TradeSize {
    readonly amount: BigNumber;
    readonly type: TradeSizeType;
    private constructor();
    static fromLot(amount: BigNumber): TradeSize;
    static fromLpc(amount: BigNumber): TradeSize;
    lots(price: BigNumber): BigNumber;
    lpc(price: BigNumber): BigNumber;
    serialize(): object;
}
export declare const marketTypes: readonly ["crypto", "forex", "commodity", "index"];
export declare const InstrumentSchema: z.ZodObject<{
    version: z.ZodNumber;
    pair: z.ZodString;
    isDefaultVisible: z.ZodDefault<z.ZodOptional<z.ZodBoolean>>;
    displayRank: z.ZodOptional<z.ZodNumber>;
    description: z.ZodOptional<z.ZodString>;
    unitName: z.ZodString;
    unitNameShort: z.ZodOptional<z.ZodString>;
    defaultLotSize: z.ZodDefault<z.ZodOptional<z.ZodNumber>>;
    marketType: z.ZodEnum<["crypto", "forex", "commodity", "index"]>;
    baseSymbolLogoUri: z.ZodOptional<z.ZodString>;
    quoteSymbolLogoUri: z.ZodOptional<z.ZodString>;
    hideQuoteSymbolLogo: z.ZodDefault<z.ZodOptional<z.ZodBoolean>>;
    hideQuoteSymbol: z.ZodDefault<z.ZodOptional<z.ZodBoolean>>;
    displayDecimals: z.ZodDefault<z.ZodOptional<z.ZodNumber>>;
    shouldUseExtendedDecimals: z.ZodOptional<z.ZodBoolean>;
    chartSymbol: z.ZodOptional<z.ZodString>;
}, "strip", z.ZodTypeAny, {
    version: number;
    pair: string;
    isDefaultVisible: boolean;
    unitName: string;
    defaultLotSize: number;
    marketType: "forex" | "index" | "crypto" | "commodity";
    hideQuoteSymbolLogo: boolean;
    hideQuoteSymbol: boolean;
    displayDecimals: number;
    displayRank?: number | undefined;
    description?: string | undefined;
    unitNameShort?: string | undefined;
    baseSymbolLogoUri?: string | undefined;
    quoteSymbolLogoUri?: string | undefined;
    shouldUseExtendedDecimals?: boolean | undefined;
    chartSymbol?: string | undefined;
}, {
    version: number;
    pair: string;
    unitName: string;
    marketType: "forex" | "index" | "crypto" | "commodity";
    isDefaultVisible?: boolean | undefined;
    displayRank?: number | undefined;
    description?: string | undefined;
    unitNameShort?: string | undefined;
    defaultLotSize?: number | undefined;
    baseSymbolLogoUri?: string | undefined;
    quoteSymbolLogoUri?: string | undefined;
    hideQuoteSymbolLogo?: boolean | undefined;
    hideQuoteSymbol?: boolean | undefined;
    displayDecimals?: number | undefined;
    shouldUseExtendedDecimals?: boolean | undefined;
    chartSymbol?: string | undefined;
}>;
export type InstrumentSchema = z.infer<typeof InstrumentSchema>;
export type Instrument = InstrumentSchema & {
    getDisplayDecimals: (useExtendedDecimals?: boolean) => number;
    getDescription: () => string;
    getUnitName: (useShortUnitName?: boolean) => string;
    getHideQuoteSymbolLogo: () => boolean;
    getChartSymbol: () => string;
};
