import { BigNumber, constants } from "ethers";
import { isSamePair, pairFromString, pairToString } from "../../utils/general";
import { PRICE_DECIMALS, AMOUNT_DECIMALS, PRICE_UNIT, AMOUNT_UNIT, formatAmount, parseAmount, } from "./reader";
import { parseUnits } from "ethers/lib/utils";
import { z } from "zod";
import { CURRENT_INSTRUMENT_SCHEMA_VERSION } from "../../config";
/// A trade pair ID, which factors in a specific trade liquidity pool.
export class TradePairId {
    constructor(pair, lpId) {
        this.pair = pair;
        this.lpId = lpId;
    }
    eq(other) {
        return isSamePair(this.pair, other.pair) && this.lpId === other.lpId;
    }
    toString() {
        return `${pairToString(this.pair)}_${this.lpId}`;
    }
    static fromString(s) {
        const [pair, lpId] = s.split("_");
        return new TradePairId(pairFromString(pair), lpId);
    }
    static fromState(state) {
        return new TradePairId(state.pair, state.lpId);
    }
}
/**
 * This is analogous to the `PairConfig` type in the
 * server.
 */
export class TradePair {
    constructor(id, initialMarginFraction, maintenanceMarginFraction, incrementalInitialMarginFraction, baselinePositionSize, incrementalPositionSize, marginFeeFraction, spreadFraction, isActive, isReduceOnly, maxOpenInterestDiff, maxOpenInterestLong, maxOpenInterestShort, borrowFeeFactor, fundingFactor, fundingExponent, h2soPairData, usePriceImpact, priceImpactFraction, skewScale) {
        this.id = id;
        this.initialMarginFraction = initialMarginFraction;
        this.maintenanceMarginFraction = maintenanceMarginFraction;
        this.incrementalInitialMarginFraction = incrementalInitialMarginFraction;
        this.baselinePositionSize = baselinePositionSize;
        this.incrementalPositionSize = incrementalPositionSize;
        this.marginFeeFraction = marginFeeFraction;
        this.spreadFraction = spreadFraction;
        this.isActive = isActive;
        this.isReduceOnly = isReduceOnly;
        this.maxOpenInterestDiff = maxOpenInterestDiff;
        this.maxOpenInterestLong = maxOpenInterestLong;
        this.maxOpenInterestShort = maxOpenInterestShort;
        this.borrowFeeFactor = borrowFeeFactor;
        this.fundingFactor = fundingFactor;
        this.fundingExponent = fundingExponent;
        this.h2soPairData = h2soPairData;
        this.usePriceImpact = usePriceImpact;
        this.priceImpactFraction = priceImpactFraction;
        this.skewScale = skewScale;
        this.pair = id.pair;
    }
    getLargerSideFundingRate(openInterest) {
        const totalOpenInterest = openInterest.total();
        if (totalOpenInterest.isZero()) {
            return constants.Zero;
        }
        const imbalance = openInterest.long.sub(openInterest.short).abs();
        const imbalanceExp = Math.pow(+formatAmount(imbalance), +formatAmount(this.fundingExponent));
        return this.fundingFactor
            .mul(parseAmount(imbalanceExp.toString()))
            .div(totalOpenInterest);
    }
    shouldUsePriceImpact() {
        return !!(this.usePriceImpact &&
            this.priceImpactFraction &&
            this.skewScale);
    }
    /**
     * @returns the funding rates with `AMOUNT_DECIMALS` precision.
     */
    getFundingRate(openInterest) {
        const largerSideRate = this.getLargerSideFundingRate(openInterest);
        return {
            long: openInterest.long.gt(openInterest.short)
                ? largerSideRate
                : largerSideRate.mul("-1"),
            short: openInterest.short.gt(openInterest.long)
                ? largerSideRate
                : largerSideRate.mul("-1"),
        };
    }
    getLongBorrowRate(openInterest) {
        if (this.maxOpenInterestLong === null)
            return constants.Zero;
        return openInterest.long
            .mul(this.borrowFeeFactor)
            .div(this.maxOpenInterestLong);
    }
    getShortBorrowRate(openInterest) {
        if (this.maxOpenInterestShort === null)
            return constants.Zero;
        return openInterest.short
            .mul(this.borrowFeeFactor)
            .div(this.maxOpenInterestShort);
    }
    /**
     * @returns the borrow rates with `AMOUNT_DECIMALS` precision.
     */
    getBorrowRate(openInterest) {
        return {
            long: this.getLongBorrowRate(openInterest),
            short: this.getShortBorrowRate(openInterest),
        };
    }
    getInitialMargin(size, price) {
        size = size.abs();
        if (size.lt(this.baselinePositionSize) ||
            this.baselinePositionSize.eq("0")) {
            return size
                .mul(price)
                .mul(this.initialMarginFraction)
                .div(TradePair.divisor);
        }
        const increments = size
            .sub(this.baselinePositionSize)
            .div(this.incrementalPositionSize);
        const fraction = this.initialMarginFraction.add(increments.mul(this.incrementalInitialMarginFraction));
        return size.mul(price).mul(fraction).div(TradePair.divisor);
    }
    getMaintenanceMargin(size, price) {
        return size
            .mul(price)
            .mul(this.maintenanceMarginFraction)
            .div(TradePair.divisor)
            .abs();
    }
    getMarginFee(size, price) {
        return size
            .mul(price)
            .mul(this.marginFeeFraction)
            .div(TradePair.divisor)
            .abs();
    }
}
TradePair.divisor = parseUnits("1", AMOUNT_DECIMALS + PRICE_DECIMALS);
// The bid price is almost always lower than the ask price.
// Typically when the bid and ask prices meet, a trade occurs.
export class MarketPrice {
    constructor(index, bestBid, bestAsk, skew) {
        this.index = index;
        this.bestBid = bestBid;
        this.bestAsk = bestAsk;
        this.skew = skew;
    }
    /**
     * Calculates the inverse of this market price.
     * @note This does not modify the current object.
     * @returns The inverse of this market price.
     */
    inverse() {
        return new MarketPrice(parseUnits("1", PRICE_DECIMALS * 2).div(this.index), parseUnits("1", PRICE_DECIMALS * 2).div(this.bestAsk), parseUnits("1", PRICE_DECIMALS * 2).div(this.bestBid));
    }
    get marketPrice() {
        return this.skew ?? this.index;
    }
    static fromIndex(index, spread = constants.Zero) {
        return new MarketPrice(index, index.mul(AMOUNT_UNIT.sub(spread)).div(AMOUNT_UNIT), index.mul(AMOUNT_UNIT.add(spread)).div(AMOUNT_UNIT));
    }
    static zero() {
        return new MarketPrice(BigNumber.from("0"), BigNumber.from("0"), BigNumber.from("0"));
    }
    getTradePrice(isBuying) {
        return isBuying ? this.bestAsk : this.bestBid;
    }
    getTradePriceFromTradeSize(tradeSize) {
        return this.getTradePrice(tradeSize.gt(0));
    }
}
export class TradeAction {
    constructor(pairId, size, price, timestamp, realisedEquity, marginFee, type, txHash, didOpenPosition, didClosePosition) {
        this.pairId = pairId;
        this.size = size;
        this.price = price;
        this.timestamp = timestamp;
        this.realisedEquity = realisedEquity;
        this.marginFee = marginFee;
        this.type = type;
        this.txHash = txHash;
        this.didOpenPosition = didOpenPosition;
        this.didClosePosition = didClosePosition;
    }
    eq(other) {
        return (this.pairId.eq(other.pairId) &&
            this.size.eq(other.size) &&
            this.price.eq(other.price) &&
            this.timestamp === other.timestamp &&
            this.realisedEquity.eq(other.realisedEquity) &&
            this.marginFee.eq(other.marginFee) &&
            this.type === other.type &&
            this.txHash.toLowerCase() === other.txHash.toLowerCase() &&
            this.didOpenPosition === other.didOpenPosition &&
            this.didClosePosition === other.didClosePosition);
    }
    isLiquidation() {
        return this.type === "liquidation";
    }
}
export class OpenInterest {
    constructor(long, short) {
        this.long = long;
        this.short = short;
    }
    total() {
        return this.long.add(this.short);
    }
    static zero() {
        return new OpenInterest(constants.Zero, constants.Zero);
    }
}
export class TradeSize {
    constructor(amount, type) {
        this.amount = amount;
        this.type = type;
    }
    static fromLot(amount) {
        return new TradeSize(amount, "Lot");
    }
    static fromLpc(amount) {
        return new TradeSize(amount, "Lpc");
    }
    /// Returns this trade size as lots.
    lots(price) {
        if (this.type == "Lot") {
            return this.amount;
        }
        return this.amount.mul(PRICE_UNIT).div(price);
    }
    /// Returns this trade size as LPC.
    lpc(price) {
        if (this.type == "Lpc") {
            return this.amount;
        }
        return this.amount.mul(price).div(PRICE_UNIT);
    }
    serialize() {
        return {
            [this.type]: formatAmount(this.amount),
        };
    }
}
export const marketTypes = ["crypto", "forex", "commodity", "index"];
export const InstrumentSchema = z.object({
    // Version number of the schema.
    version: z
        .number()
        .int()
        .gte(CURRENT_INSTRUMENT_SCHEMA_VERSION)
        .lt(CURRENT_INSTRUMENT_SCHEMA_VERSION + 1),
    // The instrument pair in the format of base/quote symbols.
    pair: z.string(),
    // May be used to indicate that the instrument is a default for a new user.
    isDefaultVisible: z.optional(z.boolean()).default(false),
    // Sequence for ordering.
    displayRank: z.optional(z.number()),
    // Description in markets and chart symbol info, e.g. CBOE Volatility Index.
    // Should be set to unitName if not supplied.
    description: z.optional(z.string()),
    // Unit name of a single lot of an instrument, e.g. ethereum, barrel of oil, euro, ounce of gold.
    unitName: z.string(),
    // Abbreviated lot unit for use when real-estate is limited.
    // E.g. "australian dollar" would be abbreviated to "au dollar".
    // Should be defaulted to unitName if not supplied.
    unitNameShort: z.optional(z.string()),
    // Basic unit of lots - typically 1 unit of unitName.
    // It could be 1000 if the unitName of a commodity was in grams so that the default lot would be priced at 1kg lots.
    defaultLotSize: z.optional(z.number()).default(1),
    // Market category/type.
    // Subset of TV SymbolType.
    marketType: z.enum(marketTypes),
    // The URI of the logo for the base symbol.
    // Should default to a default image, value or rendered icon,
    // e.g. a disc with the baseSymbol text if not supplied.
    baseSymbolLogoUri: z.optional(z.string()),
    // The URI of the logo for the quote symbol.
    // Should not be displayed if not supplied or hideQuoteSymbolLogo === true.
    quoteSymbolLogoUri: z.optional(z.string()),
    hideQuoteSymbolLogo: z.optional(z.boolean()).default(false),
    // Whether to hide the quote symbol, i.e. show XYZ instead of XYZ/USD
    hideQuoteSymbol: z.optional(z.boolean()).default(false),
    displayDecimals: z.optional(z.number()).default(4),
    // Should be used to determine whether add extra decimal place
    // when showing prices that are current/changing (as a superscript for example)
    // and as an additional decimal place in charts.
    shouldUseExtendedDecimals: z.optional(z.boolean()),
    // Symbol to use for the chart API. Should default to the pair if not set.
    chartSymbol: z.optional(z.string()),
});
