import { BigNumber } from "ethers";
import { OpenInterest } from "./interface";
import { AMOUNT_UNIT, parseAmount, PRICE_UNIT } from "./reader";
import { getCurrentUnixTimestamp } from "../../utils/general";
export class PairState {
    constructor(tradePair, sumFractionFunding, sumFractionBorrow, openInterest) {
        this.tradePair = tradePair;
        this.sumFractionFunding = sumFractionFunding;
        this.sumFractionBorrow = sumFractionBorrow;
        this.openInterest = openInterest;
    }
    static fromSerialized(serialized, tradePair) {
        const sumFractionFunding = {
            value: {
                long: parseAmount(serialized.sumFractionFunding.value.long),
                short: parseAmount(serialized.sumFractionFunding.value.short),
            },
            timestamp: serialized.sumFractionFunding.timestamp,
        };
        const sumFractionBorrow = {
            value: {
                long: parseAmount(serialized.sumFractionBorrow.value.long),
                short: parseAmount(serialized.sumFractionBorrow.value.short),
            },
            timestamp: serialized.sumFractionBorrow.timestamp,
        };
        const openInterest = new OpenInterest(parseAmount(serialized.openInterest.long), parseAmount(serialized.openInterest.short));
        return new PairState(tradePair, sumFractionFunding, sumFractionBorrow, openInterest);
    }
    static zero(tradePair) {
        const zeroTimestampedBigNumberMarketSide = {
            value: {
                long: BigNumber.from(0),
                short: BigNumber.from(0),
            },
            timestamp: 0,
        };
        return new PairState(tradePair, zeroTimestampedBigNumberMarketSide, zeroTimestampedBigNumberMarketSide, OpenInterest.zero());
    }
    getCurrentSumFractionFunding(indexPrice) {
        const currentTimestamp = getCurrentUnixTimestamp();
        const fundingRates = this.tradePair.getFundingRate(this.openInterest);
        return {
            long: calculateCurrentSumFractionNotional(indexPrice, fundingRates.long, this.sumFractionFunding.value.long, this.sumFractionFunding.timestamp, currentTimestamp),
            short: calculateCurrentSumFractionNotional(indexPrice, fundingRates.short, this.sumFractionFunding.value.short, this.sumFractionFunding.timestamp, currentTimestamp),
        };
    }
    getLpFundingRate(indexPrice) {
        if (this.openInterest.total().isZero()) {
            return BigNumber.from(0);
        }
        const fundingRates = this.tradePair.getFundingRate(this.openInterest);
        return getSmallerMarketSide(fundingRates).mul(indexPrice).div(PRICE_UNIT);
    }
    getCurrentSumFractionBorrow(indexPrice) {
        const currentTimestamp = getCurrentUnixTimestamp();
        const borrowRates = this.tradePair.getBorrowRate(this.openInterest);
        return {
            long: calculateCurrentSumFractionNotional(indexPrice, borrowRates.long, this.sumFractionBorrow.value.long, this.sumFractionBorrow.timestamp, currentTimestamp),
            short: calculateCurrentSumFractionNotional(indexPrice, borrowRates.short, this.sumFractionBorrow.value.short, this.sumFractionBorrow.timestamp, currentTimestamp),
        };
    }
}
const calculateCurrentSumFraction = (rate, snapshotValue, snapshotTimestamp, currentTimestamp) => {
    const hoursElapsed = calculateHoursElapsed(snapshotTimestamp, currentTimestamp);
    const accrued = rate.mul(hoursElapsed).div(AMOUNT_UNIT);
    return snapshotValue.add(accrued);
};
const calculateCurrentSumFractionNotional = (indexPrice, rate, snapshotValue, snapshotTimestamp, currentTimestamp) => calculateCurrentSumFraction(rate.mul(indexPrice).div(PRICE_UNIT), snapshotValue, snapshotTimestamp, currentTimestamp);
const calculateHoursElapsed = (fromUnixTimestamp, toUnixTimestamp) => parseAmount((toUnixTimestamp - fromUnixTimestamp).toString()).div(60 * 60);
const getSmallerMarketSide = (marketSide) => {
    return marketSide.short.lt(marketSide.long)
        ? marketSide.short
        : marketSide.long;
};
