import { ethers } from "ethers";
import { BASIS_POINTS_DIVISOR } from "../../../../../../constants";
import { getLatestMarketPrice } from "../prices";
import { getActualPositionHlp } from "./parseContractPosition";
import { getReversedPrice, isSamePair } from "../../../../../../utils/general";
/**
 * Gets the delta for a position given the current price of the index token.
 * @param position the position from which to calculate the delta
 * @param minProfitTime the number of seconds until the position can take profit.
 * @param minProfitBasisPoints the profit percentage a position has to reach to
 * take profit before {minProfitTime}
 * @returns the current delta, and delta percentage, and pending delta of
 * the position as well as whether the position has profit. Pending delta
 * refers to the delta of the position before the minimum profit time has
 * passed, or the minimum profit delta has been exceeded.
 */
export const getPositionDelta = (position, minProfitTime, minProfitBasisPoints, priceOverride) => {
    const actualPosition = getActualPositionHlp(position);
    const isPairReversed = !isSamePair(position.pair, actualPosition.pair);
    const indexPrice = priceOverride || getLatestMarketPrice(actualPosition.pair).index;
    const actualIndexPrice = isPairReversed
        ? getReversedPrice(indexPrice)
        : indexPrice;
    return getPositionDeltaForActualPosition(actualIndexPrice, actualPosition, minProfitTime, minProfitBasisPoints);
};
export const getPositionDeltaForActualPosition = (indexPrice, position, minProfitTime, minProfitBasisPoints) => {
    const { size, collateral, isLong, averagePrice, lastIncreasedTime } = position;
    const priceDelta = averagePrice.gt(indexPrice)
        ? averagePrice.sub(indexPrice)
        : indexPrice.sub(averagePrice);
    let delta = size.mul(priceDelta).div(averagePrice);
    const pendingDelta = delta;
    const minProfitExpired = lastIncreasedTime
        .add(minProfitTime)
        .lt(Math.floor(Date.now() / 1000));
    const hasProfit = isLong
        ? indexPrice.gt(averagePrice)
        : indexPrice.lt(averagePrice);
    const isDeltaTooLow = delta
        .mul(BASIS_POINTS_DIVISOR)
        .lte(size.mul(minProfitBasisPoints));
    if (!minProfitExpired && hasProfit && isDeltaTooLow) {
        delta = ethers.constants.Zero;
    }
    const deltaPercentage = delta.mul(BASIS_POINTS_DIVISOR).div(collateral);
    const pendingDeltaPercentage = pendingDelta
        .mul(BASIS_POINTS_DIVISOR)
        .div(collateral);
    return {
        delta,
        pendingDelta,
        pendingDeltaPercentage,
        hasProfit,
        deltaPercentage,
    };
};
