import { BigNumber } from "ethers";
import { DATA_FEED_PRICE_DECIMALS } from "../../../../../config";
import { BASIS_POINTS_DIVISOR, H2SO_PRICE_DECIMALS, } from "../../../../../constants";
import { config, h2so } from "../../../../../index";
import { cachedArbitrumPegs } from "../../../../../utils/convert";
import { allSettledResults, getReversedMarketPriceLegacy, getReversedPair, isSamePair, pairToString, transformDecimals, } from "../../../../../utils/general";
import { fetchTimedEncodedSignedQuotes } from "../../../../h2so";
import { fetchUnsignedQuote, } from "../../../../h2so/fetcher";
import { fromParsedDataFeedPair, toParsedDatafeedPair, } from "../../../../h2so/toParsedDatafeedPair";
import { HandleTokenManagerInstance } from "../../../../token-manager/HandleTokenManager";
import { PRICE_DECIMALS, } from "../../legacyInterface";
import { DEFAULT_HLP_NETWORK, getActualHlpPairIfReversed, getReversedHlpPairIfApplicable, isReversedPair, shouldHlpPairBeReversed, } from "../config";
import { getTokenSpreadBasisPoints } from "./getTokenSpreadBasisPoints";
import { getTradePairHlpFromPair, getTradePairsHlp, } from "./tokens";
/**
 * Gets a signed H2SO quote for all the input pairs and calculates the
 * price spread for each pair.
 * The prices follow the pair, returning the price of the base symbol
 * quoted against the quote symbol.
 * One of the symbols in each pair must be USD.
 * @example const [{ encoded }, aud] = await getQuotedPrices(pairFromString("AUD/USD"));
 */
export const getSignedPrices = async (pairs, network = DEFAULT_HLP_NETWORK) => {
    const provider = config.providers[network];
    // Map pairs to "actual" pairs, in case any of the inputs are reversed.
    const usdQuotedPairs = pairs.map(getActualHlpPairIfReversed);
    // Fetch H2SO quote, waiting (if needed) for the quote to become valid.
    const quote = await fetchTimedEncodedSignedQuotes(usdQuotedPairs, provider);
    // Calculate the spread for each token.
    const baseTokenAddresses = usdQuotedPairs.map((pair) => HandleTokenManagerInstance.getTokenBySymbol(fromParsedDataFeedPair(pair).baseSymbol, network).address);
    const spreads = await Promise.all(baseTokenAddresses.map((token) => getTokenSpreadBasisPoints({
        tokenAddress: token,
    })));
    const spreadPrices = quote.decoded.map(({ signatureParams: { value } }, i) => {
        const price = transformDecimals(value, H2SO_PRICE_DECIMALS, PRICE_DECIMALS);
        let marketPrice = {
            index: price,
            bestBid: getSpreadPrice(value, spreads[i], false),
            bestAsk: getSpreadPrice(value, spreads[i], true),
        };
        if (pairs[i].baseSymbol === "USD")
            marketPrice = getReversedMarketPriceLegacy(marketPrice);
        return marketPrice;
    });
    return [quote, ...spreadPrices];
};
/**
 * Gets (unsigned) H2SO prices for all the input pairs and calculates the
 * price spread for each pair, transforming prices to 30 decimals.
 * For hLP, all pairs must be quoted by USD.
 * If a price is not found, it is not included in the returned map.
 * @example const [aud] = await getUnsignedPrices(pairFromString("AUD/USD"));
 * @returns Mapping from stringified pair to the market price.
 */
export const getUnsignedPrices = async (pairs, network = DEFAULT_HLP_NETWORK) => allSettledResults(pairs.map(async (pair) => fetchUnsignedMarketPriceForPair(pair, network))).then(mapPriceArrayToDictionary);
export const fetchUnsignedMarketPriceForPair = async (pair, network = DEFAULT_HLP_NETWORK) => {
    const actualPair = getActualHlpPairIfReversed(pair);
    const tokenAddress = HandleTokenManagerInstance.getTokenBySymbol(actualPair.baseSymbol, network).address;
    const [price, spreadBasisPoints] = await Promise.all([
        await fetchUnsignedQuote(actualPair),
        getTokenSpreadBasisPoints({ tokenAddress }),
    ]);
    const priceBn = transformDecimals(BigNumber.from(price), H2SO_PRICE_DECIMALS, PRICE_DECIMALS);
    let marketPrice = {
        index: priceBn,
        bestAsk: getSpreadPrice(BigNumber.from(price), spreadBasisPoints, true),
        bestBid: getSpreadPrice(BigNumber.from(price), spreadBasisPoints, false),
    };
    if (!isSamePair(actualPair, pair))
        marketPrice = getReversedMarketPriceLegacy(marketPrice);
    return {
        pair: pairToString(pair),
        marketPrice,
    };
};
const mapPriceArrayToDictionary = (prices) => prices.reduce((object, price) => ({
    ...object,
    [price.pair]: price.marketPrice,
}), {});
export const applySpread = (pair, oraclePrice) => {
    // The maximum spread is used for both as hLP spread is symmetric.
    const { spreadBps: { maximum: spread }, } = pair;
    return {
        maximum: getSpreadPrice(oraclePrice, spread, true),
        minimum: getSpreadPrice(oraclePrice, spread, false),
    };
};
const getSpreadPrice = (oraclePrice, spread, maximise) => {
    const price = transformDecimals(oraclePrice, DATA_FEED_PRICE_DECIMALS, PRICE_DECIMALS);
    const multiplier = getSpreadMultiplierBps(spread, maximise);
    return price.mul(multiplier).div(BASIS_POINTS_DIVISOR);
};
export const getSpreadMultiplierBps = (spreadBps, maximise) => maximise
    ? spreadBps.add(BASIS_POINTS_DIVISOR)
    : spreadBps.mul(-1).add(BASIS_POINTS_DIVISOR);
const symbolToAddressCache = {};
// Subscriptions from this module to the h2so feed.
const inSubscriptions = {};
// Subscriptions from trade users to this module's market prices.
const outSubscriptions = [];
export const subscribeToMarketPrice = ({ callback, pair, }) => {
    if (!inSubscriptions[pairToString(pair)])
        subscribeToH2soFeed(pair);
    return (outSubscriptions.push({
        callback,
        arg: pair,
    }) - 1);
};
export const unsubscribeFromMarketPrice = (id) => {
    if (!outSubscriptions[id])
        throw new Error("hlp prices: not subscribed");
    outSubscriptions[id] = null;
};
const subscribeToH2soFeed = (pair) => {
    // Get the actual pair, in case the input is reversed (e.g. USD/JPY).
    pair = getActualHlpPairIfReversed(pair);
    const response = h2so.feed.subscribe([pair].map(toParsedDatafeedPair), onH2soFeedPriceUpdate);
    inSubscriptions[pairToString(pair)] = response.subscriberId;
};
const onH2soFeedPriceUpdate = async (dataFeedPair, price) => {
    const { pair, marketPrice } = getMarketPriceFromH2soIndex(dataFeedPair, price);
    for (let subscription of outSubscriptions) {
        if (!subscription || !isSamePair(subscription.arg, pair))
            continue;
        subscription.callback(pair, marketPrice);
    }
};
export const getLatestMarketPrice = (pair) => {
    const peggedPair = getPeggedPair(pair);
    if (peggedPair) {
        pair = peggedPair;
    }
    const dataFeedPair = toParsedDatafeedPair(getActualHlpPairIfReversed(pair));
    const indexPrice = h2so.feed.getLatestPrice(dataFeedPair);
    const { marketPrice } = getMarketPriceFromH2soIndex(dataFeedPair, indexPrice);
    return marketPrice;
};
const getPeggedPair = (pair) => {
    // store whether original pair is reversed
    const isReversed = pair.baseSymbol === "USD";
    const trueBaseSymbol = isReversed ? pair.quoteSymbol : pair.baseSymbol;
    const peggedToken = HandleTokenManagerInstance.getTokenBySymbol(trueBaseSymbol, DEFAULT_HLP_NETWORK);
    // get the hlp token for the pegged pair, if it exists
    const pegs = cachedArbitrumPegs.get();
    const hlpTokenAddress = pegs.find((peg) => peg.peggedToken.toLowerCase() === peggedToken.address.toLowerCase())?.fxToken;
    if (!hlpTokenAddress) {
        return undefined;
    }
    // get the trade pair for the hlp token
    const tradePairsHlp = getTradePairsHlp();
    const tradePair = tradePairsHlp.find((tradePair) => tradePair.indexAddress.toLowerCase() === hlpTokenAddress.toLowerCase());
    if (!tradePair) {
        return undefined;
    }
    // make sure the reversal for the pegged pair is the same as the original pair
    const shouldReverse = isReversed != isReversedPair(tradePair.pair);
    return shouldReverse ? getReversedPair(tradePair.pair) : tradePair.pair;
};
/**
 * Maps an h2so datafeed pair and index price to an hLP pair
 * and MarketPrice object.
 */
const getMarketPriceFromH2soIndex = (dataFeedPair, indexPrice, reversePriceIfApplicable = true) => {
    // The actual hLP pair, i.e. must not be reversed, even if the pair
    // is configured to be when facing the user.
    const pair = fromParsedDataFeedPair(dataFeedPair);
    if (!symbolToAddressCache[pair.baseSymbol])
        symbolToAddressCache[pair.baseSymbol] =
            HandleTokenManagerInstance.getTokenBySymbol(pair.baseSymbol, DEFAULT_HLP_NETWORK).address;
    // Potentially reverse the pair, e.g. from fxJPY/USD to USD/fxJPY.
    const pairParsed = getReversedHlpPairIfApplicable(pair);
    const tradePair = getTradePairHlpFromPair(pairParsed);
    const { maximum, minimum } = applySpread(tradePair, indexPrice);
    let marketPrice = {
        index: transformDecimals(indexPrice, H2SO_PRICE_DECIMALS, PRICE_DECIMALS),
        bestBid: minimum,
        bestAsk: maximum,
    };
    if (reversePriceIfApplicable && shouldHlpPairBeReversed(pair))
        marketPrice = getReversedMarketPriceLegacy(marketPrice);
    return { marketPrice, pair: pairParsed };
};
export const getHlpMarketPriceFromIndex = (pair, indexPrice) => {
    const dataFeedPair = toParsedDatafeedPair(getActualHlpPairIfReversed(pair));
    const { marketPrice } = getMarketPriceFromH2soIndex(dataFeedPair, transformDecimals(indexPrice, PRICE_DECIMALS, H2SO_PRICE_DECIMALS), false);
    return marketPrice;
};
