import { BigNumber } from "ethers";
import { gql, request } from "graphql-request";
import { config } from "../../../../..";
import { FIVE_MINUTES_MILLIS } from "../../../../../constants";
import { CachedObject } from "../../../../../utils/cachedObject";
import { cachedArbitrumPegs, getTokenPegs } from "../../../../../utils/convert";
import { getReversedPair, isSamePair, pairFromString, pairToBaseTokenAddress, pairToQuoteTokenAddress, pairToString, transformDecimals, } from "../../../../../utils/general";
import { fetchCacheApi, getCacheServerErrorMessage, } from "../../../../../utils/sdk";
import { isHlpMarketClosed } from "../../../../../utils/trade";
import { getNativeTokenInfo } from "../../../../token-manager";
import { HandleTokenManagerInstance } from "../../../../token-manager/HandleTokenManager";
import { PRICE_DECIMALS, } from "../../legacyInterface";
import { isEtherAddress } from "../../../utils";
import { DEFAULT_HLP_NETWORK, HLP_PLATFORM_NAME, fetch, getActualHlpPairIfReversed, getHlpContracts, isReversedPair, shouldHlpPairBeReversed, } from "../config";
import { getAllTokenFundingRates } from "./fundingRate";
import { getAllTokenSpreadBasisPoints } from "./getTokenSpreadBasisPoints";
import { fetchUnsignedMarketPriceForPair } from "./prices";
const UNAVAILABLE_REASON_WHITELIST = "disabled";
const UNAVAILABLE_REASON_MARKET_CLOSED = "weekend trading";
const UNAVAILABLE_REASON_MARKET_PRICE_FEED = "unavailable";
var PriceAvailabilityStatus;
(function (PriceAvailabilityStatus) {
    PriceAvailabilityStatus[PriceAvailabilityStatus["Unknown"] = 0] = "Unknown";
    PriceAvailabilityStatus[PriceAvailabilityStatus["Available"] = 1] = "Available";
    PriceAvailabilityStatus[PriceAvailabilityStatus["Unavailable"] = 2] = "Unavailable";
})(PriceAvailabilityStatus || (PriceAvailabilityStatus = {}));
const cachedTokens = new CachedObject(FIVE_MINUTES_MILLIS);
const cachedTradePairs = new CachedObject(FIVE_MINUTES_MILLIS);
export const fetchTokens = async () => cachedTokens.fetch(fetchTokensUncached);
/// Gets cached tokens; throws if cache is empty.
export const getTokens = () => cachedTokens.get();
const fetchTokensUncached = async () => {
    if (config.sdk.shouldUseCacheServer) {
        const tokens = await fetchTradePairsHlpFromServer();
        return tokens
            .map((t) => t.internals.hlpToken)
            .filter((token) => token.isWhitelisted);
    }
    const indexed = await fetchIndexedTokenConfigs();
    const liquidity = await fetchLiquidityInfo(indexed.map((i) => i.token));
    return indexed.map((raw) => ({
        address: raw.token,
        isWhitelisted: raw.isWhitelisted,
        tokenDecimals: +raw.tokenDecimals,
        tokenWeight: +raw.tokenWeight,
        minProfitBasisPoints: +raw.minProfitBasisPoints,
        maxUsdHlpAmount: BigNumber.from(raw.maxUsdgAmount),
        isStable: raw.isStable,
        isShortable: raw.isShortable,
        poolAmount: liquidity.get(raw.token.toLowerCase()).poolAmount,
        reservedAmount: liquidity.get(raw.token.toLowerCase()).reservedAmount,
    }));
};
const fetchIndexedTokenConfigs = async () => {
    const response = await request(config.theGraphEndpoints.arbitrum.trade, gql `
      query {
        tokenConfigs(first: 1000, where: { isWhitelisted: true }) {
          token
          isWhitelisted
          tokenDecimals
          tokenWeight
          minProfitBasisPoints
          maxUsdgAmount
          isStable
          isShortable
        }
      }
    `);
    if (!Array.isArray(response.tokenConfigs))
        throw new Error("Response is not an array");
    return response.tokenConfigs;
};
export const hlpTokenToUsdPair = ({ address }) => {
    const { symbol } = HandleTokenManagerInstance.getTokenByAddress(address, DEFAULT_HLP_NETWORK);
    return pairFromString(`${symbol}/USD`);
};
export const fetchTradePairsHlp = async () => cachedTradePairs
    .fetch(fetchTradePairsHlpUncached)
    .then(filterTradableTradePairsHlp);
const fetchTradePairsHlpFromServer = async () => fetchCacheApi("trade-pairs-hlp");
export const fetchStablePairsHlp = async () => cachedTradePairs
    .fetch(fetchTradePairsHlpUncached)
    .then(filterStableTradePairsHlp);
/**
 * Fetches a stable pair from the hLP.
 * If there is more than one, this will return the first stable available.
 */
export const fetchStablePairHlp = async () => {
    const pair = (await fetchStablePairsHlp())?.[0];
    if (!pair)
        throw new Error("Stable pair not found for hLP");
    return pair;
};
/// Gets cached tokens; throws if cache is empty.
export const getTradePairsHlp = () => cachedTradePairs.get();
/// Gets trade pair from cache; throws if not found.
export const getTradePairHlpFromPair = (pair) => {
    const cache = cachedTradePairs.get();
    const tradePair = cache.find((p) => isSamePair(p.pair, pair));
    if (!tradePair)
        throw new Error(`getTradePairHlpFromPair: not found: ${pairToString(pair)}`);
    return tradePair;
};
/**
 * Filters hLP pairs which are actually tradable, (e.g. excluding "stable"
 * tokens hLP tokens such as fxUSD).
 */
const filterTradableTradePairsHlp = (pairs) => pairs.filter((pair) => !pair.internals.hlpToken.isStable && pair.pair.baseSymbol !== "WETH");
/**
 * Filters hLP pairs which are quoted on USD, and "stable" tokens,
 * e.g. fxUSD/USD, and not actually tradable.
 */
const filterStableTradePairsHlp = (pairs) => pairs.filter((pair) => pair.internals.hlpToken.isStable);
/**
 * Returns all hLP pairs, even if they are not tradable (e.g. including
 * stable hLP tokens such as fxUSD).
 */
const fetchTradePairsHlpUncachedUnfiltered = async () => {
    const tokens = await fetchTokens();
    const spreads = await getAllTokenSpreadBasisPoints(DEFAULT_HLP_NETWORK, false);
    const fundingRates = await getAllTokenFundingRates();
    const hlpConfig = await fetch(DEFAULT_HLP_NETWORK);
    return tokens.map((token) => mapHlpTokenToTradePairHlp(token, spreads[token.address], fundingRates[token.address], hlpConfig));
};
export const fetchTradePairsHlpUncached = async () => {
    if (config.sdk.shouldUseCacheServer) {
        try {
            return await fetchTradePairsHlpFromServer();
        }
        catch (error) {
            console.error(getCacheServerErrorMessage(), error);
        }
    }
    return fetchTradePairsHlpUncachedUnfiltered().then(setInternalPriceAvailabilityStatus);
};
const mapHlpTokenToTradePairHlp = (hlpToken, spread, fundingRate, hlpConfig) => {
    const symbol = HandleTokenManagerInstance.getTokenByAddress(hlpToken.address, DEFAULT_HLP_NETWORK).symbol;
    if (!fundingRate)
        throw new Error("hlp getTradePairs: funding rate not found");
    if (!spread)
        throw new Error("hlp getTradePairs: spread not found");
    const internals = {
        hlpToken,
        priceAvailabilityStatus: PriceAvailabilityStatus.Unknown,
    };
    const pair = pairFromString(`${symbol}/USD`);
    return {
        pair: shouldHlpPairBeReversed(pair) ? getReversedPair(pair) : pair,
        spreadBps: {
            // Spread is symmetric for hLP.
            maximum: spread,
            minimum: spread,
        },
        indexAddress: hlpToken.address,
        internals,
        marginFeeBps: BigNumber.from(hlpConfig.marginFeeBasisPoints),
        platformName: HLP_PLATFORM_NAME,
        maxLeverageBps: BigNumber.from(hlpConfig.maxLeverage),
        liquidationFeeUsd: hlpConfig.liquidationFee,
    };
};
export const getAvailableLiquidity = async ({ pair, forceFetch = false, }) => {
    const hlpTokens = await cachedTradePairs.fetch(fetchTradePairsHlpUncached, forceFetch);
    const normalHlpToken = hlpTokens.find((t) => isSamePair(pair, t.pair));
    const stable = hlpTokens.find((t) => t.internals.hlpToken.isStable);
    if (!stable)
        throw new Error("stable token not found");
    if (!normalHlpToken)
        throw new Error("index token not found");
    const isReversed = isReversedPair(pair);
    // null checks safe here as error thrown if both don't exist
    const indexData = normalHlpToken.internals.hlpToken;
    const stableData = stable.internals.hlpToken;
    if (!isReversed) {
        return {
            long: indexData.poolAmount.sub(indexData.reservedAmount),
            short: transformDecimals(stableData.poolAmount.sub(stableData.reservedAmount), stableData.tokenDecimals, PRICE_DECIMALS),
        };
    }
    else {
        return {
            short: transformDecimals(indexData.poolAmount.sub(indexData.reservedAmount), indexData.tokenDecimals, PRICE_DECIMALS),
            long: stableData.poolAmount.sub(stableData.reservedAmount),
        };
    }
};
export const getAvailabilityHlp = (pair) => {
    const tradePair = getTradePairHlpFromPair(pair);
    const tokenSymbol = tradePair.pair.baseSymbol;
    if (!tradePair.internals.hlpToken.isWhitelisted) {
        return {
            isAvailable: false,
            reason: UNAVAILABLE_REASON_WHITELIST,
        };
    }
    if (isHlpMarketClosed(tokenSymbol)) {
        return {
            isAvailable: false,
            reason: UNAVAILABLE_REASON_MARKET_CLOSED,
        };
    }
    if (tradePair.internals.priceAvailabilityStatus ===
        PriceAvailabilityStatus.Unavailable) {
        return {
            isAvailable: false,
            reason: UNAVAILABLE_REASON_MARKET_PRICE_FEED,
        };
    }
    return {
        isAvailable: true,
    };
};
/***
 * Returns a TradePairsHlp array with available set to false if
 * the H2SO price is not available for the pair.
 */
const setInternalPriceAvailabilityStatus = async (pairs) => {
    return Promise.all(pairs.map(async (pair) => {
        // if already unavailable, return current reason
        if (pair.internals.priceAvailabilityStatus !==
            PriceAvailabilityStatus.Unknown) {
            return pair;
        }
        try {
            await fetchUnsignedMarketPriceForPair(pair.pair);
            pair.internals.priceAvailabilityStatus =
                PriceAvailabilityStatus.Available;
        }
        catch {
            pair.internals.priceAvailabilityStatus =
                PriceAvailabilityStatus.Unavailable;
        }
        return pair;
    }));
};
export const pairToTradePairHlp = async (pair) => {
    // Update cache if necessary.
    await fetchTradePairsHlp();
    return pairToTradePairHlpSync(pair);
};
export const pairToTradePairHlpSync = (pair) => {
    if (shouldHlpPairBeReversed(pair))
        pair = getReversedPair(pair);
    const pairs = getTradePairsHlp();
    const pairHlp = pairs.find((p) => isSamePair(p.pair, pair));
    if (!pairHlp)
        throw new Error(`pairToHlpTradePairSync: could not find ${pairToString(pair)}`);
    return pairHlp;
};
export const getHlpTokenFromPair = (pair) => {
    const tokens = getTokens();
    const isPairReversed = shouldHlpPairBeReversed(getActualHlpPairIfReversed(pair));
    const tokenAddress = isPairReversed
        ? pairToQuoteTokenAddress(pair)
        : pairToBaseTokenAddress(pair);
    const token = tokens.find((t) => t.address.toLowerCase() === tokenAddress.toLowerCase());
    if (!token)
        throw new Error(`getHlpTokenFromPosition: index token not found (${tokenAddress})`);
    return token;
};
/// Fetch all (tradable and stable) hLP tokens, then append pegged tokens if available.
export const getPlatformTokens = (includeEth) => {
    const tradeTokens = getTokensInfo(includeEth);
    if (!cachedArbitrumPegs.isAvailable) {
        return tradeTokens;
    }
    const peggedTokens = HandleTokenManagerInstance.getTokensByAddresses(cachedArbitrumPegs.get().map((peg) => {
        return {
            address: peg.peggedToken,
            network: DEFAULT_HLP_NETWORK,
        };
    }));
    return [...tradeTokens, ...peggedTokens];
};
export const getTokensInfo = (includeEth = true) => [
    HandleTokenManagerInstance.getHlpTokens(DEFAULT_HLP_NETWORK),
    // Also include ETH.
    includeEth ? [getNativeTokenInfo(DEFAULT_HLP_NETWORK)] : [],
].flat();
export const getCollateralTokens = (_) => {
    // trigger cache fetch
    getTokenPegs(DEFAULT_HLP_NETWORK);
    const hlpTokens = getTokensInfo();
    const peggedAddresses = cachedArbitrumPegs
        .get()
        .map((peg) => peg.peggedToken);
    const peggedTokens = HandleTokenManagerInstance.getTokensByAddresses(peggedAddresses.map((address) => ({
        address,
        network: DEFAULT_HLP_NETWORK,
    })));
    return [...hlpTokens, ...peggedTokens];
};
export const getTokenInfoByAddress = (address) => {
    const tokens = getTokensInfo();
    const isNative = isEtherAddress(address);
    const token = tokens.find((token) => !isNative
        ? token.address.toLowerCase() === address.toLowerCase()
        : token.symbol === "WETH");
    if (!token) {
        throw new Error(`could not find token with address: ${address}`);
    }
    return token;
};
export const getTradePair = ({ pair }) => pairToTradePairHlpSync(pair);
export const fetchLiquidityInfo = async (addresses, network = DEFAULT_HLP_NETWORK) => {
    const { vault } = getHlpContracts(network);
    const responses = await Promise.all(addresses.map(async (address) => {
        return {
            reservedAmount: await vault.reservedAmounts(address),
            poolAmount: await vault.poolAmounts(address),
        };
    }));
    const map = new Map();
    responses.forEach((response, i) => {
        map.set(addresses[i].toLowerCase(), response);
    });
    return map;
};
