import { BigNumber, ContractTransaction, ethers, PopulatedTransaction, Signer } from "ethers";
import { Provider } from "@ethersproject/providers";
import { Pair } from "../../../types/trade";
import { Subscription, SubscriptionId } from "../../../types/general";
import { TokenInfo } from "@uniswap/token-lists";
export declare const PRICE_DECIMALS = 30;
export declare const PRICE_UNIT: BigNumber;
export type PositionId = {
    pair: Pair;
    isLong: boolean;
    collateralAddress: string;
};
export type PositionInternal<T> = PositionId & {
    size: BigNumber;
    collateral: BigNumber;
    averagePrice: BigNumber;
    hasRealisedProfit: boolean;
    realisedPnL: BigNumber;
    lastIncreasedTime: BigNumber;
    reserveAmount: BigNumber;
    delta: BigNumber;
    hasProfit: boolean;
    leverage: BigNumber;
    fundingRatePpm: BigNumber;
    fundingFee: BigNumber;
    liquidationPrice: BigNumber;
    uid: string;
    platformName: string;
    internals: T;
};
export type Position = PositionInternal<unknown>;
export type TradePairInternal<T> = {
    pair: Pair;
    spreadBps: NumericSpread;
    marginFeeBps: BigNumber;
    indexAddress: string;
    platformName: string;
    maxLeverageBps: BigNumber;
    liquidationFeeUsd: BigNumber;
    internals: T;
};
export type TradePair = TradePairInternal<unknown>;
export type Spread<T> = {
    minimum: T;
    maximum: T;
};
export type NumericSpread = Spread<BigNumber>;
type EditPositionArgs = PositionId & {
    indexDelta: BigNumber;
    collateralDelta: BigNumber;
    signer: Signer;
    slippagePercent: number;
    receiver?: string;
    overrides?: ethers.Overrides;
};
type EditOrderArgs = Omit<EditPositionArgs, keyof PositionId> & {
    pair: Pair;
};
export type ApproveIncreasePositionArgs = Pick<EditPositionArgs, "collateralAddress" | "collateralDelta" | "signer" | "overrides" | "pair"> & {
    maximise?: boolean;
};
export type IncreasePositionArgs = EditPositionArgs;
export type DecreasePositionArgs = EditPositionArgs & {
    receiveCollateralAddress?: string;
};
type TriggerOrder = {
    triggerPrice: BigNumber;
    shouldTriggerAboveThreshold: boolean;
};
export type CreateIncreasePositionOrderArgs = EditOrderArgs & TriggerOrder & PositionId;
export type CreateDecreasePositionOrderArgs = CreateIncreasePositionOrderArgs;
export type UpdateDecreasePositionOrderArgs = EditOrderArgs & TriggerOrder & {
    orderId: string;
};
export type UpdateIncreasePositionOrderArgs = UpdateDecreasePositionOrderArgs;
export type GetAllPositionArgs = {
    account: string;
    provider: Provider;
};
export type GetPositionArgs = PositionId & GetAllPositionArgs;
export type TokenArgs = {
    tokenAddress: string;
};
export type GetTokenSpreadBasisPointsArgs = TokenArgs;
export type GetPriceArgs = {
    pair: Pair;
};
export type MarketPrice = {
    index: BigNumber;
    bestBid: BigNumber;
    bestAsk: BigNumber;
};
export type MarketPriceSubscription = Subscription<MarketPriceCallback, Pair>;
export type SubscribeMarketPriceArgs = {
    callback: MarketPriceCallback;
    pair: Pair;
};
export type MarketPriceCallback = (pair: Readonly<Pair>, price: Readonly<MarketPrice>) => any;
export type GetUpdatedPositionArgs = {
    position: Position;
    indexPriceOverride?: BigNumber;
};
export type PositionInput = {
    sizeDelta: BigNumber;
    collateralDelta: BigNumber;
    collateralAddress: string;
};
type SimulateEditPositionArgs = {
    position: Position;
    input: PositionInput;
    indexPriceOverride?: BigNumber;
};
export type SimulateIncreasePositionArgs = SimulateEditPositionArgs;
export type SimulateDecreasePositionArgs = SimulateEditPositionArgs;
export type Fee<T> = {
    amountUsd: BigNumber;
    metadata: T;
};
export type SwapFeeMetadata = {
    fromToken: TokenInfo;
    toToken: TokenInfo;
};
export type SimulatePositionFees = {
    totalAmountUsd: BigNumber;
    breakdown: {
        margin: Fee<undefined>;
        swap?: Fee<SwapFeeMetadata>;
    };
};
export type SimulatePositionResult = {
    position: Position;
    fees: SimulatePositionFees;
};
export type GetCollateralTokensArgs = {
    pair: Pair;
};
export type GetTradePairArgs = {
    pair: Pair;
};
export type GetTradeHistoryArgs = {
    account: string;
    limit?: number;
    page?: number;
};
export type TradeActionType = "increase" | "decrease" | "liquidation";
export type TradeAction = {
    id: string;
    account: string;
    collateralToken: TokenInfo;
    indexToken: TokenInfo;
    collateralDelta: BigNumber;
    sizeDelta: BigNumber;
    isLong: boolean;
    price: BigNumber;
    timestamp: number;
    type: TradeActionType;
    realisedPnl: BigNumber;
};
/**
 * Available liquidity, in the base token, for this trade pair.
 * long is returned as an amount of the index token
 * short is returned as a usd amount, to 30 decimals
 */
export type AvailableLiquidity = {
    long: BigNumber;
    short: BigNumber;
};
export type GetAvailableLiquidityArgs = {
    pair: Pair;
    forceFetch?: boolean;
};
export type GetAvailabilityArgs = {
    pair: Pair;
};
export type Availability = {
    isAvailable: true;
    reason?: undefined;
} | {
    isAvailable: false;
    reason?: string;
};
export type ActiveIncreaseOrder = PositionId & TriggerOrder & {
    purchaseTokenAmount: BigNumber;
    purchaseToken: string;
    sizeDelta: BigNumber;
    orderId: string;
};
export type ActiveDecreaseOrder = PositionInput & PositionId & TriggerOrder & {
    orderId: string;
};
export type GetActiveOrderArgs = {
    account: string;
    provider: Provider;
};
export type ActiveOrders = {
    increase: ActiveIncreaseOrder[];
    decrease: ActiveDecreaseOrder[];
};
export type GetMinimumPositionCollateralArgs = {
    pair: Pair;
    existingCollateral: BigNumber;
    isTriggerOrder: boolean;
};
/**
 * An interface for perpetual trading platforms.
 * The platform has a single collateral token for each position.
 * The platform trade pair consists of an index token and a quote asset,
 * where the quote asset does not need to be a token.
 * The platform prices use 30 decimals of precision.
 */
export interface Trade {
    name: string;
    initialise(): Promise<void>;
    /**
     * Returns an array of contract transactions for required approvals.
     * If no approvals are required, the array is returned empty.
     */
    approveIncreasePosition(args: ApproveIncreasePositionArgs): Promise<PopulatedTransaction[]>;
    increasePosition(args: IncreasePositionArgs): Promise<ContractTransaction>;
    decreasePosition(args: DecreasePositionArgs): Promise<ContractTransaction>;
    approveCreateIncreasePositionOrder(args: ApproveIncreasePositionArgs): Promise<PopulatedTransaction[]>;
    createIncreasePositionOrder(args: CreateIncreasePositionOrderArgs): Promise<ContractTransaction>;
    createDecreasePositionOrder(args: CreateIncreasePositionOrderArgs): Promise<ContractTransaction>;
    updateIncreasePositionOrder(args: UpdateIncreasePositionOrderArgs): Promise<ContractTransaction>;
    updateDecreasePositionOrder(args: UpdateDecreasePositionOrderArgs): Promise<ContractTransaction>;
    getPosition(args: GetPositionArgs): Promise<Position>;
    getAllPositions(args: GetAllPositionArgs): Promise<Position[]>;
    getUpdatedPosition(args: GetUpdatedPositionArgs): Position;
    simulateIncreasePosition(args: SimulateIncreasePositionArgs): SimulatePositionResult;
    simulateDecreasePosition(args: SimulateDecreasePositionArgs): SimulatePositionResult;
    getLpTokenPrice(): Promise<BigNumber>;
    getPrice(args: GetPriceArgs): MarketPrice;
    subscribeToMarketPrice(args: SubscribeMarketPriceArgs): SubscriptionId;
    unsubscribeFromMarketPrice(id: SubscriptionId): void;
    getTradePairs(): Promise<TradePair[]>;
    getTokens(): TokenInfo[];
    getCollateralTokens(args: GetCollateralTokensArgs): TokenInfo[];
    getTradePair(args: GetTradePairArgs): TradePair;
    getTradeHistory(args: GetTradeHistoryArgs): Promise<TradeAction[]>;
    getAvailableLiquidity(args: GetAvailableLiquidityArgs): Promise<AvailableLiquidity>;
    getPairAvailability: (args: GetAvailabilityArgs) => Availability;
    getActiveOrders(args: GetActiveOrderArgs): Promise<ActiveOrders>;
    getMinimumPositionCollateral(args: GetMinimumPositionCollateralArgs): BigNumber;
}
export type TokenInfoGlp = TokenInfo & {
    extensions: {
        isUsdg?: boolean;
        isNative?: boolean;
        isWrappedNative?: boolean;
        isShortable?: boolean;
        isStable?: boolean;
    };
};
export {};
