import { pairToString } from "../../utils/general";
import { AMOUNT_UNIT, PRICE_UNIT, } from "./reader";
import { ethers, BigNumber } from "ethers";
const ZERO = BigNumber.from(0);
const MINUS_ONE = BigNumber.from(-1);
export class Position {
    constructor(pairId, size, entryPrice, snapshotSumFractionFunding, snapshotSumFractionBorrow) {
        this.pairId = pairId;
        this.size = size;
        this.entryPrice = entryPrice;
        this.snapshotSumFractionFunding = snapshotSumFractionFunding;
        this.snapshotSumFractionBorrow = snapshotSumFractionBorrow;
    }
    /**
     * Creates an empty position.
     * @param pairId The position pair ID.
     * @param currentSumFractionFunding The current sum fraction for funding.
     * If not passed this defaults to zero which may result in the position
     * having an incorrect, non-zero accrued funding fee.
     * @param currentSumFractionBorrow Similar to the funding argument.
     */
    static Zero(pairId, currentSumFractionFunding = BigNumber.from(0), currentSumFractionBorrow = BigNumber.from(0)) {
        return new Position(pairId, ethers.constants.Zero, ethers.constants.Zero, currentSumFractionFunding, currentSumFractionBorrow);
    }
    get isLong() {
        return this.size.gt("0");
    }
    get isShort() {
        return this.size.lt("0");
    }
    get isZero() {
        return this.size.eq(0);
    }
    /// Returns the unrealized equity to be realized given the position is fully
    /// closed, i.e. the PnL for the entire position (not including fees such as
    //  the margin fee).
    calculateFullUnrealizedEquityToBeRealized(marketPrice) {
        const tradeSize = this.size.mul("-1");
        return this.calculateUnrealizedEquityToBeRealized(tradeSize, marketPrice);
    }
    /// Returns accrued account fees, i.e. borrow and funding fees.
    calculateAccruedAccountFees(pairState, indexPrice) {
        return this.calculateAccruedFundingFee(pairState, indexPrice).add(this.calculateAccruedBorrowFee(pairState, indexPrice));
    }
    calculateAccruedFundingFee(pairState, indexPrice) {
        const sumFractionFunding = pairState.getCurrentSumFractionFunding(indexPrice);
        return calculateFeeWithSumFractions(this.size, this.snapshotSumFractionFunding, this.isLong ? sumFractionFunding.long : sumFractionFunding.short);
    }
    calculateAccruedBorrowFee(pairState, indexPrice) {
        const sumFractionBorrow = pairState.getCurrentSumFractionBorrow(indexPrice);
        return calculateFeeWithSumFractions(this.size, this.snapshotSumFractionBorrow, this.isLong ? sumFractionBorrow.long : sumFractionBorrow.short);
    }
    /// Returns the equity to be realized given a trade size is filled.
    /// When a position is decreased, the user realizes its unrealized PnL
    /// proportionally to the trade size, up to the full position size.
    /// When a position is increased, no equity is realized.
    /// This does not include fees (such as the margin fee).
    calculateUnrealizedEquityToBeRealized(tradeSize, marketPrice) {
        const tradePrice = marketPrice.getTradePriceFromTradeSize(tradeSize);
        const tradeSizeEffect = this.getPositionTradeSizeEffect(tradeSize);
        return this.calculateRealizedPnl(tradeSizeEffect, tradePrice);
    }
    /// When increasing a position size, due to spread the trade may
    /// add new unrealized equity (at a loss).
    calculateNewUnrealizedEquity(tradeSize, marketPrice) {
        const tradeSizeEffect = this.getPositionTradeSizeEffect(tradeSize);
        const getIncreaseSize = () => {
            switch (tradeSizeEffect.sizeChangeType) {
                case SizeChangeType.Decrease:
                    return ZERO;
                case SizeChangeType.Increase:
                    return tradeSizeEffect.relevantTradeSize;
                case SizeChangeType.CloseAndFlipIncrease:
                    return tradeSizeEffect.relevantTradeSize;
                default:
                    throw new Error("unknown tradeSizeEffect enum variant");
            }
        };
        const isBuying = tradeSize.gt(0);
        const increaseSize = getIncreaseSize();
        const entryPrice = marketPrice.getTradePrice(isBuying);
        const markPrice = marketPrice.getTradePrice(!isBuying);
        const spreadLoss = increaseSize
            .mul(markPrice.sub(entryPrice))
            .div(PRICE_UNIT);
        if (spreadLoss.gt(0)) {
            throw new Error("spread loss cannot be positive");
        }
        return spreadLoss;
    }
    /// Calculates the REALIZED PnL from a trade.
    /// This does not account for any unrealised PnL that may be created
    /// after the trade is done, such as due to spread.
    calculateRealizedPnl(tradeSizeEffect, price) {
        return this.getSizeDecreasing(tradeSizeEffect)
            .mul(price.sub(this.entryPrice))
            .div(PRICE_UNIT);
    }
    applyTradeEffect(effect, size) {
        const nextSize = this.size.add(size);
        const tradeSizeEffect = this.getPositionTradeSizeEffect(size);
        const nextEntryPrice = this.getNextEntryPrice(tradeSizeEffect, effect.fillPrice);
        // Any changes to the position's state must happen AFTER this.
        this.size = nextSize;
        this.entryPrice = nextEntryPrice;
    }
    getPositionTradeSizeEffect(tradeSize) {
        const isIncreasing = tradeSize.gt(ZERO) == this.size.gt(ZERO);
        if (isIncreasing) {
            return {
                sizeChangeType: SizeChangeType.Increase,
                relevantTradeSize: tradeSize,
            };
        }
        const tradeSizeAbs = tradeSize.abs();
        const positionSizeAbs = this.size.abs();
        if (tradeSizeAbs.lte(positionSizeAbs)) {
            return {
                sizeChangeType: SizeChangeType.Decrease,
                relevantTradeSize: tradeSize,
            };
        }
        const increaseSizeAfterClose = tradeSize.add(this.size);
        return {
            sizeChangeType: SizeChangeType.CloseAndFlipIncrease,
            relevantTradeSize: increaseSizeAfterClose,
        };
    }
    getNextEntryPrice(tradeSizeEffect, tradePrice) {
        switch (tradeSizeEffect.sizeChangeType) {
            case SizeChangeType.Decrease:
                return this.entryPrice;
            case SizeChangeType.CloseAndFlipIncrease:
                return tradePrice;
            case SizeChangeType.Increase:
                const existingSize = {
                    amount: this.entryPrice,
                    weight: this.size,
                };
                const tradeSize = {
                    amount: tradePrice,
                    weight: tradeSizeEffect.relevantTradeSize,
                };
                return calculateWeightedAverage([existingSize, tradeSize]);
            default:
                throw new Error("unknown tradeSizeEffect enum variant");
        }
    }
    /// Returns the position size that is subtracted from this position
    /// by a `PositionTradeSizeEffect`.
    getSizeDecreasing(tradeSizeEffect) {
        switch (tradeSizeEffect.sizeChangeType) {
            case SizeChangeType.Increase:
                return ZERO;
            case SizeChangeType.Decrease:
                return tradeSizeEffect.relevantTradeSize.mul(MINUS_ONE);
            case SizeChangeType.CloseAndFlipIncrease:
                return this.size;
            default:
                throw new Error("unknown tradeSizeEffect enum variant");
        }
    }
    getNotionalValue(markPrice) {
        return this.size.mul(markPrice).div(PRICE_UNIT);
    }
    getMarkPrice(marketPrice) {
        return marketPrice.getTradePrice(this.isShort);
    }
    /// The "increase price" is the price used for trades that
    /// match the side of the position.
    /// For example, if placing a buy trade on a long position,
    /// the price used is the increase price i.e. the ask.
    /// However, if placing a sell trade on a long position,
    /// the price used is the mark price which is the "decrease" price.
    getIncreasePrice(marketPrice) {
        return marketPrice.getTradePrice(this.isLong);
    }
    async getPeriodicFees(args) {
        return args.reader.getPeriodicFeesForPosition(args.accountId, pairToString(this.pairId.pair), args.startTimestamp);
    }
    clone() {
        return new Position(this.pairId, this.size, this.entryPrice, this.snapshotSumFractionFunding, this.snapshotSumFractionBorrow);
    }
}
export var SizeChangeType;
(function (SizeChangeType) {
    SizeChangeType[SizeChangeType["Increase"] = 0] = "Increase";
    SizeChangeType[SizeChangeType["Decrease"] = 1] = "Decrease";
    SizeChangeType[SizeChangeType["CloseAndFlipIncrease"] = 2] = "CloseAndFlipIncrease";
})(SizeChangeType || (SizeChangeType = {}));
const calculateWeightedAverage = (inputs) => {
    const totalWeight = inputs
        .map((input) => input.weight)
        .reduce((sum, weight) => sum.add(weight), ZERO);
    return inputs.reduce((weightedAverage, input) => weightedAverage.add(input.weight.mul(input.amount).div(totalWeight)), ZERO);
};
const calculateFeeWithSumFractions = (size, previousSumFraction, currentSumFraction) => size.abs().mul(currentSumFraction.sub(previousSumFraction)).div(AMOUNT_UNIT);
