import { LiquidityPoolSingle } from "./liquidityPool";
import { AMOUNT_DECIMALS } from "./reader";
import { isSameAddress, pairFromString } from "../../utils/general";
import { TradePairId, TradePair, } from "./index";
import { parseUnits } from "ethers/lib/utils";
import { retryPromise } from "../../utils/sdk";
export class TradeProtocol {
    constructor(liquidityPools) {
        this.liquidityPools = liquidityPools;
    }
    static async create(reader, adapter, usdToken, priceFeed) {
        await adapter.waitForConnect();
        const liquidityPools = await retryPromise(reader.getLiquidityPools.bind(reader));
        const pairConfig = await adapter.getLpConfig();
        const tradePairs = getTradePairs(pairConfig);
        const pools = liquidityPools
            .filter((pool) => isSameAddress(usdToken.address, pool.underlyingToken))
            .map((pool) => {
            return new LiquidityPoolSingle(pool.id, tradePairs[pool.id] || [], usdToken, adapter, priceFeed);
        });
        return new TradeProtocol(pools);
    }
    getLiquidityPools() {
        return this.liquidityPools;
    }
    getLiquidityPoolsOfUnderlyingToken(tokenAddress) {
        return this.liquidityPools.filter((pool) => pool.getUnderlyingToken().address.toLowerCase() ===
            tokenAddress.toLowerCase());
    }
    getLiquidityPool(id) {
        const liquidityPool = this.liquidityPools.find((pool) => pool.id === id);
        if (!liquidityPool) {
            throw new Error(`Liquidity pool ${id} not found`);
        }
        return liquidityPool;
    }
    getTradePair(pairId) {
        return this.getLiquidityPool(pairId.lpId).getTradePair(pairId.pair);
    }
    getTradePairs() {
        return this.liquidityPools.flatMap((pool) => pool.getTradePairs());
    }
    subscribeToMarketPrice(args, lpId) {
        return this.getLiquidityPool(lpId).subscribeToMarketPrice(args);
    }
    unsubscribeFromMarketPrice(id, lpId) {
        this.getLiquidityPool(lpId).unsubscribeFromMarketPrice(id);
    }
    getPrice(pairId, size) {
        return this.getLiquidityPool(pairId.lpId).getPrice(pairId.pair, size);
    }
    tryGetPrice(pairId, size) {
        try {
            return this.getPrice(pairId, size);
        }
        catch (e) {
            return undefined;
        }
    }
    getTradePrice(lpId, args) {
        return this.getLiquidityPool(lpId).getTradePrice(args);
    }
    getOpenInterest(pairId) {
        return this.getLiquidityPool(pairId.lpId).getOpenInterest({
            pair: pairId.pair,
        });
    }
}
const getTradePairs = (lpConfig) => {
    if (lpConfig.error) {
        throw new Error(lpConfig.error);
    }
    const pairs = {};
    for (const pair of lpConfig.result.content.lpConfigUpdate) {
        if (!pairs[pair.lpId]) {
            pairs[pair.lpId] = [];
        }
        pairs[pair.lpId].push(new TradePair(new TradePairId(pairFromString(pair.pair), pair.lpId), parseUnits(pair.config.initialMarginFraction, AMOUNT_DECIMALS), parseUnits(pair.config.maintenanceMarginFraction, AMOUNT_DECIMALS), parseUnits(pair.config.incrementalInitialMarginFraction, AMOUNT_DECIMALS), parseUnits(pair.config.baselinePositionSize, AMOUNT_DECIMALS), parseUnits(pair.config.incrementalPositionSize, AMOUNT_DECIMALS), parseUnits(pair.config.marginFeeFraction, AMOUNT_DECIMALS), parseUnits(pair.config.symmetricalSpreadFraction, AMOUNT_DECIMALS), pair.config.isActive, pair.config.isReduceOnly, pair.config.maxOpenInterestDiff === null
            ? null
            : parseUnits(pair.config.maxOpenInterestDiff, AMOUNT_DECIMALS), pair.config.maxOpenInterestLong === null
            ? null
            : parseUnits(pair.config.maxOpenInterestLong, AMOUNT_DECIMALS), pair.config.maxOpenInterestShort === null
            ? null
            : parseUnits(pair.config.maxOpenInterestShort, AMOUNT_DECIMALS), parseUnits(pair.config.borrowFeeFactor, AMOUNT_DECIMALS), parseUnits(pair.config.fundingFactor, AMOUNT_DECIMALS), parseUnits(pair.config.fundingExponent, AMOUNT_DECIMALS), { pair: pairFromString(pair.pair) }, pair.config.usePriceImpact, pair.config.priceImpactFraction
            ? parseUnits(pair.config.priceImpactFraction, AMOUNT_DECIMALS)
            : null, pair.config.skewScale
            ? parseUnits(pair.config.skewScale, AMOUNT_DECIMALS)
            : null));
    }
    return pairs;
};
