import { TradeAccountRole } from "./account";
import { BigNumber } from "ethers";
export declare const AMOUNT_DECIMALS = 18;
export declare const PRICE_DECIMALS = 8;
export declare const AMOUNT_UNIT: BigNumber;
export declare const PRICE_UNIT: BigNumber;
export type ReaderAccount = {
    id: string;
    isOpen: boolean;
    realizedEquities: {
        token: string;
        value: string;
    }[];
    positions: ReaderPosition[];
};
export type ReaderMeta = {
    _meta: {
        block: {
            timestamp: number;
        };
    };
};
export type ReaderAccountResponse = {
    account: ReaderAccount | null;
} & ReaderMeta;
export type ReaderAccountRoleQuery = {
    users: {
        address: string;
        roles: string[];
    }[];
    id: string;
};
export type ReaderAccountUserAccountsQuery = {
    accountUsers: {
        account: {
            id: string;
        };
    }[];
};
export type ReaderPosition = {
    account: {
        id: string;
    };
    liquidityPool: {
        id: string;
    };
    pair: string;
    size: string;
    entryPrice: string;
    snapshotSumFractionFunding: string;
    snapshotSumFractionBorrow: string;
};
export type ReaderLiquidityPool = {
    id: string;
    owner: string;
    underlyingToken: string;
    realizedEquity: string;
    liquidityTokenSupply: string;
    accruedFees: string;
    undistributedFees: string;
    openInterests: ReaderLiquidityPoolOpenInterest[];
};
export type ReaderLiquidityPoolOpenInterest = {
    pair: string;
    long: string;
    short: string;
};
export declare enum ReaderTradeType {
    Trade = 0,
    Liquidation = 1
}
export type ReaderTrade = {
    id: string;
    account: {
        id: string;
    };
    liquidityPool: {
        id: string;
    };
    pair: string;
    size: string;
    price: string;
    marginFee: string;
    realizedEquity: string;
    tradeType: ReaderTradeType;
    transaction: {
        timestamp: string;
        hash: string;
    };
    didOpenPosition: boolean;
    didClosePosition: boolean;
};
export type ReaderAccountAssetDeposit = {
    id: string;
    account: {
        id: string;
    };
    depositorAddress: string;
    assetAddress: string;
    amount: string;
    transaction: {
        timestamp: string;
        hash: string;
    };
};
export type ReaderAccountAssetWithdrawal = {
    id: string;
    account: {
        id: string;
    };
    accountUser: string;
    assetAddress: string;
    amount: string;
    transaction: {
        timestamp: string;
        hash: string;
    };
};
export type ReaderDepositWithdrawHistory = {
    accountAssetDeposits: ReaderAccountAssetDeposit[];
    accountAssetWithdrawals: ReaderAccountAssetWithdrawal[];
};
export declare enum ReaderPeriodicPositionFeeType {
    Borrow = 0,
    Funding = 1
}
export type ReaderPeriodicPositionFeeCollection = {
    id: string;
    account: {
        id: string;
    };
    pair: string;
    liquidityPool: {
        id: string;
    };
    transaction: {
        timestamp: string;
    };
    periodicPositionFeeType: ReaderPeriodicPositionFeeType;
    amount: string;
};
export type ReaderSystemParam = {
    paramId: string;
    paramValue: string;
};
export type SystemParams = Record<string, BigNumber | undefined>;
export type ReaderGetTradesOptions = {
    accountId?: number;
    pair?: string;
    liquidityPoolId?: string;
    skipOrdering?: boolean;
};
export type ReaderGetPositionsOptions = {
    accountId?: number;
    pair?: string;
    liquidityPoolId?: string;
    skipOrdering?: boolean;
};
export type ReaderHistoricalLpPairVolume = {
    id: string;
    liquidityPool: {
        id: string;
    };
    pair: string;
    volume: string;
    volumeLpc: string;
};
export interface TradeReader {
    getAccount(accountId: number): Promise<ReaderAccountResponse>;
    getUserAccountIds(userAddress: string, userRole: TradeAccountRole): Promise<number[]>;
    getLiquidityPools(): Promise<ReaderLiquidityPool[]>;
    getTradeHistory(accountId: number, limit: number, offset: number, pair?: string): Promise<ReaderTrade[]>;
    getTrades(limit: number, offset: number, options?: ReaderGetTradesOptions): Promise<ReaderTrade[]>;
    getDepositWithdrawHistory(accountId: number, limit: number, offset: number): Promise<ReaderDepositWithdrawHistory>;
    getPeriodicFeeHistory(accountId: number, limit: number, offset: number, startTimestamp?: number): Promise<ReaderPeriodicPositionFeeCollection[]>;
    getPeriodicFeesForPosition(accountId: number, pair: string, startTimestamp: number): Promise<ReaderPeriodicPositionFeeCollection[]>;
    getSystemParams(): Promise<SystemParams>;
    getSystemParam(paramName: string): Promise<BigNumber>;
    getWithdrawGasFeeUsd(): Promise<BigNumber>;
    getTradeGasFeeUsd(): Promise<BigNumber>;
    getPositions(limit: number, offset: number, options?: ReaderGetPositionsOptions): Promise<ReaderPosition[]>;
    getHistoricalLpPairVolumes(limit: number, offset: number, liquidityPoolId: string, pair?: string): Promise<ReaderHistoricalLpPairVolume[]>;
}
export declare class TradeReaderSubgraph implements TradeReader {
    private graphUrl;
    constructor(graphUrl: string);
    getUserAccountIds(userAddress: string, userRole: TradeAccountRole): Promise<number[]>;
    getAccount(accountId: number): Promise<ReaderAccountResponse>;
    getLiquidityPools(): Promise<ReaderLiquidityPool[]>;
    /**
     * Gets the trade history for a trade account.
     */
    getTradeHistory(accountId: number, limit: number, offset: number, pair?: string): Promise<ReaderTrade[]>;
    getTrades(limit: number, offset: number, { accountId, pair, liquidityPoolId, skipOrdering, }?: ReaderGetTradesOptions): Promise<ReaderTrade[]>;
    getDepositWithdrawHistory(accountId: number, limit: number, offset: number): Promise<ReaderDepositWithdrawHistory>;
    getPeriodicFeeHistory(accountId: number, limit: number, offset: number, startTimestamp?: number): Promise<ReaderPeriodicPositionFeeCollection[]>;
    getPeriodicFeesForPosition(accountId: number, pair: string, startTimestamp: number): Promise<ReaderPeriodicPositionFeeCollection[]>;
    getSystemParams(): Promise<SystemParams>;
    getSystemParam(paramId: string): Promise<BigNumber>;
    getWithdrawGasFeeUsd(): Promise<BigNumber>;
    getTradeGasFeeUsd(): Promise<BigNumber>;
    getPositions(limit: number, offset: number, { accountId, pair, liquidityPoolId, skipOrdering, }?: ReaderGetPositionsOptions): Promise<ReaderPosition[]>;
    getHistoricalLpPairVolumes(limit: number, offset: number, liquidityPoolId: string, pair?: string): Promise<ReaderHistoricalLpPairVolume[]>;
}
export declare const parseAmount: (stringified: string) => BigNumber;
export declare const parsePrice: (stringified: string) => BigNumber;
export declare const formatAmount: (amount: BigNumber) => string;
export declare const formatPrice: (price: BigNumber) => string;
