import { formatUnits, parseUnits } from "ethers/lib/utils";
import { request, gql } from "graphql-request";
import { H2SO_PRICE_DECIMALS } from "../../constants";
import { BigNumber } from "ethers";
export const AMOUNT_DECIMALS = 18;
export const PRICE_DECIMALS = H2SO_PRICE_DECIMALS;
export const AMOUNT_UNIT = parseUnits("1", AMOUNT_DECIMALS);
export const PRICE_UNIT = parseUnits("1", PRICE_DECIMALS);
export var ReaderTradeType;
(function (ReaderTradeType) {
    ReaderTradeType[ReaderTradeType["Trade"] = 0] = "Trade";
    ReaderTradeType[ReaderTradeType["Liquidation"] = 1] = "Liquidation";
})(ReaderTradeType || (ReaderTradeType = {}));
export var ReaderPeriodicPositionFeeType;
(function (ReaderPeriodicPositionFeeType) {
    ReaderPeriodicPositionFeeType[ReaderPeriodicPositionFeeType["Borrow"] = 0] = "Borrow";
    ReaderPeriodicPositionFeeType[ReaderPeriodicPositionFeeType["Funding"] = 1] = "Funding";
})(ReaderPeriodicPositionFeeType || (ReaderPeriodicPositionFeeType = {}));
export class TradeReaderSubgraph {
    constructor(graphUrl) {
        this.graphUrl = graphUrl;
    }
    async getUserAccountIds(userAddress, userRole) {
        userAddress = userAddress.toLowerCase();
        const query = gql `
      query ownedAccountsForAddress($address: String!, $role: String!) {
        accountUsers(where: { address: $address, roles_contains: [$role] }) {
          account {
            id
          }
        }
      }
    `;
        const response = await request(this.graphUrl, query, { address: userAddress, role: String(+userRole) });
        if (!response.accountUsers || response.accountUsers.length === 0) {
            return [];
        }
        return response.accountUsers.map((user) => +user.account.id);
    }
    async getAccount(accountId) {
        const query = gql `
      query getAccount($accountId: Int!) {
        account(id: $accountId) {
          id
          isOpen
          realizedEquities {
            token
            value
          }
          positions {
            account {
              id
            }
            pair
            size
            entryPrice
            snapshotSumFractionFunding
            snapshotSumFractionBorrow
            liquidityPool {
              id
            }
          }
        }
        _meta {
          block {
            timestamp
          }
        }
      }
    `;
        return request(this.graphUrl, query, {
            accountId,
        });
    }
    async getLiquidityPools() {
        const query = gql `
      query getLiquidityPools {
        liquidityPools {
          owner
          id
          underlyingToken
          realizedEquity
          liquidityTokenSupply
          accruedFees
          undistributedFees
          openInterests {
            pair
            short
            long
          }
        }
      }
    `;
        const response = await request(this.graphUrl, query);
        return response.liquidityPools;
    }
    /**
     * Gets the trade history for a trade account.
     */
    async getTradeHistory(accountId, limit, offset, pair) {
        return this.getTrades(limit, offset, { accountId, pair });
    }
    async getTrades(limit, offset, { accountId, pair, liquidityPoolId, skipOrdering, } = {}) {
        const { filterString, variableString, ordering } = getQueryArgStrings(accountId?.toString(), pair, liquidityPoolId, skipOrdering);
        const query = gql `
      query getTradeHistory(${variableString}) {
        trades(
          where: { ${filterString} }
          ${ordering}   
          first: $limit
          skip: $offset
        ) {
          id
          account {
            id
          }
          liquidityPool {
            id
          }
          pair
          size
          price
          marginFee
          realizedEquity
          tradeType
          transaction {
            timestamp
            hash
          }
          didOpenPosition
          didClosePosition
        }
      }
    `;
        const response = await request(this.graphUrl, query, {
            limit,
            offset,
            pair,
            account: accountId?.toString(),
            liquidityPool: liquidityPoolId?.toString(),
        });
        return response.trades;
    }
    async getDepositWithdrawHistory(accountId, limit, offset) {
        const query = gql `
      query getDepositWithdrawHistory(
        $accountId: String!
        $limit: Int
        $offset: Int
      ) {
        accountAssetDeposits(
          where: { account: $accountId }
          orderBy: transaction__timestamp
          orderDirection: desc
          first: $limit
          skip: $offset
        ) {
          id
          account {
            id
          }
          depositorAddress
          assetAddress
          amount
          transaction {
            hash
            timestamp
          }
        }
        accountAssetWithdrawals(
          where: { account: $accountId }
          orderBy: transaction__timestamp
          orderDirection: desc
          first: $limit
          skip: $offset
        ) {
          id
          account {
            id
          }
          accountUser
          amount
          transaction {
            hash
            timestamp
          }
        }
      }
    `;
        return request(this.graphUrl, query, {
            accountId: accountId.toString(),
            limit,
            offset,
        });
    }
    async getPeriodicFeeHistory(accountId, limit, offset, startTimestamp) {
        const query = gql `
      query getPeriodicFeeHistory(
        $accountId: String!
        $limit: Int
        $offset: Int
        ${startTimestamp ? "$startTimestamp: Int" : ""}
      ) {
        periodicPositionFeeCollections(
          where: { account: $accountId${startTimestamp
            ? ", transaction_: { timestamp_gte: $startTimestamp }"
            : ""} }
          orderBy: transaction__timestamp
          orderDirection: desc
          first: $limit
          skip: $offset
        ) {
          id
          timestamp
          account {
            id
          }
          pair
          liquidityPool {
            id
          }
          transaction {
            timestamp
          }
          amount
          periodicPositionFeeType
        }
      }
    `;
        const response = await request(this.graphUrl, query, {
            accountId: accountId.toString(),
            limit,
            offset,
            startTimestamp: startTimestamp?.toString(),
        });
        return response.periodicPositionFeeCollections;
    }
    async getPeriodicFeesForPosition(accountId, pair, startTimestamp) {
        const query = gql `
      query getPeriodicFeeHistory(
        $accountId: String!
        $pair: String!
        $startTimestamp: String!
      ) {
        periodicPositionFeeCollections(
          where: {
            account: $accountId
            pair: $pair
            transaction_: { timestamp_gte: $startTimestamp }
          }
          orderBy: transaction__timestamp
          orderDirection: desc
        ) {
          id
          timestamp
          account {
            id
          }
          pair
          liquidityPool {
            id
          }
          transaction {
            timestamp
          }
          amount
          periodicPositionFeeType
        }
      }
    `;
        const response = await request(this.graphUrl, query, {
            accountId: accountId.toString(),
            pair,
            startTimestamp: startTimestamp.toString(),
        });
        return response.periodicPositionFeeCollections;
    }
    async getSystemParams() {
        const query = gql `
      query getSystemParams {
        systemParams(first: 1000) {
          paramId
          paramValue
        }
      }
    `;
        const response = await request(this.graphUrl, query);
        return Object.fromEntries(response.systemParams.map((param) => {
            if (!param.paramValue) {
                throw new Error(`Invalid param value: ${param.paramValue}`);
            }
            return [param.paramId, BigNumber.from(param.paramValue)];
        }));
    }
    async getSystemParam(paramId) {
        const query = gql `
      query getSystemParam($paramId: String!) {
        systemParams(where: { paramId: $paramId }) {
          paramId
          paramValue
        }
      }
    `;
        const response = await request(this.graphUrl, query, {
            paramId,
        });
        const systemParam = response.systemParams[0];
        if (!systemParam) {
            throw new Error(`Invalid param value: ${systemParam}`);
        }
        return BigNumber.from(systemParam.paramValue);
    }
    async getWithdrawGasFeeUsd() {
        const withdrawGasUnit = await this.getSystemParam("gas.unit.withdraw");
        const gasUsdUnit = await this.getSystemParam("gas.usd.unit");
        return withdrawGasUnit.mul(gasUsdUnit);
    }
    async getTradeGasFeeUsd() {
        const tradeGasUnit = await this.getSystemParam("gas.unit.trade");
        const gasUsdUnit = await this.getSystemParam("gas.usd.unit");
        return tradeGasUnit.mul(gasUsdUnit);
    }
    async getPositions(limit, offset, { accountId, pair, liquidityPoolId, skipOrdering, } = {}) {
        const { filterString, variableString, ordering } = getQueryArgStrings(accountId?.toString(), pair, liquidityPoolId, skipOrdering);
        const query = gql `
      query getPositions(${variableString}) {
        positions(
          where: { ${filterString} }
          ${ordering}   
          first: $limit
          skip: $offset
      ) {
          account {
            id
          }
          entryPrice
          id
          pair
          size
          snapshotSumFractionFunding
          snapshotSumFractionBorrow
          liquidityPool {
            id
          }
        }
      }
    `;
        const response = await request(this.graphUrl, query, {
            limit,
            offset,
            pair,
            account: accountId?.toString(),
            liquidityPool: liquidityPoolId?.toString(),
        });
        return response.positions;
    }
    async getHistoricalLpPairVolumes(limit, offset, liquidityPoolId, pair) {
        const { filterString, variableString } = getQueryArgStrings(undefined, pair, liquidityPoolId, true);
        const query = gql `
      query getHistoricalLpPairVolumes(${variableString}) {
        historicalLpPairVolumes(
          where: { ${filterString} }
          first: $limit
          skip: $offset
        ) {
          id
          liquidityPool {
            id
          }
          pair
          volume
          volumeLpc
        }
      }
    `;
        const response = await request(this.graphUrl, query, {
            liquidityPool: liquidityPoolId,
            pair: pair,
            limit,
            offset,
        });
        return response.historicalLpPairVolumes;
    }
}
export const parseAmount = (stringified) => parseUnits(stringified, AMOUNT_DECIMALS);
export const parsePrice = (stringified) => parseUnits(stringified, PRICE_DECIMALS);
export const formatAmount = (amount) => formatUnits(amount, AMOUNT_DECIMALS);
export const formatPrice = (price) => formatUnits(price, PRICE_DECIMALS);
const getGqQueryVariableExpressions = (variables) => variables.reduce((str, variable) => {
    if (!variable.isDefined) {
        return str;
    }
    const leftPrefix = variable.isVariableOnLeft ? "$" : "";
    const rightPrefix = variable.isVariableOnLeft ? "" : "$";
    return str.length === 0
        ? `${leftPrefix}${variable.left}: ${rightPrefix}${variable.right}`
        : `${str}, ${leftPrefix}${variable.left}: ${rightPrefix}${variable.right}`;
}, "");
const getQueryArgStrings = (accountId, pair, liquidityPoolId, skipOrdering) => {
    const optionalVariables = [
        {
            left: "account",
            right: "String!",
            isDefined: !!accountId,
            isVariableOnLeft: true,
        },
        {
            left: "pair",
            right: "String!",
            isDefined: !!pair,
            isVariableOnLeft: true,
        },
        {
            left: "liquidityPool",
            right: "String!",
            isDefined: !!liquidityPoolId,
            isVariableOnLeft: true,
        },
    ];
    const optionalVariableFilters = optionalVariables.map((v) => ({
        ...v,
        // In a filter, both sides are the same
        // since the variable name is the same as the property.
        right: v.left,
        isVariableOnLeft: false,
    }));
    return {
        filterString: getGqQueryVariableExpressions(optionalVariableFilters),
        variableString: getGqQueryVariableExpressions([
            {
                left: "limit",
                right: "Int",
                isDefined: true,
                isVariableOnLeft: true,
            },
            {
                left: "offset",
                right: "Int",
                isDefined: true,
                isVariableOnLeft: true,
            },
            ...optionalVariables,
        ]),
        ordering: `${!skipOrdering
            ? `
        orderBy: transaction__timestamp
        orderDirection: desc
        `
            : ""}`,
    };
};
