import { ethers, } from "ethers";
import { BENTOBOX_ADDRESS, KASHI_ADDRESS } from "@sushiswap/core-sdk";
import { ERC20__factory, ERC2612__factory, SushiBento__factory, } from "../contracts";
import { NETWORK_NAME_TO_CHAIN_ID } from "../constants";
import { didUserCancelTransaction } from "./web3";
import config from "../config";
import { splitSignature } from "ethers/lib/utils";
import axios from "axios";
const PERMIT_TOKEN_SYMBOLS = ["USDC", "USDC.e", "DAI", "USDT", "FRAX"];
const PERMIT_IMPLEMENTATION_NAMES = ["ERC2612"];
const PERMIT_TOKEN_SYMBOL_BY_IMPLEMENTATION = {
    ERC2612: ["USDC", "USDC.e", "DAI", "USDT", "FRAX"],
};
/**
 * Checks if the account has enough allowance for the given ERC20 token.
 * If not, it will request the allowance.
 * This will try to use permits if the contract supports ERC2612.
 * @param account The account to ensure allowance for.
 * @param tokenInfo The ERC20 to approve.
 * @param allowanceTarget The spender (EOA or contract) address for the allowance.
 * @param signer The signer of the account.
 * @param minimumAllowance The minimum amount that must be approved.
 * @param currentAllowance If provided, it will skip the allowance check and use this value instead.
 * @returns an AllowanceResult struct.
 */
export const ensureHasAllowance = async (account, tokenInfo, allowanceTarget, signer, minimumAllowance, currentAllowance) => {
    const token = ERC20__factory.connect(tokenInfo.address, signer);
    const allowance = currentAllowance ?? (await token.allowance(account, allowanceTarget));
    if (allowance.gte(minimumAllowance)) {
        return {
            didAlreadyHaveEnough: true,
        };
    }
    const doesImplementPermit = PERMIT_TOKEN_SYMBOLS.includes(tokenInfo.symbol);
    if (doesImplementPermit && signer.provider) {
        const result = await tryPermit(tokenInfo, token, allowanceTarget, signer, signer.provider, minimumAllowance);
        if (result) {
            return result;
        }
    }
    const getApprovalTxToExecute = () => token.approve(allowanceTarget, minimumAllowance);
    return {
        didAlreadyHaveEnough: false,
        getApprovalTxToExecute,
    };
};
const tryPermit = async (tokenInfo, token, allowanceTarget, signer, provider, minimumAllowance) => {
    try {
        await permit(tokenInfo, ERC2612__factory.connect(token.address, signer), allowanceTarget, signer, provider, minimumAllowance);
        return {
            didAlreadyHaveEnough: false,
        };
    }
    catch (error) {
        if (didUserCancelTransaction(error)) {
            throw error;
        }
        // Fall back to the approval implementation.
    }
    return null;
};
const permit = async (tokenInfo, token, allowanceTarget, signer, provider, minimumAllowance) => {
    const implementation = findPermitImplementation(tokenInfo.symbol);
    if (!implementation) {
        throw new Error(`${tokenInfo.symbol} is not configured for permits`);
    }
    const handler = PERMIT_HANDLERS[implementation];
    return handler(token, allowanceTarget, signer, provider, minimumAllowance);
};
const findPermitImplementation = (tokenSymbol) => {
    for (const key in PERMIT_TOKEN_SYMBOL_BY_IMPLEMENTATION) {
        const implementation = key;
        const values = PERMIT_TOKEN_SYMBOL_BY_IMPLEMENTATION[implementation];
        const isCurrentImplementation = values?.includes(tokenSymbol);
        if (isCurrentImplementation) {
            return implementation;
        }
    }
    return null;
};
const permitErc2612 = async (token, allowanceTarget, signer, provider, minimumAllowance) => {
    const account = await signer.getAddress();
    const chainId = await provider.getNetwork().then((n) => n.chainId);
    const nonce = await token.nonces(account);
    const name = await token.name();
    // Note that the version field is not available on all token contracts,
    // so it falls back to 1.
    const version = await token.version().catch((_) => "1");
    const deadline = Math.floor(Date.now() / 1000) + 5 * 60;
    const domain = {
        name: name,
        verifyingContract: token.address,
        version,
        chainId: chainId,
    };
    const types = {
        Permit: [
            { name: "owner", type: "address" },
            { name: "spender", type: "address" },
            { name: "value", type: "uint256" },
            { name: "nonce", type: "uint256" },
            { name: "deadline", type: "uint256" },
        ],
    };
    // Define transaction
    const values = {
        owner: account,
        spender: allowanceTarget,
        value: minimumAllowance,
        nonce,
        deadline,
    };
    const jsonRpcSigner = signer;
    if (!jsonRpcSigner._signTypedData) {
        throw new Error("_signTypedData is not available on the signer");
    }
    const signature = await jsonRpcSigner
        ._signTypedData(domain, types, values)
        .then((signature) => splitSignature(signature));
    const recovered = ethers.utils.verifyTypedData(domain, types, values, signature);
    if (recovered.toLowerCase() != account.toLowerCase()) {
        throw new Error("permit signature is invalid");
    }
    await executePermit({
        tokenAddress: token.address,
        tokenHolderAddress: account,
        signature,
        allowanceTarget,
        minimumAllowance,
        deadline,
        signer,
    });
};
export const PERMIT_HANDLERS = {
    ERC2612: permitErc2612,
};
const executePermit = async (args) => {
    if (config.sdk.shouldUseCacheServer) {
        const success = await tryExecutePermitRemotely(args);
        if (success) {
            return;
        }
    }
    await executePermitLocally(args);
};
const tryExecutePermitRemotely = async (args) => {
    try {
        const endpoint = `${config.api.baseUrl}/permit`;
        const payload = {
            tokenAddress: args.tokenAddress,
            tokenHolderAddress: args.tokenHolderAddress,
            signature: args.signature,
            allowanceTarget: args.allowanceTarget,
            minimumAllowanceHex: args.minimumAllowance.toHexString(),
            deadline: args.deadline,
            chainId: await args.signer.getChainId(),
        };
        const response = await axios.post(endpoint, payload);
        return !!response?.data?.success;
    }
    catch (error) {
        if (config.sdk.printLogs) {
            console.error(error);
        }
        // Fallback to local execution.
    }
    return false;
};
export const executePermitLocally = (args) => ERC2612__factory.connect(args.tokenAddress, args.signer).permit(args.tokenHolderAddress, args.allowanceTarget, args.minimumAllowance, args.deadline, args.signature.v, args.signature.r, args.signature.s);
export const getIsKashiApproved = async (account, network, signer) => {
    const chainId = NETWORK_NAME_TO_CHAIN_ID[network];
    const bentoBoxAddress = BENTOBOX_ADDRESS[chainId];
    const kashiAddress = KASHI_ADDRESS[chainId];
    const contract = SushiBento__factory.connect(bentoBoxAddress, signer);
    return contract.masterContractApproved(kashiAddress, account);
};
export const signKashiApproval = async (account, network, signer) => {
    const chainId = NETWORK_NAME_TO_CHAIN_ID[network];
    const bentoBoxAddress = BENTOBOX_ADDRESS[chainId];
    const kashiAddress = KASHI_ADDRESS[chainId];
    const contract = SushiBento__factory.connect(bentoBoxAddress, signer);
    const warning = "Give FULL access to funds in (and approved to) BentoBox?";
    const nonce = await contract.nonces(account);
    const message = {
        warning,
        user: account,
        masterContract: kashiAddress,
        approved: true,
        nonce,
    };
    const typedData = {
        types: {
            SetMasterContractApproval: [
                { name: "warning", type: "string" },
                { name: "user", type: "address" },
                { name: "masterContract", type: "address" },
                { name: "approved", type: "bool" },
                { name: "nonce", type: "uint256" },
            ],
        },
        primaryType: "SetMasterContractApproval",
        domain: {
            name: "BentoBox V1",
            chainId: chainId,
            verifyingContract: bentoBoxAddress,
        },
        message: message,
    };
    const jsonRpcSigner = signer;
    const signature = await jsonRpcSigner._signTypedData(typedData.domain, typedData.types, typedData.message);
    return ethers.utils.splitSignature(signature);
};
