import { Contract as MultiCallContract, Provider as MultiCallProvider, } from "ethers-multicall";
import { Handle__factory, VaultLibrary__factory, Comptroller__factory, Treasury__factory, GovernanceLock__factory, RewardPool__factory, BalancerVault__factory, } from "../contracts";
import handleAbi from "../abis/handle/Handle.json";
import vaultLibraryAbi from "../abis/handle/VaultLibrary.json";
import comptrollerAbi from "../abis/handle/Comptroller.json";
import fxKeeperPoolAbi from "../abis/handle/fxKeeperPool.json";
import governanceLockAbi from "../abis/handle/GovernanceLock.json";
import rewardPoolAbi from "../abis/handle/RewardPool.json";
import erc20Abi from "../abis/ERC20.json";
import sdkConfig, { BALANCER_VAULT_ADDRESS_ARBITRUM, } from "../config";
import { getProvider } from "./web3";
import { NETWORK_NAME_TO_CHAIN_ID } from "../constants";
import { parseMainNetwork } from "../types/network";
export const createMultiCallContract = (address, abi) => {
    if (!Array.isArray(abi))
        abi = abi.abi;
    return new MultiCallContract(address, abi);
};
export const createERC20MulticallContract = (address) => createMultiCallContract(address, erc20Abi);
export const getMulticallProtocolContracts = (signer, network = "arbitrum") => {
    const protocol = sdkConfig.protocol[parseMainNetwork(network)]?.protocol;
    return createMulticallProtocolContracts(protocol, NETWORK_NAME_TO_CHAIN_ID[network], signer ?? sdkConfig.providers[network]);
};
export const createMulticallProtocolContracts = (protocolAddresses, chainId, signerOrProvider) => {
    const provider = new MultiCallProvider(getProvider(signerOrProvider), chainId);
    const contracts = {
        handle: createMultiCallContract(protocolAddresses.handle, handleAbi.abi),
        vaultLibrary: createMultiCallContract(protocolAddresses.vaultLibrary, vaultLibraryAbi.abi),
        comptroller: createMultiCallContract(protocolAddresses.comptroller, comptrollerAbi),
        fxKeeperPool: createMultiCallContract(protocolAddresses.fxKeeperPool, fxKeeperPoolAbi),
        governanceLock: createMultiCallContract(protocolAddresses.governanceLock, governanceLockAbi),
        governanceLockRetired: createMultiCallContract(protocolAddresses.governanceLockRetired, governanceLockAbi),
        rewardPool: createMultiCallContract(protocolAddresses.rewardPool, rewardPoolAbi),
    };
    return { provider, contracts };
};
export const getCdpContracts = (protocolAddresses, signer) => {
    return {
        handle: Handle__factory.connect(protocolAddresses.handle, signer),
        vaultLibrary: VaultLibrary__factory.connect(protocolAddresses.vaultLibrary, signer),
        comptroller: Comptroller__factory.connect(protocolAddresses.comptroller, signer),
        treasury: Treasury__factory.connect(protocolAddresses.treasury, signer),
    };
};
export const getGovernanceLockContract = (signerOrProvider, isRetired = false) => GovernanceLock__factory.connect(!isRetired
    ? sdkConfig.protocol.arbitrum.protocol.governanceLock
    : sdkConfig.protocol.arbitrum.protocol.governanceLockRetired, signerOrProvider ?? sdkConfig.providers.arbitrum);
export const getRewardPoolContract = (signerOrProvider) => RewardPool__factory.connect(sdkConfig.protocol.arbitrum.protocol.rewardPool, signerOrProvider ?? sdkConfig.providers.arbitrum);
export const callMulticallObject = async (callObject, provider) => {
    const properties = Object.keys(callObject);
    const calls = Object.values(callObject);
    const response = await provider.all(calls);
    return properties.reduce((progress, property, index) => {
        return {
            ...progress,
            [property]: response[index],
        };
    }, {});
};
export const callMulticallObjects = async (callObjects, provider) => {
    const calls = callObjects.reduce((progress, callObject) => [
        ...progress,
        ...Object.values(callObject),
    ], []);
    const response = await provider.all(calls);
    return multicallResponsesToObjects(callObjects, response);
};
export const fetchPromisifiedObject = async (object) => {
    const properties = Object.keys(object);
    const promises = Object.values(object);
    const response = await Promise.all(promises);
    return properties.reduce((progress, property, index) => ({
        ...progress,
        [property]: response[index],
    }), {});
};
export const multicallResponsesToObjects = (callObjects, results) => {
    let resultIndex = 0;
    return callObjects.map((callObject) => {
        const properties = Object.keys(callObject);
        return properties.reduce((progress, property) => {
            const newProgess = {
                ...progress,
                [property]: results[resultIndex],
            };
            resultIndex++;
            return newProgess;
        }, {});
    });
};
export const getBalancerVaultContract = () => BalancerVault__factory.connect(BALANCER_VAULT_ADDRESS_ARBITRUM, sdkConfig.providers.arbitrum);
