import axios from "axios";
import { ethers } from "ethers";
import config from "../config";
import { CHAIN_ID_TO_NETWORK_NAME } from "../constants";
import { fetchUnsignedQuote } from "../components/h2so/fetcher";
export const getCoingeckoPrices = async (tokens, network) => {
    if (tokens.length === 0)
        return {};
    const query = {
        contract_addresses: tokens.join(","),
        vs_currencies: "usd",
    };
    try {
        const { data } = await axios.get(`${config.coingecko.tokenPriceUrl}${config.coingecko.networkAlias[network]}`, { params: query });
        return Object.entries(data).reduce((prev, current) => {
            return {
                ...prev,
                [current[0].toLowerCase()]: current[1].usd,
            };
        }, {});
    }
    catch (e) {
        // if there is an error, it shouldn't stop the quote by erroring, but no price can be used
        console.error(e);
        return {};
    }
};
/**
 * Gets the usd price of a token
 * @param token the token for which to get the usd value
 * @returns the usd price of a token, or undefined if it cannot be fetched
 */
export const getUsdPrice = async (token) => {
    if (token.extensions?.isFxToken ||
        token.extensions?.isHlpToken ||
        token.extensions?.isNative) {
        try {
            const price = await fetchUnsignedQuote({
                baseSymbol: token.symbol,
                quoteSymbol: "USD",
            });
            return price / 1e8;
        }
        catch (e) {
            // this will leave price as undefined on error
            console.error(e);
            return;
        }
    }
    else {
        const response = await getCoingeckoPrices([token.address], CHAIN_ID_TO_NETWORK_NAME[token.chainId]);
        return response[token.address.toLowerCase()];
    }
};
/**
 * Gets the usd value for an amount of a token
 * @param token the token for which to get the usd value
 * @param amount the amount of token for which to get the value of
 * @returns the usd value of the amount of token, or undefined if a price cannot be fetched
 */
export const getUsdValue = async (token, amount) => {
    const price = await getUsdPrice(token);
    if (!price)
        return;
    const precision = 1e10; // 10 decimal places
    const adjustedAmount = amount
        .mul(precision)
        .div(ethers.utils.parseUnits("1", token.decimals));
    return (+adjustedAmount * price) / precision;
};
/**
 * Returns the usd value difference as a ratio.
 * @param usdValueIn the usd value initially
 * @param usdValueOut the usd value output
 * @returns usd value difference as a ratio
 */
export const getPriceImpact = (usdValueIn, usdValueOut) => 1 - usdValueOut / usdValueIn;
